@@ -15,6 +15,8 @@
  */
 package io.netty.buffer;
 
+import io.netty.util.ByteProcessor;
+import io.netty.util.CharsetUtil;
 import io.netty.util.IllegalReferenceCountException;
 import io.netty.util.ResourceLeakDetector;
 import io.netty.util.ResourceLeakDetectorFactory;
@@ -29,6 +31,7 @@ import java.io.InputStream;
 import java.io.OutputStream;
 import java.nio.ByteBuffer;
 import java.nio.ByteOrder;
+import java.nio.channels.FileChannel;
 import java.nio.channels.GatheringByteChannel;
 import java.nio.channels.ScatteringByteChannel;
 import java.nio.charset.Charset;
@@ -57,11 +60,8 @@ public abstract class AbstractByteBuf extends ByteBuf {
     int writerIndex;
     private int markedReaderIndex;
     private int markedWriterIndex;
-
     private int maxCapacity;
 
-    private SwappedByteBuf swappedBuf;
-
     protected AbstractByteBuf(int maxCapacity) {
         if (maxCapacity < 0) {
             throw new IllegalArgumentException("maxCapacity: " + maxCapacity + " (expected: >= 0)");
@@ -69,6 +69,20 @@ public abstract class AbstractByteBuf extends ByteBuf {
         this.maxCapacity = maxCapacity;
     }
 
+    @Override
+    public boolean isReadOnly() {
+        return false;
+    }
+
+    @SuppressWarnings("deprecation")
+    @Override
+    public ByteBuf asReadOnly() {
+        if (isReadOnly()) {
+            return this;
+        }
+        return Unpooled.unmodifiableBuffer(this);
+    }
+
     @Override
     public int maxCapacity() {
         return maxCapacity;
@@ -264,7 +278,7 @@ public abstract class AbstractByteBuf extends ByteBuf {
         }
 
         // Normalize the current capacity to the power of 2.
-        int newCapacity = calculateNewCapacity(writerIndex + minWritableBytes);
+        int newCapacity = alloc().calculateNewCapacity(writerIndex + minWritableBytes, maxCapacity);
 
         // Adjust to the new capacity.
         capacity(newCapacity);
@@ -293,41 +307,13 @@ public abstract class AbstractByteBuf extends ByteBuf {
         }
 
         // Normalize the current capacity to the power of 2.
-        int newCapacity = calculateNewCapacity(writerIndex + minWritableBytes);
+        int newCapacity = alloc().calculateNewCapacity(writerIndex + minWritableBytes, maxCapacity);
 
         // Adjust to the new capacity.
         capacity(newCapacity);
         return 2;
     }
 
-    private int calculateNewCapacity(int minNewCapacity) {
-        final int maxCapacity = this.maxCapacity;
-        final int threshold = 1048576 * 4; // 4 MiB page
-
-        if (minNewCapacity == threshold) {
-            return threshold;
-        }
-
-        // If over threshold, do not double but just increase by threshold.
-        if (minNewCapacity > threshold) {
-            int newCapacity = minNewCapacity / threshold * threshold;
-            if (newCapacity > maxCapacity - threshold) {
-                newCapacity = maxCapacity;
-            } else {
-                newCapacity += threshold;
-            }
-            return newCapacity;
-        }
-
-        // Not over threshold. Double up to 4 MiB, starting from 64.
-        int newCapacity = 64;
-        while (newCapacity < minNewCapacity) {
-            newCapacity <<= 1;
-        }
-
-        return Math.min(newCapacity, maxCapacity);
-    }
-
     @Override
     public ByteBuf order(ByteOrder endianness) {
         if (endianness == null) {
@@ -336,12 +322,7 @@ public abstract class AbstractByteBuf extends ByteBuf {
         if (endianness == order()) {
             return this;
         }
-
-        SwappedByteBuf swappedBuf = this.swappedBuf;
-        if (swappedBuf == null) {
-            this.swappedBuf = swappedBuf = newSwappedByteBuf();
-        }
-        return swappedBuf;
+        return newSwappedByteBuf();
     }
 
     /**
@@ -377,11 +358,24 @@ public abstract class AbstractByteBuf extends ByteBuf {
 
     protected abstract short _getShort(int index);
 
+    @Override
+    public short getShortLE(int index) {
+        checkIndex(index, 2);
+        return _getShortLE(index);
+    }
+
+    protected abstract short _getShortLE(int index);
+
     @Override
     public int getUnsignedShort(int index) {
         return getShort(index) & 0xFFFF;
     }
 
+    @Override
+    public int getUnsignedShortLE(int index) {
+        return getShortLE(index) & 0xFFFF;
+    }
+
     @Override
     public int getUnsignedMedium(int index) {
         checkIndex(index, 3);
@@ -390,6 +384,14 @@ public abstract class AbstractByteBuf extends ByteBuf {
 
     protected abstract int _getUnsignedMedium(int index);
 
+    @Override
+    public int getUnsignedMediumLE(int index) {
+        checkIndex(index, 3);
+        return _getUnsignedMediumLE(index);
+    }
+
+    protected abstract int _getUnsignedMediumLE(int index);
+
     @Override
     public int getMedium(int index) {
         int value = getUnsignedMedium(index);
@@ -399,6 +401,15 @@ public abstract class AbstractByteBuf extends ByteBuf {
         return value;
     }
 
+    @Override
+    public int getMediumLE(int index) {
+        int value = getUnsignedMediumLE(index);
+        if ((value & 0x800000) != 0) {
+            value |= 0xff000000;
+        }
+        return value;
+    }
+
     @Override
     public int getInt(int index) {
         checkIndex(index, 4);
@@ -407,11 +418,24 @@ public abstract class AbstractByteBuf extends ByteBuf {
 
     protected abstract int _getInt(int index);
 
+    @Override
+    public int getIntLE(int index) {
+        checkIndex(index, 4);
+        return _getIntLE(index);
+    }
+
+    protected abstract int _getIntLE(int index);
+
     @Override
     public long getUnsignedInt(int index) {
         return getInt(index) & 0xFFFFFFFFL;
     }
 
+    @Override
+    public long getUnsignedIntLE(int index) {
+        return getIntLE(index) & 0xFFFFFFFFL;
+    }
+
     @Override
     public long getLong(int index) {
         checkIndex(index, 8);
@@ -420,6 +444,14 @@ public abstract class AbstractByteBuf extends ByteBuf {
 
     protected abstract long _getLong(int index);
 
+    @Override
+    public long getLongLE(int index) {
+        checkIndex(index, 8);
+        return _getLongLE(index);
+    }
+
+    protected abstract long _getLongLE(int index);
+
     @Override
     public char getChar(int index) {
         return (char) getShort(index);
@@ -454,6 +486,19 @@ public abstract class AbstractByteBuf extends ByteBuf {
         return this;
     }
 
+    @Override
+    public CharSequence getCharSequence(int index, int length, Charset charset) {
+        // TODO: We could optimize this for UTF8 and US_ASCII
+        return toString(index, length, charset);
+    }
+
+    @Override
+    public CharSequence readCharSequence(int length, Charset charset) {
+        CharSequence sequence = getCharSequence(readerIndex, length, charset);
+        readerIndex += length;
+        return sequence;
+    }
+
     @Override
     public ByteBuf setByte(int index, int value) {
         checkIndex(index);
@@ -478,6 +523,15 @@ public abstract class AbstractByteBuf extends ByteBuf {
 
     protected abstract void _setShort(int index, int value);
 
+    @Override
+    public ByteBuf setShortLE(int index, int value) {
+        checkIndex(index, 2);
+        _setShortLE(index, value);
+        return this;
+    }
+
+    protected abstract void _setShortLE(int index, int value);
+
     @Override
     public ByteBuf setChar(int index, int value) {
         setShort(index, value);
@@ -493,6 +547,15 @@ public abstract class AbstractByteBuf extends ByteBuf {
 
     protected abstract void _setMedium(int index, int value);
 
+    @Override
+    public ByteBuf setMediumLE(int index, int value) {
+        checkIndex(index, 3);
+        _setMediumLE(index, value);
+        return this;
+    }
+
+    protected abstract void _setMediumLE(int index, int value);
+
     @Override
     public ByteBuf setInt(int index, int value) {
         checkIndex(index, 4);
@@ -502,6 +565,15 @@ public abstract class AbstractByteBuf extends ByteBuf {
 
     protected abstract void _setInt(int index, int value);
 
+    @Override
+    public ByteBuf setIntLE(int index, int value) {
+        checkIndex(index, 4);
+        _setIntLE(index, value);
+        return this;
+    }
+
+    protected abstract void _setIntLE(int index, int value);
+
     @Override
     public ByteBuf setFloat(int index, float value) {
         setInt(index, Float.floatToRawIntBits(value));
@@ -517,6 +589,15 @@ public abstract class AbstractByteBuf extends ByteBuf {
 
     protected abstract void _setLong(int index, long value);
 
+    @Override
+    public ByteBuf setLongLE(int index, long value) {
+        checkIndex(index, 8);
+        _setLongLE(index, value);
+        return this;
+    }
+
+    protected abstract void _setLongLE(int index, long value);
+
     @Override
     public ByteBuf setDouble(int index, double value) {
         setLong(index, Double.doubleToRawLongBits(value));
@@ -584,6 +665,23 @@ public abstract class AbstractByteBuf extends ByteBuf {
         return this;
     }
 
+    @Override
+    public int setCharSequence(int index, CharSequence sequence, Charset charset) {
+        if (charset.equals(CharsetUtil.UTF_8)) {
+            ensureWritable(ByteBufUtil.utf8MaxBytes(sequence));
+            return ByteBufUtil.writeUtf8(this, index, sequence, sequence.length());
+        }
+        if (charset.equals(CharsetUtil.US_ASCII)) {
+            int len = sequence.length();
+            ensureWritable(len);
+            return ByteBufUtil.writeAscii(this, index, sequence, len);
+        }
+        byte[] bytes = sequence.toString().getBytes(charset);
+        ensureWritable(bytes.length);
+        setBytes(index, bytes);
+        return bytes.length;
+    }
+
     @Override
     public byte readByte() {
         checkReadableBytes0(1);
@@ -611,11 +709,24 @@ public abstract class AbstractByteBuf extends ByteBuf {
         return v;
     }
 
+    @Override
+    public short readShortLE() {
+        checkReadableBytes0(2);
+        short v = _getShortLE(readerIndex);
+        readerIndex += 2;
+        return v;
+    }
+
     @Override
     public int readUnsignedShort() {
         return readShort() & 0xFFFF;
     }
 
+    @Override
+    public int readUnsignedShortLE() {
+        return readShortLE() & 0xFFFF;
+    }
+
     @Override
     public int readMedium() {
         int value = readUnsignedMedium();
@@ -625,6 +736,15 @@ public abstract class AbstractByteBuf extends ByteBuf {
         return value;
     }
 
+    @Override
+    public int readMediumLE() {
+        int value = readUnsignedMediumLE();
+        if ((value & 0x800000) != 0) {
+            value |= 0xff000000;
+        }
+        return value;
+    }
+
     @Override
     public int readUnsignedMedium() {
         checkReadableBytes0(3);
@@ -633,6 +753,14 @@ public abstract class AbstractByteBuf extends ByteBuf {
         return v;
     }
 
+    @Override
+    public int readUnsignedMediumLE() {
+        checkReadableBytes0(3);
+        int v = _getUnsignedMediumLE(readerIndex);
+        readerIndex += 3;
+        return v;
+    }
+
     @Override
     public int readInt() {
         checkReadableBytes0(4);
@@ -641,11 +769,24 @@ public abstract class AbstractByteBuf extends ByteBuf {
         return v;
     }
 
+    @Override
+    public int readIntLE() {
+        checkReadableBytes0(4);
+        int v = _getIntLE(readerIndex);
+        readerIndex += 4;
+        return v;
+    }
+
     @Override
     public long readUnsignedInt() {
         return readInt() & 0xFFFFFFFFL;
     }
 
+    @Override
+    public long readUnsignedIntLE() {
+        return readIntLE() & 0xFFFFFFFFL;
+    }
+
     @Override
     public long readLong() {
         checkReadableBytes0(8);
@@ -654,6 +795,14 @@ public abstract class AbstractByteBuf extends ByteBuf {
         return v;
     }
 
+    @Override
+    public long readLongLE() {
+        checkReadableBytes0(8);
+        long v = _getLongLE(readerIndex);
+        readerIndex += 8;
+        return v;
+    }
+
     @Override
     public char readChar() {
         return (char) readShort();
@@ -689,6 +838,13 @@ public abstract class AbstractByteBuf extends ByteBuf {
         return slice;
     }
 
+    @Override
+    public ByteBuf readRetainedSlice(int length) {
+        ByteBuf slice = retainedSlice(readerIndex, length);
+        readerIndex += length;
+        return slice;
+    }
+
     @Override
     public ByteBuf readBytes(byte[] dst, int dstIndex, int length) {
         checkReadableBytes(length);
@@ -746,6 +902,15 @@ public abstract class AbstractByteBuf extends ByteBuf {
         return readBytes;
     }
 
+    @Override
+    public int readBytes(FileChannel out, long position, int length)
+            throws IOException {
+        checkReadableBytes(length);
+        int readBytes = getBytes(readerIndex, out, position, length);
+        readerIndex += readBytes;
+        return readBytes;
+    }
+
     @Override
     public ByteBuf readBytes(OutputStream out, int length) throws IOException {
         checkReadableBytes(length);
@@ -784,6 +949,15 @@ public abstract class AbstractByteBuf extends ByteBuf {
         return this;
     }
 
+    @Override
+    public ByteBuf writeShortLE(int value) {
+        ensureAccessible();
+        ensureWritable0(2);
+        _setShortLE(writerIndex, value);
+        writerIndex += 2;
+        return this;
+    }
+
     @Override
     public ByteBuf writeMedium(int value) {
         ensureAccessible();
@@ -793,6 +967,15 @@ public abstract class AbstractByteBuf extends ByteBuf {
         return this;
     }
 
+    @Override
+    public ByteBuf writeMediumLE(int value) {
+        ensureAccessible();
+        ensureWritable0(3);
+        _setMediumLE(writerIndex, value);
+        writerIndex += 3;
+        return this;
+    }
+
     @Override
     public ByteBuf writeInt(int value) {
         ensureAccessible();
@@ -802,6 +985,15 @@ public abstract class AbstractByteBuf extends ByteBuf {
         return this;
     }
 
+    @Override
+    public ByteBuf writeIntLE(int value) {
+        ensureAccessible();
+        ensureWritable0(4);
+        _setIntLE(writerIndex, value);
+        writerIndex += 4;
+        return this;
+    }
+
     @Override
     public ByteBuf writeLong(long value) {
         ensureAccessible();
@@ -811,6 +1003,15 @@ public abstract class AbstractByteBuf extends ByteBuf {
         return this;
     }
 
+    @Override
+    public ByteBuf writeLongLE(long value) {
+        ensureAccessible();
+        ensureWritable0(8);
+        _setLongLE(writerIndex, value);
+        writerIndex += 8;
+        return this;
+    }
+
     @Override
     public ByteBuf writeChar(int value) {
         writeShort(value);
@@ -903,6 +1104,17 @@ public abstract class AbstractByteBuf extends ByteBuf {
         return writtenBytes;
     }
 
+    @Override
+    public int writeBytes(FileChannel in, long position, int length) throws IOException {
+        ensureAccessible();
+        ensureWritable(length);
+        int writtenBytes = setBytes(writerIndex, in, position, length);
+        if (writtenBytes > 0) {
+            writerIndex += writtenBytes;
+        }
+        return writtenBytes;
+    }
+
     @Override
     public ByteBuf writeZero(int length) {
         if (length == 0) {
@@ -939,6 +1151,13 @@ public abstract class AbstractByteBuf extends ByteBuf {
         return this;
     }
 
+    @Override
+    public int writeCharSequence(CharSequence sequence, Charset charset) {
+        int written = setCharSequence(writerIndex, sequence, charset);
+        writerIndex += written;
+        return written;
+    }
+
     @Override
     public ByteBuf copy() {
         return copy(readerIndex, readableBytes());
@@ -946,7 +1165,12 @@ public abstract class AbstractByteBuf extends ByteBuf {
 
     @Override
     public ByteBuf duplicate() {
-        return new DuplicatedAbstractByteBuf(this);
+        return new UnpooledDuplicatedByteBuf(this);
+    }
+
+    @Override
+    public ByteBuf retainedDuplicate() {
+        return duplicate().retain();
     }
 
     @Override
@@ -954,9 +1178,19 @@ public abstract class AbstractByteBuf extends ByteBuf {
         return slice(readerIndex, readableBytes());
     }
 
+    @Override
+    public ByteBuf retainedSlice() {
+        return slice().retain();
+    }
+
     @Override
     public ByteBuf slice(int index, int length) {
-        return new SlicedAbstractByteBuf(this, index, length);
+        return new UnpooledSlicedByteBuf(this, index, length);
+    }
+
+    @Override
+    public ByteBuf retainedSlice(int index, int length) {
+        return slice(index, length).retain();
     }
 
     @Override
@@ -1005,7 +1239,7 @@ public abstract class AbstractByteBuf extends ByteBuf {
     }
 
     @Override
-    public int forEachByte(ByteBufProcessor processor) {
+    public int forEachByte(ByteProcessor processor) {
         ensureAccessible();
         try {
             return forEachByteAsc0(readerIndex, writerIndex, processor);
@@ -1016,7 +1250,7 @@ public abstract class AbstractByteBuf extends ByteBuf {
     }
 
     @Override
-    public int forEachByte(int index, int length, ByteBufProcessor processor) {
+    public int forEachByte(int index, int length, ByteProcessor processor) {
         checkIndex(index, length);
         try {
             return forEachByteAsc0(index, index + length, processor);
@@ -1026,7 +1260,7 @@ public abstract class AbstractByteBuf extends ByteBuf {
         }
     }
 
-    private int forEachByteAsc0(int start, int end, ByteBufProcessor processor) throws Exception {
+    private int forEachByteAsc0(int start, int end, ByteProcessor processor) throws Exception {
         for (; start < end; ++start) {
             if (!processor.process(_getByte(start))) {
                 return start;
@@ -1037,10 +1271,10 @@ public abstract class AbstractByteBuf extends ByteBuf {
     }
 
     @Override
-    public int forEachByteDesc(ByteBufProcessor processor) {
+    public int forEachByteDesc(ByteProcessor processor) {
         ensureAccessible();
         try {
-            return forEachByteDesc0(writerIndex, readerIndex, processor);
+            return forEachByteDesc0(writerIndex - 1, readerIndex, processor);
         } catch (Exception e) {
             PlatformDependent.throwException(e);
             return -1;
@@ -1048,7 +1282,7 @@ public abstract class AbstractByteBuf extends ByteBuf {
     }
 
     @Override
-    public int forEachByteDesc(int index, int length, ByteBufProcessor processor) {
+    public int forEachByteDesc(int index, int length, ByteProcessor processor) {
         checkIndex(index, length);
         try {
             return forEachByteDesc0(index + length - 1, index, processor);
@@ -1058,7 +1292,7 @@ public abstract class AbstractByteBuf extends ByteBuf {
         }
     }
 
-    private int forEachByteDesc0(int rStart, final int rEnd, ByteBufProcessor processor) throws Exception {
+    private int forEachByteDesc0(int rStart, final int rEnd, ByteProcessor processor) throws Exception {
         for (; rStart >= rEnd; --rStart) {
             if (!processor.process(_getByte(rStart))) {
                 return rStart;
@@ -1074,13 +1308,7 @@ public abstract class AbstractByteBuf extends ByteBuf {
 
     @Override
     public boolean equals(Object o) {
-        if (this == o) {
-            return true;
-        }
-        if (o instanceof ByteBuf) {
-            return ByteBufUtil.equals(this, (ByteBuf) o);
-        }
-        return false;
+        return this == o || (o instanceof ByteBuf && ByteBufUtil.equals(this, (ByteBuf) o));
     }
 
     @Override
