@@ -15,6 +15,8 @@
  */
 package io.netty.buffer;
 
+import io.netty.util.AsciiString;
+import io.netty.util.ByteProcessor;
 import io.netty.util.CharsetUtil;
 import io.netty.util.Recycler;
 import io.netty.util.Recycler.Handle;
@@ -33,6 +35,8 @@ import java.nio.charset.Charset;
 import java.nio.charset.CharsetDecoder;
 import java.nio.charset.CharsetEncoder;
 import java.nio.charset.CoderResult;
+import java.nio.charset.CodingErrorAction;
+import java.util.Arrays;
 import java.util.Locale;
 
 import static io.netty.util.internal.MathUtil.isOutOfBounds;
@@ -54,13 +58,19 @@ public final class ByteBufUtil {
         }
     };
 
+    private static final byte WRITE_UTF_UNKNOWN = (byte) '?';
     private static final int MAX_CHAR_BUFFER_SIZE;
     private static final int THREAD_LOCAL_BUFFER_SIZE;
+    private static final int MAX_BYTES_PER_CHAR_UTF8 =
+            (int) CharsetUtil.encoder(CharsetUtil.UTF_8).maxBytesPerChar();
 
     static final ByteBufAllocator DEFAULT_ALLOCATOR;
 
     static {
-        String allocType = SystemPropertyUtil.get("io.netty.allocator.type", "unpooled").toLowerCase(Locale.US).trim();
+        String allocType = SystemPropertyUtil.get(
+                "io.netty.allocator.type", PlatformDependent.isAndroid() ? "unpooled" : "pooled");
+        allocType = allocType.toLowerCase(Locale.US).trim();
+
         ByteBufAllocator alloc;
         if ("unpooled".equals(allocType)) {
             alloc = UnpooledByteBufAllocator.DEFAULT;
@@ -69,8 +79,8 @@ public final class ByteBufUtil {
             alloc = PooledByteBufAllocator.DEFAULT;
             logger.debug("-Dio.netty.allocator.type: {}", allocType);
         } else {
-            alloc = UnpooledByteBufAllocator.DEFAULT;
-            logger.debug("-Dio.netty.allocator.type: unpooled (unknown: {})", allocType);
+            alloc = PooledByteBufAllocator.DEFAULT;
+            logger.debug("-Dio.netty.allocator.type: pooled (unknown: {})", allocType);
         }
 
         DEFAULT_ALLOCATOR = alloc;
@@ -99,8 +109,6 @@ public final class ByteBufUtil {
     }
 
     /**
-<<<<<<< HEAD
-=======
      * Returns a <a href="http://en.wikipedia.org/wiki/Hex_dump">hex dump</a>
      * of the specified byte array.
      */
@@ -117,7 +125,6 @@ public final class ByteBufUtil {
     }
 
     /**
->>>>>>> e5386b0... Move Hex dump related util from ByteBufUtil to inner class
      * Calculates the hash code of the specified buffer.  This method is
      * useful when implementing a new buffer type.
      */
@@ -153,7 +160,55 @@ public final class ByteBufUtil {
 
     /**
      * Returns {@code true} if and only if the two specified buffers are
-     * identical to each other as described in {@code ChannelBuffer#equals(Object)}.
+     * identical to each other for {@code length} bytes starting at {@code aStartIndex}
+     * index for the {@code a} buffer and {@code bStartIndex} index for the {@code b} buffer.
+     * A more compact way to express this is:
+     * <p>
+     * {@code a[aStartIndex : aStartIndex + length] == b[bStartIndex : bStartIndex + length]}
+     */
+    public static boolean equals(ByteBuf a, int aStartIndex, ByteBuf b, int bStartIndex, int length) {
+        if (aStartIndex < 0 || bStartIndex < 0 || length < 0) {
+            throw new IllegalArgumentException("All indexes and lengths must be non-negative");
+        }
+        if (a.writerIndex() - length < aStartIndex || b.writerIndex() - length < bStartIndex) {
+            return false;
+        }
+
+        final int longCount = length >>> 3;
+        final int byteCount = length & 7;
+
+        if (a.order() == b.order()) {
+            for (int i = longCount; i > 0; i --) {
+                if (a.getLong(aStartIndex) != b.getLong(bStartIndex)) {
+                    return false;
+                }
+                aStartIndex += 8;
+                bStartIndex += 8;
+            }
+        } else {
+            for (int i = longCount; i > 0; i --) {
+                if (a.getLong(aStartIndex) != swapLong(b.getLong(bStartIndex))) {
+                    return false;
+                }
+                aStartIndex += 8;
+                bStartIndex += 8;
+            }
+        }
+
+        for (int i = byteCount; i > 0; i --) {
+            if (a.getByte(aStartIndex) != b.getByte(bStartIndex)) {
+                return false;
+            }
+            aStartIndex ++;
+            bStartIndex ++;
+        }
+
+        return true;
+    }
+
+    /**
+     * Returns {@code true} if and only if the two specified buffers are
+     * identical to each other as described in {@link ByteBuf#equals(Object)}.
      * This method is useful when implementing a new buffer type.
      */
     public static boolean equals(ByteBuf bufferA, ByteBuf bufferB) {
@@ -161,40 +216,7 @@ public final class ByteBufUtil {
         if (aLen != bufferB.readableBytes()) {
             return false;
         }
-
-        final int longCount = aLen >>> 3;
-        final int byteCount = aLen & 7;
-
-        int aIndex = bufferA.readerIndex();
-        int bIndex = bufferB.readerIndex();
-
-        if (bufferA.order() == bufferB.order()) {
-            for (int i = longCount; i > 0; i --) {
-                if (bufferA.getLong(aIndex) != bufferB.getLong(bIndex)) {
-                    return false;
-                }
-                aIndex += 8;
-                bIndex += 8;
-            }
-        } else {
-            for (int i = longCount; i > 0; i --) {
-                if (bufferA.getLong(aIndex) != swapLong(bufferB.getLong(bIndex))) {
-                    return false;
-                }
-                aIndex += 8;
-                bIndex += 8;
-            }
-        }
-
-        for (int i = byteCount; i > 0; i --) {
-            if (bufferA.getByte(aIndex) != bufferB.getByte(bIndex)) {
-                return false;
-            }
-            aIndex ++;
-            bIndex ++;
-        }
-
-        return true;
+        return equals(bufferA, bufferA.readerIndex(), bufferB, bufferB.readerIndex(), aLen);
     }
 
     /**
@@ -207,54 +229,83 @@ public final class ByteBufUtil {
         final int minLength = Math.min(aLen, bLen);
         final int uintCount = minLength >>> 2;
         final int byteCount = minLength & 3;
-
         int aIndex = bufferA.readerIndex();
         int bIndex = bufferB.readerIndex();
 
-        if (bufferA.order() == bufferB.order()) {
-            for (int i = uintCount; i > 0; i --) {
-                long va = bufferA.getUnsignedInt(aIndex);
-                long vb = bufferB.getUnsignedInt(bIndex);
-                if (va > vb) {
-                    return 1;
-                }
-                if (va < vb) {
-                    return -1;
-                }
-                aIndex += 4;
-                bIndex += 4;
+        if (uintCount > 0) {
+            boolean bufferAIsBigEndian = bufferA.order() == ByteOrder.BIG_ENDIAN;
+            final long res;
+            int uintCountIncrement = uintCount << 2;
+
+            if (bufferA.order() == bufferB.order()) {
+                res = bufferAIsBigEndian ? compareUintBigEndian(bufferA, bufferB, aIndex, bIndex, uintCountIncrement) :
+                        compareUintLittleEndian(bufferA, bufferB, aIndex, bIndex, uintCountIncrement);
+            } else {
+                res = bufferAIsBigEndian ? compareUintBigEndianA(bufferA, bufferB, aIndex, bIndex, uintCountIncrement) :
+                        compareUintBigEndianB(bufferA, bufferB, aIndex, bIndex, uintCountIncrement);
             }
-        } else {
-            for (int i = uintCount; i > 0; i --) {
-                long va = bufferA.getUnsignedInt(aIndex);
-                long vb = swapInt(bufferB.getInt(bIndex)) & 0xFFFFFFFFL;
-                if (va > vb) {
-                    return 1;
-                }
-                if (va < vb) {
-                    return -1;
-                }
-                aIndex += 4;
-                bIndex += 4;
+            if (res != 0) {
+                // Ensure we not overflow when cast
+                return (int) Math.min(Integer.MAX_VALUE, Math.max(Integer.MIN_VALUE, res));
             }
+            aIndex += uintCountIncrement;
+            bIndex += uintCountIncrement;
         }
 
-        for (int i = byteCount; i > 0; i --) {
-            short va = bufferA.getUnsignedByte(aIndex);
-            short vb = bufferB.getUnsignedByte(bIndex);
-            if (va > vb) {
-                return 1;
+        for (int aEnd = aIndex + byteCount; aIndex < aEnd; ++aIndex, ++bIndex) {
+            int comp = bufferA.getUnsignedByte(aIndex) - bufferB.getUnsignedByte(bIndex);
+            if (comp != 0) {
+                return comp;
             }
-            if (va < vb) {
-                return -1;
-            }
-            aIndex ++;
-            bIndex ++;
         }
 
         return aLen - bLen;
     }
 
+    private static long compareUintBigEndian(
+            ByteBuf bufferA, ByteBuf bufferB, int aIndex, int bIndex, int uintCountIncrement) {
+        for (int aEnd = aIndex + uintCountIncrement; aIndex < aEnd; aIndex += 4, bIndex += 4) {
+            long comp = bufferA.getUnsignedInt(aIndex) - bufferB.getUnsignedInt(bIndex);
+            if (comp != 0) {
+                return comp;
+            }
+        }
+        return 0;
+    }
+
+    private static long compareUintLittleEndian(
+            ByteBuf bufferA, ByteBuf bufferB, int aIndex, int bIndex, int uintCountIncrement) {
+        for (int aEnd = aIndex + uintCountIncrement; aIndex < aEnd; aIndex += 4, bIndex += 4) {
+            long comp = bufferA.getUnsignedIntLE(aIndex) - bufferB.getUnsignedIntLE(bIndex);
+            if (comp != 0) {
+                return comp;
+            }
+        }
+        return 0;
+    }
+
+    private static long compareUintBigEndianA(
+            ByteBuf bufferA, ByteBuf bufferB, int aIndex, int bIndex, int uintCountIncrement) {
+        for (int aEnd = aIndex + uintCountIncrement; aIndex < aEnd; aIndex += 4, bIndex += 4) {
+            long comp =  bufferA.getUnsignedInt(aIndex) - bufferB.getUnsignedIntLE(bIndex);
+            if (comp != 0) {
+                return comp;
+            }
+        }
+        return 0;
+    }
+
+    private static long compareUintBigEndianB(
+            ByteBuf bufferA, ByteBuf bufferB, int aIndex, int bIndex, int uintCountIncrement) {
+        for (int aEnd = aIndex + uintCountIncrement; aIndex < aEnd; aIndex += 4, bIndex += 4) {
+            long comp =  bufferA.getUnsignedIntLE(aIndex) - bufferB.getUnsignedInt(bIndex);
+            if (comp != 0) {
+                return comp;
+            }
+        }
+        return 0;
+    }
+
     /**
      * The default implementation of {@link ByteBuf#indexOf(int, int, byte)}.
      * This method is useful when implementing a new buffer type.
@@ -322,7 +373,7 @@ public final class ByteBufUtil {
             return -1;
         }
 
-        return buffer.forEachByte(fromIndex, toIndex - fromIndex, new IndexOfProcessor(value));
+        return buffer.forEachByte(fromIndex, toIndex - fromIndex, new ByteProcessor.IndexOfProcessor(value));
     }
 
     private static int lastIndexOf(ByteBuf buffer, int fromIndex, int toIndex, byte value) {
@@ -331,7 +382,22 @@ public final class ByteBufUtil {
             return -1;
         }
 
-        return buffer.forEachByteDesc(toIndex, fromIndex - toIndex, new IndexOfProcessor(value));
+        return buffer.forEachByteDesc(toIndex, fromIndex - toIndex, new ByteProcessor.IndexOfProcessor(value));
+    }
+
+    /**
+     * Encode a {@link CharSequence} in <a href="http://en.wikipedia.org/wiki/UTF-8">UTF-8</a> and write
+     * it to a {@link ByteBuf} allocated with {@code alloc}.
+     * @param alloc The allocator used to allocate a new {@link ByteBuf}.
+     * @param seq The characters to write into a buffer.
+     * @return The {@link ByteBuf} which contains the <a href="http://en.wikipedia.org/wiki/UTF-8">UTF-8</a> encoded
+     * result.
+     */
+    public static ByteBuf writeUtf8(ByteBufAllocator alloc, CharSequence seq) {
+        // UTF-8 uses max. 3 bytes per char, so calculate the worst case.
+        ByteBuf buf = alloc.buffer(utf8MaxBytes(seq));
+        writeUtf8(buf, seq);
+        return buf;
     }
 
     /**
@@ -341,20 +407,15 @@ public final class ByteBufUtil {
      * This method returns the actual number of bytes written.
      */
     public static int writeUtf8(ByteBuf buf, CharSequence seq) {
-        if (buf == null) {
-            throw new NullPointerException("buf");
-        }
-        if (seq == null) {
-            throw new NullPointerException("seq");
-        }
-        // UTF-8 uses max. 3 bytes per char, so calculate the worst case.
         final int len = seq.length();
-        final int maxSize = len * 3;
-        buf.ensureWritable(maxSize);
+        buf.ensureWritable(utf8MaxBytes(seq));
 
         for (;;) {
             if (buf instanceof AbstractByteBuf) {
-                return writeUtf8((AbstractByteBuf) buf, seq, len);
+                AbstractByteBuf byteBuf = (AbstractByteBuf) buf;
+                int written = writeUtf8(byteBuf, byteBuf.writerIndex, seq, len);
+                byteBuf.writerIndex += written;
+                return written;
             } else if (buf instanceof WrappedByteBuf) {
                 // Unwrap as the wrapped buffer may be an AbstractByteBuf and so we can use fast-path.
                 buf = buf.unwrap();
@@ -367,9 +428,8 @@ public final class ByteBufUtil {
     }
 
     // Fast-Path implementation
-    private static int writeUtf8(AbstractByteBuf buffer, CharSequence seq, int len) {
-        int oldWriterIndex = buffer.writerIndex;
-        int writerIndex = oldWriterIndex;
+    static int writeUtf8(AbstractByteBuf buffer, int writerIndex, CharSequence seq, int len) {
+        int oldWriterIndex = writerIndex;
 
         // We can use the _set methods as these not need to do any index checks and reference checks.
         // This is possible as we called ensureWritable(...) before.
@@ -382,8 +442,8 @@ public final class ByteBufUtil {
                 buffer._setByte(writerIndex++, (byte) (0x80 | (c & 0x3f)));
             } else if (isSurrogate(c)) {
                 if (!Character.isHighSurrogate(c)) {
-                    throw new IllegalArgumentException("Invalid encoding. " +
-                            "Expected high (leading) surrogate at index " + i + " but got " + c);
+                    buffer._setByte(writerIndex++, WRITE_UTF_UNKNOWN);
+                    continue;
                 }
                 final char c2;
                 try {
@@ -392,12 +452,13 @@ public final class ByteBufUtil {
                     // re-throw a more informative exception describing the problem.
                     c2 = seq.charAt(++i);
                 } catch (IndexOutOfBoundsException e) {
-                    throw new IllegalArgumentException("Underflow. " +
-                            "Expected low (trailing) surrogate at index " + i + " but no more characters found.", e);
+                    buffer._setByte(writerIndex++, WRITE_UTF_UNKNOWN);
+                    break;
                 }
                 if (!Character.isLowSurrogate(c2)) {
-                    throw new IllegalArgumentException("Invalid encoding. " +
-                            "Expected low (trailing) surrogate at index " + i + " but got " + c2);
+                    buffer._setByte(writerIndex++, WRITE_UTF_UNKNOWN);
+                    buffer._setByte(writerIndex++, Character.isHighSurrogate(c2) ? WRITE_UTF_UNKNOWN : c2);
+                    continue;
                 }
                 int codePoint = Character.toCodePoint(c, c2);
                 // See http://www.unicode.org/versions/Unicode7.0.0/ch03.pdf#G2630.
@@ -411,11 +472,31 @@ public final class ByteBufUtil {
                 buffer._setByte(writerIndex++, (byte) (0x80 | (c & 0x3f)));
             }
         }
-        // update the writerIndex without any extra checks for performance reasons
-        buffer.writerIndex = writerIndex;
         return writerIndex - oldWriterIndex;
     }
 
+    /**
+     * Returns max bytes length of UTF8 character sequence.
+     */
+    public static int utf8MaxBytes(CharSequence seq) {
+        return seq.length() * MAX_BYTES_PER_CHAR_UTF8;
+    }
+
+    /**
+     * Encode a {@link CharSequence} in <a href="http://en.wikipedia.org/wiki/ASCII">ASCII</a> and write
+     * it to a {@link ByteBuf} allocated with {@code alloc}.
+     * @param alloc The allocator used to allocate a new {@link ByteBuf}.
+     * @param seq The characters to write into a buffer.
+     * @return The {@link ByteBuf} which contains the <a href="http://en.wikipedia.org/wiki/ASCII">ASCII</a> encoded
+     * result.
+     */
+    public static ByteBuf writeAscii(ByteBufAllocator alloc, CharSequence seq) {
+        // ASCII uses 1 byte per char
+        ByteBuf buf = alloc.buffer(seq.length());
+        writeAscii(buf, seq);
+        return buf;
+    }
+
     /**
      * Encode a {@link CharSequence} in <a href="http://en.wikipedia.org/wiki/ASCII">ASCII</a> and write it
      * to a {@link ByteBuf}.
@@ -423,40 +504,39 @@ public final class ByteBufUtil {
      * This method returns the actual number of bytes written.
      */
     public static int writeAscii(ByteBuf buf, CharSequence seq) {
-        if (buf == null) {
-            throw new NullPointerException("buf");
-        }
-        if (seq == null) {
-            throw new NullPointerException("seq");
-        }
         // ASCII uses 1 byte per char
         final int len = seq.length();
         buf.ensureWritable(len);
-        for (;;) {
-            if (buf instanceof AbstractByteBuf) {
-                writeAscii((AbstractByteBuf) buf, seq, len);
-                break;
-            } else if (buf instanceof WrappedByteBuf) {
-                // Unwrap as the wrapped buffer may be an AbstractByteBuf and so we can use fast-path.
-                buf = buf.unwrap();
-            } else {
-                buf.writeBytes(seq.toString().getBytes(CharsetUtil.US_ASCII));
+        if (seq instanceof AsciiString) {
+            AsciiString asciiString = (AsciiString) seq;
+            buf.writeBytes(asciiString.array(), asciiString.arrayOffset(), asciiString.length());
+        } else {
+            for (;;) {
+                if (buf instanceof AbstractByteBuf) {
+                    AbstractByteBuf byteBuf = (AbstractByteBuf) buf;
+                    int written = writeAscii(byteBuf, byteBuf.writerIndex, seq, len);
+                    byteBuf.writerIndex += written;
+                    return written;
+                } else if (buf instanceof WrappedByteBuf) {
+                    // Unwrap as the wrapped buffer may be an AbstractByteBuf and so we can use fast-path.
+                    buf = buf.unwrap();
+                } else {
+                    buf.writeBytes(seq.toString().getBytes(CharsetUtil.US_ASCII));
+                }
             }
         }
         return len;
     }
 
     // Fast-Path implementation
-    private static void writeAscii(AbstractByteBuf buffer, CharSequence seq, int len) {
-        int writerIndex = buffer.writerIndex;
+    static int writeAscii(AbstractByteBuf buffer, int writerIndex, CharSequence seq, int len) {
 
         // We can use the _set methods as these not need to do any index checks and reference checks.
         // This is possible as we called ensureWritable(...) before.
         for (int i = 0; i < len; i++) {
             buffer._setByte(writerIndex++, (byte) seq.charAt(i));
         }
-        // update the writerIndex without any extra checks for performance reasons
-        buffer.writerIndex = writerIndex;
+        return len;
     }
 
     /**
@@ -464,12 +544,26 @@ public final class ByteBufUtil {
      * is allocated via the {@link ByteBufAllocator}.
      */
     public static ByteBuf encodeString(ByteBufAllocator alloc, CharBuffer src, Charset charset) {
-        return encodeString0(alloc, false, src, charset);
+        return encodeString0(alloc, false, src, charset, 0);
     }
 
-    static ByteBuf encodeString0(ByteBufAllocator alloc, boolean enforceHeap, CharBuffer src, Charset charset) {
-        final CharsetEncoder encoder = CharsetUtil.getEncoder(charset);
-        int length = (int) ((double) src.remaining() * encoder.maxBytesPerChar());
+    /**
+     * Encode the given {@link CharBuffer} using the given {@link Charset} into a new {@link ByteBuf} which
+     * is allocated via the {@link ByteBufAllocator}.
+     *
+     * @param alloc The {@link ByteBufAllocator} to allocate {@link ByteBuf}.
+     * @param src The {@link CharBuffer} to encode.
+     * @param charset The specified {@link Charset}.
+     * @param extraCapacity the extra capacity to alloc except the space for decoding.
+     */
+    public static ByteBuf encodeString(ByteBufAllocator alloc, CharBuffer src, Charset charset, int extraCapacity) {
+        return encodeString0(alloc, false, src, charset, extraCapacity);
+    }
+
+    static ByteBuf encodeString0(ByteBufAllocator alloc, boolean enforceHeap, CharBuffer src, Charset charset,
+                                 int extraCapacity) {
+        final CharsetEncoder encoder = CharsetUtil.encoder(charset);
+        int length = (int) ((double) src.remaining() * encoder.maxBytesPerChar()) + extraCapacity;
         boolean release = true;
         final ByteBuf dst;
         if (enforceHeap) {
@@ -504,7 +598,7 @@ public final class ByteBufUtil {
         if (len == 0) {
             return StringUtil.EMPTY_STRING;
         }
-        final CharsetDecoder decoder = CharsetUtil.getDecoder(charset);
+        final CharsetDecoder decoder = CharsetUtil.decoder(charset);
         final int maxLength = (int) ((double) len * decoder.maxCharsPerByte());
         CharBuffer dst = CHAR_BUFFERS.get();
         if (dst.length() < maxLength) {
@@ -549,6 +643,98 @@ public final class ByteBufUtil {
         }
     }
 
+    /**
+     * Returns a cached thread-local direct buffer, if available.
+     *
+     * @return a cached thread-local direct buffer, if available.  {@code null} otherwise.
+     */
+    public static ByteBuf threadLocalDirectBuffer() {
+        if (THREAD_LOCAL_BUFFER_SIZE <= 0) {
+            return null;
+        }
+
+        if (PlatformDependent.hasUnsafe()) {
+            return ThreadLocalUnsafeDirectByteBuf.newInstance();
+        } else {
+            return ThreadLocalDirectByteBuf.newInstance();
+        }
+    }
+
+    /**
+     * Create a copy of the underlying storage from {@code buf} into a byte array.
+     * The copy will start at {@link ByteBuf#readerIndex()} and copy {@link ByteBuf#readableBytes()} bytes.
+     */
+    public static byte[] getBytes(ByteBuf buf) {
+        return getBytes(buf,  buf.readerIndex(), buf.readableBytes());
+    }
+
+    /**
+     * Create a copy of the underlying storage from {@code buf} into a byte array.
+     * The copy will start at {@code start} and copy {@code length} bytes.
+     */
+    public static byte[] getBytes(ByteBuf buf, int start, int length) {
+        return getBytes(buf, start, length, true);
+    }
+
+    /**
+     * Return an array of the underlying storage from {@code buf} into a byte array.
+     * The copy will start at {@code start} and copy {@code length} bytes.
+     * If {@code copy} is true a copy will be made of the memory.
+     * If {@code copy} is false the underlying storage will be shared, if possible.
+     */
+    public static byte[] getBytes(ByteBuf buf, int start, int length, boolean copy) {
+        if (isOutOfBounds(start, length, buf.capacity())) {
+            throw new IndexOutOfBoundsException("expected: " + "0 <= start(" + start + ") <= start + length(" + length
+                    + ") <= " + "buf.capacity(" + buf.capacity() + ')');
+        }
+
+        if (buf.hasArray()) {
+            if (copy || start != 0 || length != buf.capacity()) {
+                int baseOffset = buf.arrayOffset() + start;
+                return Arrays.copyOfRange(buf.array(), baseOffset, baseOffset + length);
+            } else {
+                return buf.array();
+            }
+        }
+
+        byte[] v = new byte[length];
+        buf.getBytes(start, v);
+        return v;
+    }
+
+    /**
+     * Copies the content of {@code src} to a {@link ByteBuf} using {@link ByteBuf#writeBytes(byte[], int, int)}.
+     * @param src The source of the data to copy.
+     * @param srcIdx the starting offset of characters to copy.
+     * @param dst the destination byte array.
+     * @param dstIdx the starting offset in the destination byte array.
+     * @param length the number of characters to copy.
+     */
+    public static void copy(AsciiString src, int srcIdx, ByteBuf dst, int dstIdx, int length) {
+        if (isOutOfBounds(srcIdx, length, src.length())) {
+            throw new IndexOutOfBoundsException("expected: " + "0 <= srcIdx(" + srcIdx + ") <= srcIdx + length("
+                            + length + ") <= srcLen(" + src.length() + ')');
+        }
+
+        checkNotNull(dst, "dst").setBytes(dstIdx, src.array(), srcIdx + src.arrayOffset(), length);
+    }
+
+    /**
+     * Copies the content of {@code src} to a {@link ByteBuf} using {@link ByteBuf#writeBytes(byte[], int, int)}.
+     * @param src The source of the data to copy.
+     * @param srcIdx the starting offset of characters to copy.
+     * @param dst the destination byte array.
+     * @param length the number of characters to copy.
+     */
+    public static void copy(AsciiString src, int srcIdx, ByteBuf dst, int length) {
+        if (isOutOfBounds(srcIdx, length, src.length())) {
+            throw new IndexOutOfBoundsException("expected: " + "0 <= srcIdx(" + srcIdx + ") <= srcIdx + length("
+                            + length + ") <= srcLen(" + src.length() + ')');
+        }
+
+        checkNotNull(dst, "dst").writeBytes(src.array(), srcIdx + src.arrayOffset(), length);
+    }
+
     /**
      * Returns a multi-line hexadecimal dump of the specified {@link ByteBuf} that is easy to read by humans.
      */
@@ -776,29 +962,12 @@ public final class ByteBufUtil {
         }
     }
 
-    /**
-     * Returns a cached thread-local direct buffer, if available.
-     *
-     * @return a cached thread-local direct buffer, if available.  {@code null} otherwise.
-     */
-    public static ByteBuf threadLocalDirectBuffer() {
-        if (THREAD_LOCAL_BUFFER_SIZE <= 0) {
-            return null;
-        }
-
-        if (PlatformDependent.hasUnsafe()) {
-            return ThreadLocalUnsafeDirectByteBuf.newInstance();
-        } else {
-            return ThreadLocalDirectByteBuf.newInstance();
-        }
-    }
-
     static final class ThreadLocalUnsafeDirectByteBuf extends UnpooledUnsafeDirectByteBuf {
 
         private static final Recycler<ThreadLocalUnsafeDirectByteBuf> RECYCLER =
                 new Recycler<ThreadLocalUnsafeDirectByteBuf>() {
                     @Override
-                    protected ThreadLocalUnsafeDirectByteBuf newObject(Handle handle) {
+                    protected ThreadLocalUnsafeDirectByteBuf newObject(Handle<ThreadLocalUnsafeDirectByteBuf> handle) {
                         return new ThreadLocalUnsafeDirectByteBuf(handle);
                     }
                 };
@@ -809,9 +978,9 @@ public final class ByteBufUtil {
             return buf;
         }
 
-        private final Handle handle;
+        private final Handle<ThreadLocalUnsafeDirectByteBuf> handle;
 
-        private ThreadLocalUnsafeDirectByteBuf(Handle handle) {
+        private ThreadLocalUnsafeDirectByteBuf(Handle<ThreadLocalUnsafeDirectByteBuf> handle) {
             super(UnpooledByteBufAllocator.DEFAULT, 256, Integer.MAX_VALUE);
             this.handle = handle;
         }
@@ -822,7 +991,7 @@ public final class ByteBufUtil {
                 super.deallocate();
             } else {
                 clear();
-                RECYCLER.recycle(this, handle);
+                handle.recycle(this);
             }
         }
     }
@@ -831,7 +1000,7 @@ public final class ByteBufUtil {
 
         private static final Recycler<ThreadLocalDirectByteBuf> RECYCLER = new Recycler<ThreadLocalDirectByteBuf>() {
             @Override
-            protected ThreadLocalDirectByteBuf newObject(Handle handle) {
+            protected ThreadLocalDirectByteBuf newObject(Handle<ThreadLocalDirectByteBuf> handle) {
                 return new ThreadLocalDirectByteBuf(handle);
             }
         };
@@ -842,9 +1011,9 @@ public final class ByteBufUtil {
             return buf;
         }
 
-        private final Handle handle;
+        private final Handle<ThreadLocalDirectByteBuf> handle;
 
-        private ThreadLocalDirectByteBuf(Handle handle) {
+        private ThreadLocalDirectByteBuf(Handle<ThreadLocalDirectByteBuf> handle) {
             super(UnpooledByteBufAllocator.DEFAULT, 256, Integer.MAX_VALUE);
             this.handle = handle;
         }
@@ -855,22 +1024,206 @@ public final class ByteBufUtil {
                 super.deallocate();
             } else {
                 clear();
-                RECYCLER.recycle(this, handle);
+                handle.recycle(this);
             }
         }
     }
 
-    private static class IndexOfProcessor implements ByteBufProcessor {
-        private final byte byteToFind;
+    /**
+     * Returns {@code true} if the given {@link ByteBuf} is valid text using the given {@link Charset},
+     * otherwise return {@code false}.
+     *
+     * @param buf The given {@link ByteBuf}.
+     * @param charset The specified {@link Charset}.
+     */
+    public static boolean isText(ByteBuf buf, Charset charset) {
+        return isText(buf, buf.readerIndex(), buf.readableBytes(), charset);
+    }
 
-        public IndexOfProcessor(byte byteToFind) {
-            this.byteToFind = byteToFind;
+    /**
+     * Returns {@code true} if the specified {@link ByteBuf} starting at {@code index} with {@code length} is valid
+     * text using the given {@link Charset}, otherwise return {@code false}.
+     *
+     * @param buf The given {@link ByteBuf}.
+     * @param index The start index of the specified buffer.
+     * @param length The length of the specified buffer.
+     * @param charset The specified {@link Charset}.
+     *
+     * @throws IndexOutOfBoundsException if {@code index} + {@code length} is greater than {@code buf.readableBytes}
+     */
+    public static boolean isText(ByteBuf buf, int index, int length, Charset charset) {
+        checkNotNull(buf, "buf");
+        checkNotNull(charset, "charset");
+        final int maxIndex = buf.readerIndex() + buf.readableBytes();
+        if (index < 0 || length < 0 || index > maxIndex - length) {
+            throw new IndexOutOfBoundsException("index: " + index + " length: " + length);
         }
+        if (charset.equals(CharsetUtil.UTF_8)) {
+            return isUtf8(buf, index, length);
+        } else if (charset.equals(CharsetUtil.US_ASCII)) {
+            return isAscii(buf, index, length);
+        } else {
+            CharsetDecoder decoder = CharsetUtil.decoder(charset, CodingErrorAction.REPORT, CodingErrorAction.REPORT);
+            try {
+                if (buf.nioBufferCount() == 1) {
+                    decoder.decode(buf.internalNioBuffer(index, length));
+                } else {
+                    ByteBuf heapBuffer =  buf.alloc().heapBuffer(length);
+                    try {
+                        heapBuffer.writeBytes(buf, index, length);
+                        decoder.decode(heapBuffer.internalNioBuffer(0, length));
+                    } finally {
+                        heapBuffer.release();
+                    }
+                }
+                return true;
+            } catch (CharacterCodingException ignore) {
+                return false;
+            }
+        }
+    }
 
+    /**
+     * Aborts on a byte which is not a valid ASCII character.
+     */
+    private static final ByteProcessor FIND_NON_ASCII = new ByteProcessor() {
         @Override
         public boolean process(byte value) {
-            return value != byteToFind;
+            return value >= 0;
         }
+    };
+
+    /**
+     * Returns {@code true} if the specified {@link ByteBuf} starting at {@code index} with {@code length} is valid
+     * ASCII text, otherwise return {@code false}.
+     *
+     * @param buf    The given {@link ByteBuf}.
+     * @param index  The start index of the specified buffer.
+     * @param length The length of the specified buffer.
+     */
+    private static boolean isAscii(ByteBuf buf, int index, int length) {
+        return buf.forEachByte(index, length, FIND_NON_ASCII) == -1;
+    }
+
+    /**
+     * Returns {@code true} if the specified {@link ByteBuf} starting at {@code index} with {@code length} is valid
+     * UTF8 text, otherwise return {@code false}.
+     *
+     * @param buf The given {@link ByteBuf}.
+     * @param index The start index of the specified buffer.
+     * @param length The length of the specified buffer.
+     *
+     * @see
+     * <a href=http://www.ietf.org/rfc/rfc3629.txt>UTF-8 Definition</a>
+     *
+     * <pre>
+     * 1. Bytes format of UTF-8
+     *
+     * The table below summarizes the format of these different octet types.
+     * The letter x indicates bits available for encoding bits of the character number.
+     *
+     * Char. number range  |        UTF-8 octet sequence
+     *    (hexadecimal)    |              (binary)
+     * --------------------+---------------------------------------------
+     * 0000 0000-0000 007F | 0xxxxxxx
+     * 0000 0080-0000 07FF | 110xxxxx 10xxxxxx
+     * 0000 0800-0000 FFFF | 1110xxxx 10xxxxxx 10xxxxxx
+     * 0001 0000-0010 FFFF | 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
+     * </pre>
+     *
+     * <pre>
+     * 2. Syntax of UTF-8 Byte Sequences
+     *
+     * UTF8-octets = *( UTF8-char )
+     * UTF8-char   = UTF8-1 / UTF8-2 / UTF8-3 / UTF8-4
+     * UTF8-1      = %x00-7F
+     * UTF8-2      = %xC2-DF UTF8-tail
+     * UTF8-3      = %xE0 %xA0-BF UTF8-tail /
+     *               %xE1-EC 2( UTF8-tail ) /
+     *               %xED %x80-9F UTF8-tail /
+     *               %xEE-EF 2( UTF8-tail )
+     * UTF8-4      = %xF0 %x90-BF 2( UTF8-tail ) /
+     *               %xF1-F3 3( UTF8-tail ) /
+     *               %xF4 %x80-8F 2( UTF8-tail )
+     * UTF8-tail   = %x80-BF
+     * </pre>
+     */
+    private static boolean isUtf8(ByteBuf buf, int index, int length) {
+        final int endIndex = index + length;
+        while (index < endIndex) {
+            byte b1 = buf.getByte(index++);
+            byte b2, b3, b4;
+            if ((b1 & 0x80) == 0) {
+                // 1 byte
+                continue;
+            }
+            if ((b1 & 0xE0) == 0xC0) {
+                // 2 bytes
+                //
+                // Bit/Byte pattern
+                // 110xxxxx    10xxxxxx
+                // C2..DF      80..BF
+                if (index >= endIndex) { // no enough bytes
+                    return false;
+                }
+                b2 = buf.getByte(index++);
+                if ((b2 & 0xC0) != 0x80) { // 2nd byte not starts with 10
+                    return false;
+                }
+                if ((b1 & 0xFF) < 0xC2) { // out of lower bound
+                    return false;
+                }
+            } else if ((b1 & 0xF0) == 0xE0) {
+                // 3 bytes
+                //
+                // Bit/Byte pattern
+                // 1110xxxx    10xxxxxx    10xxxxxx
+                // E0          A0..BF      80..BF
+                // E1..EC      80..BF      80..BF
+                // ED          80..9F      80..BF
+                // E1..EF      80..BF      80..BF
+                if (index > endIndex - 2) { // no enough bytes
+                    return false;
+                }
+                b2 = buf.getByte(index++);
+                b3 = buf.getByte(index++);
+                if ((b2 & 0xC0) != 0x80 || (b3 & 0xC0) != 0x80) { // 2nd or 3rd bytes not start with 10
+                    return false;
+                }
+                if ((b1 & 0x0F) == 0x00 && (b2 & 0xFF) < 0xA0) { // out of lower bound
+                    return false;
+                }
+                if ((b1 & 0x0F) == 0x0D && (b2 & 0xFF) > 0x9F) { // out of upper bound
+                    return false;
+                }
+            } else if ((b1 & 0xF8) == 0xF0) {
+                // 4 bytes
+                //
+                // Bit/Byte pattern
+                // 11110xxx    10xxxxxx    10xxxxxx    10xxxxxx
+                // F0          90..BF      80..BF      80..BF
+                // F1..F3      80..BF      80..BF      80..BF
+                // F4          80..8F      80..BF      80..BF
+                if (index > endIndex - 3) { // no enough bytes
+                    return false;
+                }
+                b2 = buf.getByte(index++);
+                b3 = buf.getByte(index++);
+                b4 = buf.getByte(index++);
+                if ((b2 & 0xC0) != 0x80 || (b3 & 0xC0) != 0x80 || (b4 & 0xC0) != 0x80) {
+                    // 2nd, 3rd or 4th bytes not start with 10
+                    return false;
+                }
+                if ((b1 & 0xFF) > 0xF4 // b1 invalid
+                        || (b1 & 0xFF) == 0xF0 && (b2 & 0xFF) < 0x90    // b2 out of lower bound
+                        || (b1 & 0xFF) == 0xF4 && (b2 & 0xFF) > 0x8F) { // b2 out of upper bound
+                    return false;
+                }
+            } else {
+                return false;
+            }
+        }
+        return true;
     }
 
     private ByteBufUtil() { }
