@@ -23,6 +23,7 @@ import java.io.OutputStream;
 import java.nio.ByteBuffer;
 import java.nio.ByteOrder;
 import java.nio.channels.ClosedChannelException;
+import java.nio.channels.FileChannel;
 import java.nio.channels.GatheringByteChannel;
 import java.nio.channels.ScatteringByteChannel;
 
@@ -232,6 +233,11 @@ public class UnpooledDirectByteBuf extends AbstractReferenceCountedByteBuf {
         return buffer.getShort(index);
     }
 
+    @Override
+    protected short _getShortLE(int index) {
+        return ByteBufUtil.swapShort(buffer.getShort(index));
+    }
+
     @Override
     public int getUnsignedMedium(int index) {
         ensureAccessible();
@@ -240,7 +246,16 @@ public class UnpooledDirectByteBuf extends AbstractReferenceCountedByteBuf {
 
     @Override
     protected int _getUnsignedMedium(int index) {
-        return (getByte(index) & 0xff) << 16 | (getByte(index + 1) & 0xff) << 8 | getByte(index + 2) & 0xff;
+        return (getByte(index) & 0xff)     << 16 |
+               (getByte(index + 1) & 0xff) << 8  |
+               getByte(index + 2) & 0xff;
+    }
+
+    @Override
+    protected int _getUnsignedMediumLE(int index) {
+        return getByte(index) & 0xff             |
+               (getByte(index + 1) & 0xff) << 8  |
+               (getByte(index + 2) & 0xff) << 16;
     }
 
     @Override
@@ -254,6 +269,11 @@ public class UnpooledDirectByteBuf extends AbstractReferenceCountedByteBuf {
         return buffer.getInt(index);
     }
 
+    @Override
+    protected int _getIntLE(int index) {
+        return ByteBufUtil.swapInt(buffer.getInt(index));
+    }
+
     @Override
     public long getLong(int index) {
         ensureAccessible();
@@ -265,6 +285,11 @@ public class UnpooledDirectByteBuf extends AbstractReferenceCountedByteBuf {
         return buffer.getLong(index);
     }
 
+    @Override
+    protected long _getLongLE(int index) {
+        return ByteBufUtil.swapLong(buffer.getLong(index));
+    }
+
     @Override
     public ByteBuf getBytes(int index, ByteBuf dst, int dstIndex, int length) {
         checkDstIndex(index, length, dstIndex, dst.capacity());
@@ -316,19 +341,15 @@ public class UnpooledDirectByteBuf extends AbstractReferenceCountedByteBuf {
     }
 
     private void getBytes(int index, ByteBuffer dst, boolean internal) {
-        checkIndex(index);
-        if (dst == null) {
-            throw new NullPointerException("dst");
-        }
+        checkIndex(index, dst.remaining());
 
-        int bytesToCopy = Math.min(capacity() - index, dst.remaining());
         ByteBuffer tmpBuf;
         if (internal) {
             tmpBuf = internalNioBuffer();
         } else {
             tmpBuf = buffer.duplicate();
         }
-        tmpBuf.clear().position(index).limit(index + bytesToCopy);
+        tmpBuf.clear().position(index).limit(index + dst.remaining());
         dst.put(tmpBuf);
     }
 
@@ -365,6 +386,11 @@ public class UnpooledDirectByteBuf extends AbstractReferenceCountedByteBuf {
         buffer.putShort(index, (short) value);
     }
 
+    @Override
+    protected void _setShortLE(int index, int value) {
+        buffer.putShort(index, ByteBufUtil.swapShort((short) value));
+    }
+
     @Override
     public ByteBuf setMedium(int index, int value) {
         ensureAccessible();
@@ -379,6 +405,13 @@ public class UnpooledDirectByteBuf extends AbstractReferenceCountedByteBuf {
         setByte(index + 2, (byte) value);
     }
 
+    @Override
+    protected void _setMediumLE(int index, int value) {
+        setByte(index, (byte) value);
+        setByte(index + 1, (byte) (value >>> 8));
+        setByte(index + 2, (byte) (value >>> 16));
+    }
+
     @Override
     public ByteBuf setInt(int index, int value) {
         ensureAccessible();
@@ -391,6 +424,11 @@ public class UnpooledDirectByteBuf extends AbstractReferenceCountedByteBuf {
         buffer.putInt(index, value);
     }
 
+    @Override
+    protected void _setIntLE(int index, int value) {
+        buffer.putInt(index, ByteBufUtil.swapInt(value));
+    }
+
     @Override
     public ByteBuf setLong(int index, long value) {
         ensureAccessible();
@@ -403,6 +441,11 @@ public class UnpooledDirectByteBuf extends AbstractReferenceCountedByteBuf {
         buffer.putLong(index, value);
     }
 
+    @Override
+    protected void _setLongLE(int index, long value) {
+        buffer.putLong(index, ByteBufUtil.swapLong(value));
+    }
+
     @Override
     public ByteBuf setBytes(int index, ByteBuf src, int srcIndex, int length) {
         checkSrcIndex(index, length, srcIndex, src.capacity());
@@ -497,6 +540,22 @@ public class UnpooledDirectByteBuf extends AbstractReferenceCountedByteBuf {
         return out.write(tmpBuf);
     }
 
+    @Override
+    public int getBytes(int index, FileChannel out, long position, int length) throws IOException {
+        return getBytes(index, out, position, length, false);
+    }
+
+    private int getBytes(int index, FileChannel out, long position, int length, boolean internal) throws IOException {
+        ensureAccessible();
+        if (length == 0) {
+            return 0;
+        }
+
+        ByteBuffer tmpBuf = internal ? internalNioBuffer() : buffer.duplicate();
+        tmpBuf.clear().position(index).limit(index + length);
+        return out.write(tmpBuf, position);
+    }
+
     @Override
     public int readBytes(GatheringByteChannel out, int length) throws IOException {
         checkReadableBytes(length);
@@ -505,6 +564,14 @@ public class UnpooledDirectByteBuf extends AbstractReferenceCountedByteBuf {
         return readBytes;
     }
 
+    @Override
+    public int readBytes(FileChannel out, long position, int length) throws IOException {
+        checkReadableBytes(length);
+        int readBytes = getBytes(readerIndex, out, position, length, true);
+        readerIndex += readBytes;
+        return readBytes;
+    }
+
     @Override
     public int setBytes(int index, InputStream in, int length) throws IOException {
         ensureAccessible();
@@ -535,6 +602,18 @@ public class UnpooledDirectByteBuf extends AbstractReferenceCountedByteBuf {
         }
     }
 
+    @Override
+    public int setBytes(int index, FileChannel in, long position, int length) throws IOException {
+        ensureAccessible();
+        ByteBuffer tmpBuf = internalNioBuffer();
+        tmpBuf.clear().position(index).limit(index + length);
+        try {
+            return in.read(tmpNioBuf, position);
+        } catch (ClosedChannelException ignored) {
+            return -1;
+        }
+    }
+
     @Override
     public int nioBufferCount() {
         return 1;
