@@ -26,14 +26,14 @@ import java.nio.ReadOnlyBufferException;
 
 import static io.netty.util.internal.MathUtil.isOutOfBounds;
 import static io.netty.util.internal.ObjectUtil.checkNotNull;
+import static io.netty.util.internal.PlatformDependent.BIG_ENDIAN_NATIVE_ORDER;
 
 /**
  * All operations get and set as {@link ByteOrder#BIG_ENDIAN}.
  */
 final class UnsafeByteBufUtil {
-
-    static final boolean BIG_ENDIAN_NATIVE_ORDER = ByteOrder.nativeOrder() == ByteOrder.BIG_ENDIAN;
     private static final boolean UNALIGNED = PlatformDependent.isUnaligned();
+    private static final byte ZERO = 0;
 
     static byte getByte(long address) {
         return PlatformDependent.getByte(address);
@@ -47,6 +47,14 @@ final class UnsafeByteBufUtil {
         return (short) (PlatformDependent.getByte(address) << 8 | PlatformDependent.getByte(address + 1) & 0xff);
     }
 
+    static short getShortLE(long address) {
+        if (UNALIGNED) {
+            short v = PlatformDependent.getShort(address);
+            return BIG_ENDIAN_NATIVE_ORDER ? Short.reverseBytes(v) : v;
+        }
+        return (short) (PlatformDependent.getByte(address) & 0xff | PlatformDependent.getByte(address + 1) << 8);
+    }
+
     static int getUnsignedMedium(long address) {
         if (UNALIGNED) {
             if (BIG_ENDIAN_NATIVE_ORDER) {
@@ -61,6 +69,20 @@ final class UnsafeByteBufUtil {
                PlatformDependent.getByte(address + 2) & 0xff;
     }
 
+    static int getUnsignedMediumLE(long address) {
+        if (UNALIGNED) {
+            if (BIG_ENDIAN_NATIVE_ORDER) {
+                return (Short.reverseBytes(PlatformDependent.getShort(address)) & 0xffff) << 8 |
+                       PlatformDependent.getByte(address + 2) & 0xff;
+            }
+            return (PlatformDependent.getByte(address)      & 0xff) |
+                   (PlatformDependent.getShort(address + 1) & 0xffff) << 8;
+        }
+        return PlatformDependent.getByte(address)      & 0xff       |
+               (PlatformDependent.getByte(address + 1) & 0xff) << 8 |
+               (PlatformDependent.getByte(address + 1) & 0xff) << 16 ;
+    }
+
     static int getInt(long address) {
         if (UNALIGNED) {
             int v = PlatformDependent.getInt(address);
@@ -72,6 +94,17 @@ final class UnsafeByteBufUtil {
                PlatformDependent.getByte(address + 3) & 0xff;
     }
 
+    static int getIntLE(long address) {
+        if (UNALIGNED) {
+            int v = PlatformDependent.getInt(address);
+            return BIG_ENDIAN_NATIVE_ORDER ? Integer.reverseBytes(v) : v;
+        }
+        return PlatformDependent.getByte(address) & 0xff |
+               (PlatformDependent.getByte(address + 1) & 0xff) <<  8 |
+               (PlatformDependent.getByte(address + 2) & 0xff) << 16 |
+               PlatformDependent.getByte(address + 3) << 24;
+    }
+
     static long getLong(long address) {
         if (UNALIGNED) {
             long v = PlatformDependent.getLong(address);
@@ -87,6 +120,21 @@ final class UnsafeByteBufUtil {
                (long) PlatformDependent.getByte(address + 7) & 0xff;
     }
 
+    static long getLongLE(long address) {
+        if (UNALIGNED) {
+            long v = PlatformDependent.getLong(address);
+            return BIG_ENDIAN_NATIVE_ORDER ? Long.reverseBytes(v) : v;
+        }
+        return (long) PlatformDependent.getByte(address)      & 0xff        |
+               ((long) PlatformDependent.getByte(address + 1) & 0xff) <<  8 |
+               ((long) PlatformDependent.getByte(address + 2) & 0xff) << 16 |
+               ((long) PlatformDependent.getByte(address + 3) & 0xff) << 24 |
+               ((long) PlatformDependent.getByte(address + 4) & 0xff) << 32 |
+               ((long) PlatformDependent.getByte(address + 5) & 0xff) << 40 |
+               ((long) PlatformDependent.getByte(address + 6) & 0xff) << 48 |
+               (long) PlatformDependent.getByte(address  + 7)         << 56;
+    }
+
     static void setByte(long address, int value) {
         PlatformDependent.putByte(address, (byte) value);
     }
@@ -101,6 +149,16 @@ final class UnsafeByteBufUtil {
         }
     }
 
+    static void setShortLE(long address, int value) {
+        if (UNALIGNED) {
+            PlatformDependent.putShort(
+                address, BIG_ENDIAN_NATIVE_ORDER ? Short.reverseBytes((short) value) : (short) value);
+        } else {
+            PlatformDependent.putByte(address, (byte) value);
+            PlatformDependent.putByte(address + 1, (byte) (value >>> 8));
+        }
+    }
+
     static void setMedium(long address, int value) {
         if (UNALIGNED) {
             if (BIG_ENDIAN_NATIVE_ORDER) {
@@ -117,6 +175,22 @@ final class UnsafeByteBufUtil {
         }
     }
 
+    static void setMediumLE(long address, int value) {
+        if (UNALIGNED) {
+            if (BIG_ENDIAN_NATIVE_ORDER) {
+                PlatformDependent.putShort(address, Short.reverseBytes((short) (value >>> 8)));
+                PlatformDependent.putByte(address + 2, (byte) value);
+            } else {
+                PlatformDependent.putByte(address, (byte) value);
+                PlatformDependent.putShort(address + 1, (short) (value >>> 8));
+            }
+        } else {
+            PlatformDependent.putByte(address, (byte) value);
+            PlatformDependent.putByte(address + 1, (byte) (value >>> 8));
+            PlatformDependent.putByte(address + 2, (byte) (value >>> 16));
+        }
+    }
+
     static void setInt(long address, int value) {
         if (UNALIGNED) {
             PlatformDependent.putInt(address, BIG_ENDIAN_NATIVE_ORDER ? value : Integer.reverseBytes(value));
@@ -128,6 +202,17 @@ final class UnsafeByteBufUtil {
         }
     }
 
+    static void setIntLE(long address, int value) {
+        if (UNALIGNED) {
+            PlatformDependent.putInt(address, BIG_ENDIAN_NATIVE_ORDER ? Integer.reverseBytes(value) : value);
+        } else {
+            PlatformDependent.putByte(address, (byte) value);
+            PlatformDependent.putByte(address + 1, (byte) (value >>> 8));
+            PlatformDependent.putByte(address + 2, (byte) (value >>> 16));
+            PlatformDependent.putByte(address + 3, (byte) (value >>> 24));
+        }
+    }
+
     static void setLong(long address, long value) {
         if (UNALIGNED) {
             PlatformDependent.putLong(address, BIG_ENDIAN_NATIVE_ORDER ? value : Long.reverseBytes(value));
@@ -143,6 +228,21 @@ final class UnsafeByteBufUtil {
         }
     }
 
+    static void setLongLE(long address, long value) {
+        if (UNALIGNED) {
+            PlatformDependent.putLong(address, BIG_ENDIAN_NATIVE_ORDER ? Long.reverseBytes(value) : value);
+        } else {
+            PlatformDependent.putByte(address, (byte) value);
+            PlatformDependent.putByte(address + 1, (byte) (value >>> 8));
+            PlatformDependent.putByte(address + 2, (byte) (value >>> 16));
+            PlatformDependent.putByte(address + 3, (byte) (value >>> 24));
+            PlatformDependent.putByte(address + 4, (byte) (value >>> 32));
+            PlatformDependent.putByte(address + 5, (byte) (value >>> 40));
+            PlatformDependent.putByte(address + 6, (byte) (value >>> 48));
+            PlatformDependent.putByte(address + 7, (byte) (value >>> 56));
+        }
+    }
+
     static byte getByte(byte[] array, int index) {
         return PlatformDependent.getByte(array, index);
     }
@@ -152,7 +252,16 @@ final class UnsafeByteBufUtil {
             short v = PlatformDependent.getShort(array, index);
             return BIG_ENDIAN_NATIVE_ORDER ? v : Short.reverseBytes(v);
         }
-        return (short) (PlatformDependent.getByte(index) << 8 | PlatformDependent.getByte(index + 1) & 0xff);
+        return (short) (PlatformDependent.getByte(array, index) << 8 |
+                PlatformDependent.getByte(array, index + 1) & 0xff);
+    }
+
+    static short getShortLE(byte[] array, int index) {
+        if (UNALIGNED) {
+            short v = PlatformDependent.getShort(array, index);
+            return BIG_ENDIAN_NATIVE_ORDER ? Short.reverseBytes(v) : v;
+        }
+        return (short) (PlatformDependent.getByte(index) & 0xff | PlatformDependent.getByte(index + 1) << 8);
     }
 
     static int getUnsignedMedium(byte[] array, int index) {
@@ -169,6 +278,20 @@ final class UnsafeByteBufUtil {
                 PlatformDependent.getByte(array, index + 2) & 0xff;
     }
 
+    static int getUnsignedMediumLE(byte[] array, int index) {
+        if (UNALIGNED) {
+            if (BIG_ENDIAN_NATIVE_ORDER) {
+                return (Short.reverseBytes(PlatformDependent.getShort(array, index)) & 0xffff) << 8 |
+                       PlatformDependent.getByte(array, index + 2) & 0xff;
+            }
+            return (PlatformDependent.getByte(array, index) & 0xff) |
+                   (PlatformDependent.getShort(array, index + 1) & 0xffff) << 8;
+        }
+        return  PlatformDependent.getByte(array, index) & 0xff         |
+                (PlatformDependent.getByte(array, index + 1) & 0xff) <<  8 |
+                (PlatformDependent.getByte(array, index + 2) & 0xff) << 16;
+    }
+
     static int getInt(byte[] array, int index) {
         if (UNALIGNED) {
             int v = PlatformDependent.getInt(array, index);
@@ -180,9 +303,20 @@ final class UnsafeByteBufUtil {
                 PlatformDependent.getByte(array, index + 3) & 0xff;
     }
 
+    static int getIntLE(byte[] array, int index) {
+        if (UNALIGNED) {
+            int v = PlatformDependent.getInt(array, index);
+            return BIG_ENDIAN_NATIVE_ORDER ? Integer.reverseBytes(v) : v;
+        }
+        return PlatformDependent.getByte(array, index)      & 0xff        |
+               (PlatformDependent.getByte(array, index + 1) & 0xff) <<  8 |
+               (PlatformDependent.getByte(array, index + 2) & 0xff) << 16 |
+               PlatformDependent.getByte(array,  index + 3) << 24;
+    }
+
     static long getLong(byte[] array, int index) {
         if (UNALIGNED) {
-            long v =  PlatformDependent.getLong(array, index);
+            long v = PlatformDependent.getLong(array, index);
             return BIG_ENDIAN_NATIVE_ORDER ? v : Long.reverseBytes(v);
         }
         return (long) PlatformDependent.getByte(array, index) << 56 |
@@ -195,6 +329,21 @@ final class UnsafeByteBufUtil {
                 (long) PlatformDependent.getByte(array, index + 7) & 0xff;
     }
 
+    static long getLongLE(byte[] array, int index) {
+        if (UNALIGNED) {
+            long v = PlatformDependent.getLong(array, index);
+            return BIG_ENDIAN_NATIVE_ORDER ? Long.reverseBytes(v) : v;
+        }
+        return (long) PlatformDependent.getByte(array, index)      & 0xff        |
+               ((long) PlatformDependent.getByte(array, index + 1) & 0xff) <<  8 |
+               ((long) PlatformDependent.getByte(array, index + 2) & 0xff) << 16 |
+               ((long) PlatformDependent.getByte(array, index + 3) & 0xff) << 24 |
+               ((long) PlatformDependent.getByte(array, index + 4) & 0xff) << 32 |
+               ((long) PlatformDependent.getByte(array, index + 5) & 0xff) << 40 |
+               ((long) PlatformDependent.getByte(array, index + 6) & 0xff) << 48 |
+               (long) PlatformDependent.getByte(array,  index + 7) << 56;
+    }
+
     static void setByte(byte[] array, int index, int value) {
         PlatformDependent.putByte(array, index, (byte) value);
     }
@@ -209,6 +358,16 @@ final class UnsafeByteBufUtil {
         }
     }
 
+    static void setShortLE(byte[] array, int index, int value) {
+        if (UNALIGNED) {
+            PlatformDependent.putShort(
+                array, index, BIG_ENDIAN_NATIVE_ORDER ? Short.reverseBytes((short) value) : (short) value);
+        } else {
+            PlatformDependent.putByte(array, index, (byte) value);
+            PlatformDependent.putByte(array, index + 1, (byte) (value >>> 8));
+        }
+    }
+
     static void setMedium(byte[] array, int index, int   value) {
         if (UNALIGNED) {
             if (BIG_ENDIAN_NATIVE_ORDER) {
@@ -225,6 +384,22 @@ final class UnsafeByteBufUtil {
         }
     }
 
+    static void setMediumLE(byte[] array, int index, int   value) {
+        if (UNALIGNED) {
+            if (BIG_ENDIAN_NATIVE_ORDER) {
+                PlatformDependent.putShort(array, index, Short.reverseBytes((short) (value >>> 8)));
+                PlatformDependent.putByte(array, index + 2, (byte) value);
+            } else {
+                PlatformDependent.putByte(array, index, (byte) value);
+                PlatformDependent.putShort(array, index + 1, (short) (value >>> 8));
+            }
+        } else {
+            PlatformDependent.putByte(array, index, (byte) value);
+            PlatformDependent.putByte(array, index + 1, (byte) (value >>> 8));
+            PlatformDependent.putByte(array, index + 2, (byte) (value >>> 16));
+        }
+    }
+
     static void setInt(byte[] array, int index, int   value) {
         if (UNALIGNED) {
             PlatformDependent.putInt(array, index, BIG_ENDIAN_NATIVE_ORDER ? value : Integer.reverseBytes(value));
@@ -236,6 +411,17 @@ final class UnsafeByteBufUtil {
         }
     }
 
+    static void setIntLE(byte[] array, int index, int   value) {
+        if (UNALIGNED) {
+            PlatformDependent.putInt(array, index, BIG_ENDIAN_NATIVE_ORDER ? Integer.reverseBytes(value) : value);
+        } else {
+            PlatformDependent.putByte(array, index, (byte) value);
+            PlatformDependent.putByte(array, index + 1, (byte) (value >>> 8));
+            PlatformDependent.putByte(array, index + 2, (byte) (value >>> 16));
+            PlatformDependent.putByte(array, index + 3, (byte) (value >>> 24));
+        }
+    }
+
     static void setLong(byte[] array, int index, long  value) {
         if (UNALIGNED) {
             PlatformDependent.putLong(array, index, BIG_ENDIAN_NATIVE_ORDER ? value : Long.reverseBytes(value));
@@ -251,6 +437,28 @@ final class UnsafeByteBufUtil {
         }
     }
 
+    static void setLongLE(byte[] array, int index, long  value) {
+        if (UNALIGNED) {
+            PlatformDependent.putLong(array, index, BIG_ENDIAN_NATIVE_ORDER ? Long.reverseBytes(value) : value);
+        } else {
+            PlatformDependent.putByte(array, index, (byte) value);
+            PlatformDependent.putByte(array, index + 1, (byte) (value >>> 8));
+            PlatformDependent.putByte(array, index + 2, (byte) (value >>> 16));
+            PlatformDependent.putByte(array, index + 3, (byte) (value >>> 24));
+            PlatformDependent.putByte(array, index + 4, (byte) (value >>> 32));
+            PlatformDependent.putByte(array, index + 5, (byte) (value >>> 40));
+            PlatformDependent.putByte(array, index + 6, (byte) (value >>> 48));
+            PlatformDependent.putByte(array, index + 7, (byte) (value >>> 56));
+        }
+    }
+
+    static void setZero(byte[] array, int index, int length) {
+        if (length == 0) {
+            return;
+        }
+        PlatformDependent.setMemory(array, index, length, ZERO);
+    }
+
     static ByteBuf copy(AbstractByteBuf buf, long addr, int index, int length) {
         buf.checkIndex(index, length);
         ByteBuf copy = buf.alloc().directBuffer(length, buf.maxCapacity());
@@ -309,9 +517,8 @@ final class UnsafeByteBufUtil {
     }
 
     static void getBytes(AbstractByteBuf buf, long addr, int index, ByteBuffer dst) {
-        buf.checkIndex(index);
-        int bytesToCopy = Math.min(buf.capacity() - index, dst.remaining());
-        if (bytesToCopy == 0) {
+        buf.checkIndex(index, dst.remaining());
+        if (dst.remaining() == 0) {
             return;
         }
 
@@ -322,12 +529,12 @@ final class UnsafeByteBufUtil {
             }
             // Copy to direct memory
             long dstAddress = PlatformDependent.directBufferAddress(dst);
-            PlatformDependent.copyMemory(addr, dstAddress + dst.position(), bytesToCopy);
-            dst.position(dst.position() + bytesToCopy);
+            PlatformDependent.copyMemory(addr, dstAddress + dst.position(), dst.remaining());
+            dst.position(dst.position() + dst.remaining());
         } else if (dst.hasArray()) {
             // Copy to array
-            PlatformDependent.copyMemory(addr, dst.array(), dst.arrayOffset() + dst.position(), bytesToCopy);
-            dst.position(dst.position() + bytesToCopy);
+            PlatformDependent.copyMemory(addr, dst.array(), dst.arrayOffset() + dst.position(), dst.remaining());
+            dst.position(dst.position() + dst.remaining());
         } else  {
             dst.put(buf.nioBuffer());
         }
@@ -380,7 +587,7 @@ final class UnsafeByteBufUtil {
             try {
                 byte[] tmp = tmpBuf.array();
                 src.get(tmp, tmpBuf.arrayOffset(), length); // moves the src position too
-                PlatformDependent.copyMemory(tmp, 0, addr, length);
+                PlatformDependent.copyMemory(tmp, tmpBuf.arrayOffset(), addr, length);
             } finally {
                 tmpBuf.release();
             }
@@ -402,5 +609,22 @@ final class UnsafeByteBufUtil {
         }
     }
 
+    static void setZero(AbstractByteBuf buf, long addr, int index, int length) {
+        if (length == 0) {
+            return;
+        }
+
+        buf.checkIndex(index, length);
+        PlatformDependent.setMemory(addr, length, ZERO);
+    }
+
+    static UnpooledUnsafeDirectByteBuf newUnsafeDirectByteBuf(
+            ByteBufAllocator alloc, int initialCapacity, int maxCapacity) {
+        if (PlatformDependent.useDirectBufferNoCleaner()) {
+            return new UnpooledUnsafeNoCleanerDirectByteBuf(alloc, initialCapacity, maxCapacity);
+        }
+        return new UnpooledUnsafeDirectByteBuf(alloc, initialCapacity, maxCapacity);
+    }
+
     private UnsafeByteBufUtil() { }
 }
