@@ -16,6 +16,7 @@
 
 package io.netty.buffer;
 
+import io.netty.util.ByteProcessor;
 import io.netty.util.internal.StringUtil;
 
 import java.io.IOException;
@@ -23,6 +24,7 @@ import java.io.InputStream;
 import java.io.OutputStream;
 import java.nio.ByteBuffer;
 import java.nio.ByteOrder;
+import java.nio.channels.FileChannel;
 import java.nio.channels.GatheringByteChannel;
 import java.nio.channels.ScatteringByteChannel;
 import java.nio.charset.Charset;
@@ -91,6 +93,16 @@ class WrappedByteBuf extends ByteBuf {
         return buf;
     }
 
+    @Override
+    public ByteBuf asReadOnly() {
+        return buf.asReadOnly();
+    }
+
+    @Override
+    public boolean isReadOnly() {
+        return buf.isReadOnly();
+    }
+
     @Override
     public final boolean isDirect() {
         return buf.isDirect();
@@ -222,36 +234,71 @@ class WrappedByteBuf extends ByteBuf {
         return buf.getShort(index);
     }
 
+    @Override
+    public short getShortLE(int index) {
+        return buf.getShortLE(index);
+    }
+
     @Override
     public int getUnsignedShort(int index) {
         return buf.getUnsignedShort(index);
     }
 
+    @Override
+    public int getUnsignedShortLE(int index) {
+        return buf.getUnsignedShortLE(index);
+    }
+
     @Override
     public int getMedium(int index) {
         return buf.getMedium(index);
     }
 
+    @Override
+    public int getMediumLE(int index) {
+        return buf.getMediumLE(index);
+    }
+
     @Override
     public int getUnsignedMedium(int index) {
         return buf.getUnsignedMedium(index);
     }
 
+    @Override
+    public int getUnsignedMediumLE(int index) {
+        return buf.getUnsignedMediumLE(index);
+    }
+
     @Override
     public int getInt(int index) {
         return buf.getInt(index);
     }
 
+    @Override
+    public int getIntLE(int index) {
+        return buf.getIntLE(index);
+    }
+
     @Override
     public long getUnsignedInt(int index) {
         return buf.getUnsignedInt(index);
     }
 
+    @Override
+    public long getUnsignedIntLE(int index) {
+        return buf.getUnsignedIntLE(index);
+    }
+
     @Override
     public long getLong(int index) {
         return buf.getLong(index);
     }
 
+    @Override
+    public long getLongLE(int index) {
+        return buf.getLongLE(index);
+    }
+
     @Override
     public char getChar(int index) {
         return buf.getChar(index);
@@ -314,6 +361,16 @@ class WrappedByteBuf extends ByteBuf {
         return buf.getBytes(index, out, length);
     }
 
+    @Override
+    public int getBytes(int index, FileChannel out, long position, int length) throws IOException {
+        return buf.getBytes(index, out, position, length);
+    }
+
+    @Override
+    public CharSequence getCharSequence(int index, int length, Charset charset) {
+        return buf.getCharSequence(index, length, charset);
+    }
+
     @Override
     public ByteBuf setBoolean(int index, boolean value) {
         buf.setBoolean(index, value);
@@ -332,24 +389,48 @@ class WrappedByteBuf extends ByteBuf {
         return this;
     }
 
+    @Override
+    public ByteBuf setShortLE(int index, int value) {
+        buf.setShortLE(index, value);
+        return this;
+    }
+
     @Override
     public ByteBuf setMedium(int index, int value) {
         buf.setMedium(index, value);
         return this;
     }
 
+    @Override
+    public ByteBuf setMediumLE(int index, int value) {
+        buf.setMediumLE(index, value);
+        return this;
+    }
+
     @Override
     public ByteBuf setInt(int index, int value) {
         buf.setInt(index, value);
         return this;
     }
 
+    @Override
+    public ByteBuf setIntLE(int index, int value) {
+        buf.setIntLE(index, value);
+        return this;
+    }
+
     @Override
     public ByteBuf setLong(int index, long value) {
         buf.setLong(index, value);
         return this;
     }
 
+    @Override
+    public ByteBuf setLongLE(int index, long value) {
+        buf.setLongLE(index, value);
+        return this;
+    }
+
     @Override
     public ByteBuf setChar(int index, int value) {
         buf.setChar(index, value);
@@ -414,12 +495,22 @@ class WrappedByteBuf extends ByteBuf {
         return buf.setBytes(index, in, length);
     }
 
+    @Override
+    public int setBytes(int index, FileChannel in, long position, int length) throws IOException {
+        return buf.setBytes(index, in, position, length);
+    }
+
     @Override
     public ByteBuf setZero(int index, int length) {
         buf.setZero(index, length);
         return this;
     }
 
+    @Override
+    public int setCharSequence(int index, CharSequence sequence, Charset charset) {
+        return buf.setCharSequence(index, sequence, charset);
+    }
+
     @Override
     public boolean readBoolean() {
         return buf.readBoolean();
@@ -440,36 +531,71 @@ class WrappedByteBuf extends ByteBuf {
         return buf.readShort();
     }
 
+    @Override
+    public short readShortLE() {
+        return buf.readShortLE();
+    }
+
     @Override
     public int readUnsignedShort() {
         return buf.readUnsignedShort();
     }
 
+    @Override
+    public int readUnsignedShortLE() {
+        return buf.readUnsignedShortLE();
+    }
+
     @Override
     public int readMedium() {
         return buf.readMedium();
     }
 
+    @Override
+    public int readMediumLE() {
+        return buf.readMediumLE();
+    }
+
     @Override
     public int readUnsignedMedium() {
         return buf.readUnsignedMedium();
     }
 
+    @Override
+    public int readUnsignedMediumLE() {
+        return buf.readUnsignedMediumLE();
+    }
+
     @Override
     public int readInt() {
         return buf.readInt();
     }
 
+    @Override
+    public int readIntLE() {
+        return buf.readIntLE();
+    }
+
     @Override
     public long readUnsignedInt() {
         return buf.readUnsignedInt();
     }
 
+    @Override
+    public long readUnsignedIntLE() {
+        return buf.readUnsignedIntLE();
+    }
+
     @Override
     public long readLong() {
         return buf.readLong();
     }
 
+    @Override
+    public long readLongLE() {
+        return buf.readLongLE();
+    }
+
     @Override
     public char readChar() {
         return buf.readChar();
@@ -495,6 +621,11 @@ class WrappedByteBuf extends ByteBuf {
         return buf.readSlice(length);
     }
 
+    @Override
+    public ByteBuf readRetainedSlice(int length) {
+        return buf.readRetainedSlice(length);
+    }
+
     @Override
     public ByteBuf readBytes(ByteBuf dst) {
         buf.readBytes(dst);
@@ -542,6 +673,16 @@ class WrappedByteBuf extends ByteBuf {
         return buf.readBytes(out, length);
     }
 
+    @Override
+    public int readBytes(FileChannel out, long position, int length) throws IOException {
+        return buf.readBytes(out, position, length);
+    }
+
+    @Override
+    public CharSequence readCharSequence(int length, Charset charset) {
+        return buf.readCharSequence(length, charset);
+    }
+
     @Override
     public ByteBuf skipBytes(int length) {
         buf.skipBytes(length);
@@ -566,24 +707,48 @@ class WrappedByteBuf extends ByteBuf {
         return this;
     }
 
+    @Override
+    public ByteBuf writeShortLE(int value) {
+        buf.writeShortLE(value);
+        return this;
+    }
+
     @Override
     public ByteBuf writeMedium(int value) {
         buf.writeMedium(value);
         return this;
     }
 
+    @Override
+    public ByteBuf writeMediumLE(int value) {
+        buf.writeMediumLE(value);
+        return this;
+    }
+
     @Override
     public ByteBuf writeInt(int value) {
         buf.writeInt(value);
         return this;
     }
 
+    @Override
+    public ByteBuf writeIntLE(int value) {
+        buf.writeIntLE(value);
+        return this;
+    }
+
     @Override
     public ByteBuf writeLong(long value) {
         buf.writeLong(value);
         return this;
     }
 
+    @Override
+    public ByteBuf writeLongLE(long value) {
+        buf.writeLongLE(value);
+        return this;
+    }
+
     @Override
     public ByteBuf writeChar(int value) {
         buf.writeChar(value);
@@ -648,12 +813,22 @@ class WrappedByteBuf extends ByteBuf {
         return buf.writeBytes(in, length);
     }
 
+    @Override
+    public int writeBytes(FileChannel in, long position, int length) throws IOException {
+        return buf.writeBytes(in, position, length);
+    }
+
     @Override
     public ByteBuf writeZero(int length) {
         buf.writeZero(length);
         return this;
     }
 
+    @Override
+    public int writeCharSequence(CharSequence sequence, Charset charset) {
+        return buf.writeCharSequence(sequence, charset);
+    }
+
     @Override
     public int indexOf(int fromIndex, int toIndex, byte value) {
         return buf.indexOf(fromIndex, toIndex, value);
@@ -675,22 +850,22 @@ class WrappedByteBuf extends ByteBuf {
     }
 
     @Override
-    public int forEachByte(ByteBufProcessor processor) {
+    public int forEachByte(ByteProcessor processor) {
         return buf.forEachByte(processor);
     }
 
     @Override
-    public int forEachByte(int index, int length, ByteBufProcessor processor) {
+    public int forEachByte(int index, int length, ByteProcessor processor) {
         return buf.forEachByte(index, length, processor);
     }
 
     @Override
-    public int forEachByteDesc(ByteBufProcessor processor) {
+    public int forEachByteDesc(ByteProcessor processor) {
         return buf.forEachByteDesc(processor);
     }
 
     @Override
-    public int forEachByteDesc(int index, int length, ByteBufProcessor processor) {
+    public int forEachByteDesc(int index, int length, ByteProcessor processor) {
         return buf.forEachByteDesc(index, length, processor);
     }
 
@@ -709,16 +884,31 @@ class WrappedByteBuf extends ByteBuf {
         return buf.slice();
     }
 
+    @Override
+    public ByteBuf retainedSlice() {
+        return buf.retainedSlice();
+    }
+
     @Override
     public ByteBuf slice(int index, int length) {
         return buf.slice(index, length);
     }
 
+    @Override
+    public ByteBuf retainedSlice(int index, int length) {
+        return buf.retainedSlice(index, length);
+    }
+
     @Override
     public ByteBuf duplicate() {
         return buf.duplicate();
     }
 
+    @Override
+    public ByteBuf retainedDuplicate() {
+        return buf.retainedDuplicate();
+    }
+
     @Override
     public int nioBufferCount() {
         return buf.nioBufferCount();
@@ -807,6 +997,18 @@ class WrappedByteBuf extends ByteBuf {
         return this;
     }
 
+    @Override
+    public ByteBuf touch() {
+        buf.touch();
+        return this;
+    }
+
+    @Override
+    public ByteBuf touch(Object hint) {
+        buf.touch(hint);
+        return this;
+    }
+
     @Override
     public final boolean isReadable(int size) {
         return buf.isReadable(size);
