@@ -19,6 +19,7 @@ import io.netty.handler.codec.DecoderResult;
 
 public class DefaultHttpObject implements HttpObject {
 
+    private static final int HASH_CODE_PRIME = 31;
     private DecoderResult decoderResult = DecoderResult.SUCCESS;
 
     protected DefaultHttpObject() {
@@ -26,10 +27,16 @@ public class DefaultHttpObject implements HttpObject {
     }
 
     @Override
-    public DecoderResult getDecoderResult() {
+    public DecoderResult decoderResult() {
         return decoderResult;
     }
 
+    @Override
+    @Deprecated
+    public DecoderResult getDecoderResult() {
+        return decoderResult();
+    }
+
     @Override
     public void setDecoderResult(DecoderResult decoderResult) {
         if (decoderResult == null) {
@@ -37,4 +44,22 @@ public class DefaultHttpObject implements HttpObject {
         }
         this.decoderResult = decoderResult;
     }
+
+    @Override
+    public int hashCode() {
+        int result = 1;
+        result = HASH_CODE_PRIME * result + decoderResult.hashCode();
+        return result;
+    }
+
+    @Override
+    public boolean equals(Object o) {
+        if (!(o instanceof DefaultHttpObject)) {
+            return false;
+        }
+
+        DefaultHttpObject other = (DefaultHttpObject) o;
+
+        return decoderResult().equals(other.decoderResult());
+    }
 }
