@@ -16,14 +16,14 @@
 package io.netty.handler.codec.http;
 
 import io.netty.buffer.ByteBuf;
-import io.netty.buffer.ByteBufProcessor;
 import io.netty.buffer.Unpooled;
 import io.netty.channel.ChannelHandlerContext;
 import io.netty.channel.ChannelPipeline;
-import io.netty.channel.socket.ChannelInputShutdownEvent;
 import io.netty.handler.codec.ByteToMessageDecoder;
 import io.netty.handler.codec.DecoderResult;
+import io.netty.handler.codec.PrematureChannelClosureException;
 import io.netty.handler.codec.TooLongFrameException;
+import io.netty.util.ByteProcessor;
 import io.netty.util.internal.AppendableCharSequence;
 
 import java.util.List;
@@ -162,28 +162,33 @@ public abstract class HttpObjectDecoder extends ByteToMessageDecoder {
     protected HttpObjectDecoder(
             int maxInitialLineLength, int maxHeaderSize, int maxChunkSize,
             boolean chunkedSupported, boolean validateHeaders) {
+        this(maxInitialLineLength, maxHeaderSize, maxChunkSize, chunkedSupported, validateHeaders, 128);
+    }
 
+    protected HttpObjectDecoder(
+            int maxInitialLineLength, int maxHeaderSize, int maxChunkSize,
+            boolean chunkedSupported, boolean validateHeaders, int initialBufferSize) {
         if (maxInitialLineLength <= 0) {
             throw new IllegalArgumentException(
                     "maxInitialLineLength must be a positive integer: " +
-                            maxInitialLineLength);
+                     maxInitialLineLength);
         }
         if (maxHeaderSize <= 0) {
             throw new IllegalArgumentException(
                     "maxHeaderSize must be a positive integer: " +
-                            maxHeaderSize);
+                    maxHeaderSize);
         }
         if (maxChunkSize <= 0) {
             throw new IllegalArgumentException(
                     "maxChunkSize must be a positive integer: " +
-                            maxChunkSize);
+                    maxChunkSize);
         }
+        AppendableCharSequence seq = new AppendableCharSequence(initialBufferSize);
+        lineParser = new LineParser(seq, maxInitialLineLength);
+        headerParser = new HeaderParser(seq, maxHeaderSize);
         this.maxChunkSize = maxChunkSize;
         this.chunkedSupported = chunkedSupported;
         this.validateHeaders = validateHeaders;
-        AppendableCharSequence seq = new AppendableCharSequence(128);
-        lineParser = new LineParser(seq, maxInitialLineLength);
-        headerParser = new HeaderParser(seq, maxHeaderSize);
     }
 
     @Override
@@ -193,220 +198,233 @@ public abstract class HttpObjectDecoder extends ByteToMessageDecoder {
         }
 
         switch (currentState) {
-            case SKIP_CONTROL_CHARS: {
-                if (!skipControlCharacters(buffer)) {
-                    return;
-                }
-                currentState = State.READ_INITIAL;
-            }
-            case READ_INITIAL: try {
-                AppendableCharSequence line = lineParser.parse(buffer);
-                if (line == null) {
-                    return;
-                }
-                String[] initialLine = splitInitialLine(line);
-                if (initialLine.length < 3) {
-                    // Invalid initial line - ignore.
-                    currentState = State.SKIP_CONTROL_CHARS;
-                    return;
-                }
-
-                message = createMessage(initialLine);
-                currentState = State.READ_HEADER;
-                // fall-through
-            } catch (Exception e) {
-                out.add(invalidMessage(buffer, e));
+        case SKIP_CONTROL_CHARS: {
+            if (!skipControlCharacters(buffer)) {
                 return;
             }
-            case READ_HEADER: try {
-                State nextState = readHeaders(buffer);
-                if (nextState == null) {
-                    return;
-                }
-                currentState = nextState;
-                switch (nextState) {
-                    case SKIP_CONTROL_CHARS:
-                        // fast-path
-                        // No content is expected.
-                        out.add(message);
-                        out.add(LastHttpContent.EMPTY_LAST_CONTENT);
-                        resetNow();
-                        return;
-                    case READ_CHUNK_SIZE:
-                        if (!chunkedSupported) {
-                            throw new IllegalArgumentException("Chunked messages not supported");
-                        }
-                        // Chunked encoding - generate HttpMessage first.  HttpChunks will follow.
-                        out.add(message);
-                        return;
-                    default:
-                        /**
-                         * <a href="https://tools.ietf.org/html/rfc7230#section-3.3.3">RFC 7230, 3.3.3</a> states that
-                         * if a request does not have either a transfer-encoding or a content-length header then the
-                         * message body length is 0. However for a response the body length is the number of octets
-                         * received prior to the server closing the connection. So we treat this as variable length
-                         * chunked encoding.
-                         */
-                        long contentLength = contentLength();
-                        if (contentLength == 0 || contentLength == -1 && isDecodingRequest()) {
-                            out.add(message);
-                            out.add(LastHttpContent.EMPTY_LAST_CONTENT);
-                            resetNow();
-                            return;
-                        }
-
-                        assert nextState == State.READ_FIXED_LENGTH_CONTENT ||
-                                nextState == State.READ_VARIABLE_LENGTH_CONTENT;
-
-                        out.add(message);
-
-                        if (nextState == State.READ_FIXED_LENGTH_CONTENT) {
-                            // chunkSize will be decreased as the READ_FIXED_LENGTH_CONTENT state reads data chunk by
-                            // chunk.
-                            chunkSize = contentLength;
-                        }
-
-                        // We return here, this forces decode to be called again where we will decode the content
-                        return;
-                }
-            } catch (Exception e) {
-                out.add(invalidMessage(buffer, e));
+            currentState = State.READ_INITIAL;
+        }
+        case READ_INITIAL: try {
+            AppendableCharSequence line = lineParser.parse(buffer);
+            if (line == null) {
                 return;
             }
-            case READ_VARIABLE_LENGTH_CONTENT: {
-                // Keep reading data as a chunk until the end of connection is reached.
-                int toRead = Math.min(buffer.readableBytes(), maxChunkSize);
-                if (toRead > 0) {
-                    ByteBuf content = buffer.readSlice(toRead).retain();
-                    out.add(new DefaultHttpContent(content));
-                }
+            String[] initialLine = splitInitialLine(line);
+            if (initialLine.length < 3) {
+                // Invalid initial line - ignore.
+                currentState = State.SKIP_CONTROL_CHARS;
                 return;
             }
-            case READ_FIXED_LENGTH_CONTENT: {
-                int readLimit = buffer.readableBytes();
 
-                // Check if the buffer is readable first as we use the readable byte count
-                // to create the HttpChunk. This is needed as otherwise we may end up with
-                // create a HttpChunk instance that contains an empty buffer and so is
-                // handled like it is the last HttpChunk.
-                //
-                // See https://github.com/netty/netty/issues/433
-                if (readLimit == 0) {
-                    return;
-                }
-
-                int toRead = Math.min(readLimit, maxChunkSize);
-                if (toRead > chunkSize) {
-                    toRead = (int) chunkSize;
-                }
-                ByteBuf content = buffer.readSlice(toRead).retain();
-                chunkSize -= toRead;
-
-                if (chunkSize == 0) {
-                    // Read all content.
-                    out.add(new DefaultLastHttpContent(content, validateHeaders));
-                    resetNow();
-                } else {
-                    out.add(new DefaultHttpContent(content));
-                }
+            message = createMessage(initialLine);
+            currentState = State.READ_HEADER;
+            // fall-through
+        } catch (Exception e) {
+            out.add(invalidMessage(buffer, e));
+            return;
+        }
+        case READ_HEADER: try {
+            State nextState = readHeaders(buffer);
+            if (nextState == null) {
                 return;
             }
-            /**
-             * everything else after this point takes care of reading chunked content. basically, read chunk size,
-             * read chunk, read and ignore the CRLF and repeat until 0
-             */
-            case READ_CHUNK_SIZE: try {
-                AppendableCharSequence line = lineParser.parse(buffer);
-                if (line == null) {
-                    return;
-                }
-                int chunkSize = getChunkSize(line.toString());
-                this.chunkSize = chunkSize;
-                if (chunkSize == 0) {
-                    currentState = State.READ_CHUNK_FOOTER;
-                    return;
-                }
-                currentState = State.READ_CHUNKED_CONTENT;
-                // fall-through
-            } catch (Exception e) {
-                out.add(invalidChunk(buffer, e));
-                return;
-            }
-            case READ_CHUNKED_CONTENT: {
-                assert chunkSize <= Integer.MAX_VALUE;
-                int toRead = Math.min((int) chunkSize, maxChunkSize);
-                toRead = Math.min(toRead, buffer.readableBytes());
-                if (toRead == 0) {
-                    return;
-                }
-                HttpContent chunk = new DefaultHttpContent(buffer.readSlice(toRead).retain());
-                chunkSize -= toRead;
-
-                out.add(chunk);
-
-                if (chunkSize != 0) {
-                    return;
-                }
-                currentState = State.READ_CHUNK_DELIMITER;
-                // fall-through
-            }
-            case READ_CHUNK_DELIMITER: {
-                final int wIdx = buffer.writerIndex();
-                int rIdx = buffer.readerIndex();
-                while (wIdx > rIdx) {
-                    byte next = buffer.getByte(rIdx++);
-                    if (next == HttpConstants.LF) {
-                        currentState = State.READ_CHUNK_SIZE;
-                        break;
-                    }
-                }
-                buffer.readerIndex(rIdx);
-                return;
-            }
-            case READ_CHUNK_FOOTER: try {
-                LastHttpContent trailer = readTrailingHeaders(buffer);
-                if (trailer == null) {
-                    return;
-                }
-                out.add(trailer);
+            currentState = nextState;
+            switch (nextState) {
+            case SKIP_CONTROL_CHARS:
+                // fast-path
+                // No content is expected.
+                out.add(message);
+                out.add(LastHttpContent.EMPTY_LAST_CONTENT);
                 resetNow();
                 return;
-            } catch (Exception e) {
-                out.add(invalidChunk(buffer, e));
+            case READ_CHUNK_SIZE:
+                if (!chunkedSupported) {
+                    throw new IllegalArgumentException("Chunked messages not supported");
+                }
+                // Chunked encoding - generate HttpMessage first.  HttpChunks will follow.
+                out.add(message);
+                return;
+            default:
+                /**
+                 * <a href="https://tools.ietf.org/html/rfc7230#section-3.3.3">RFC 7230, 3.3.3</a> states that if a
+                 * request does not have either a transfer-encoding or a content-length header then the message body
+                 * length is 0. However for a response the body length is the number of octets received prior to the
+                 * server closing the connection. So we treat this as variable length chunked encoding.
+                 */
+                long contentLength = contentLength();
+                if (contentLength == 0 || contentLength == -1 && isDecodingRequest()) {
+                    out.add(message);
+                    out.add(LastHttpContent.EMPTY_LAST_CONTENT);
+                    resetNow();
+                    return;
+                }
+
+                assert nextState == State.READ_FIXED_LENGTH_CONTENT ||
+                        nextState == State.READ_VARIABLE_LENGTH_CONTENT;
+
+                out.add(message);
+
+                if (nextState == State.READ_FIXED_LENGTH_CONTENT) {
+                    // chunkSize will be decreased as the READ_FIXED_LENGTH_CONTENT state reads data chunk by chunk.
+                    chunkSize = contentLength;
+                }
+
+                // We return here, this forces decode to be called again where we will decode the content
                 return;
             }
-            case BAD_MESSAGE: {
-                // Keep discarding until disconnection.
-                buffer.skipBytes(buffer.readableBytes());
-                break;
+        } catch (Exception e) {
+            out.add(invalidMessage(buffer, e));
+            return;
+        }
+        case READ_VARIABLE_LENGTH_CONTENT: {
+            // Keep reading data as a chunk until the end of connection is reached.
+            int toRead = Math.min(buffer.readableBytes(), maxChunkSize);
+            if (toRead > 0) {
+                ByteBuf content = buffer.readRetainedSlice(toRead);
+                out.add(new DefaultHttpContent(content));
             }
-            case UPGRADED: {
-                int readableBytes = buffer.readableBytes();
-                if (readableBytes > 0) {
-                    // Keep on consuming as otherwise we may trigger an DecoderException,
-                    // other handler will replace this codec with the upgraded protocol codec to
-                    // take the traffic over at some point then.
-                    // See https://github.com/netty/netty/issues/2173
-                    out.add(buffer.readBytes(readableBytes));
+            return;
+        }
+        case READ_FIXED_LENGTH_CONTENT: {
+            int readLimit = buffer.readableBytes();
+
+            // Check if the buffer is readable first as we use the readable byte count
+            // to create the HttpChunk. This is needed as otherwise we may end up with
+            // create a HttpChunk instance that contains an empty buffer and so is
+            // handled like it is the last HttpChunk.
+            //
+            // See https://github.com/netty/netty/issues/433
+            if (readLimit == 0) {
+                return;
+            }
+
+            int toRead = Math.min(readLimit, maxChunkSize);
+            if (toRead > chunkSize) {
+                toRead = (int) chunkSize;
+            }
+            ByteBuf content = buffer.readRetainedSlice(toRead);
+            chunkSize -= toRead;
+
+            if (chunkSize == 0) {
+                // Read all content.
+                out.add(new DefaultLastHttpContent(content, validateHeaders));
+                resetNow();
+            } else {
+                out.add(new DefaultHttpContent(content));
+            }
+            return;
+        }
+        /**
+         * everything else after this point takes care of reading chunked content. basically, read chunk size,
+         * read chunk, read and ignore the CRLF and repeat until 0
+         */
+        case READ_CHUNK_SIZE: try {
+            AppendableCharSequence line = lineParser.parse(buffer);
+            if (line == null) {
+                return;
+            }
+            int chunkSize = getChunkSize(line.toString());
+            this.chunkSize = chunkSize;
+            if (chunkSize == 0) {
+                currentState = State.READ_CHUNK_FOOTER;
+                return;
+            }
+            currentState = State.READ_CHUNKED_CONTENT;
+            // fall-through
+        } catch (Exception e) {
+            out.add(invalidChunk(buffer, e));
+            return;
+        }
+        case READ_CHUNKED_CONTENT: {
+            assert chunkSize <= Integer.MAX_VALUE;
+            int toRead = Math.min((int) chunkSize, maxChunkSize);
+            toRead = Math.min(toRead, buffer.readableBytes());
+            if (toRead == 0) {
+                return;
+            }
+            HttpContent chunk = new DefaultHttpContent(buffer.readRetainedSlice(toRead));
+            chunkSize -= toRead;
+
+            out.add(chunk);
+
+            if (chunkSize != 0) {
+                return;
+            }
+            currentState = State.READ_CHUNK_DELIMITER;
+            // fall-through
+        }
+        case READ_CHUNK_DELIMITER: {
+            final int wIdx = buffer.writerIndex();
+            int rIdx = buffer.readerIndex();
+            while (wIdx > rIdx) {
+                byte next = buffer.getByte(rIdx++);
+                if (next == HttpConstants.LF) {
+                    currentState = State.READ_CHUNK_SIZE;
+                    break;
                 }
-                break;
             }
+            buffer.readerIndex(rIdx);
+            return;
+        }
+        case READ_CHUNK_FOOTER: try {
+            LastHttpContent trailer = readTrailingHeaders(buffer);
+            if (trailer == null) {
+                return;
+            }
+            out.add(trailer);
+            resetNow();
+            return;
+        } catch (Exception e) {
+            out.add(invalidChunk(buffer, e));
+            return;
+        }
+        case BAD_MESSAGE: {
+            // Keep discarding until disconnection.
+            buffer.skipBytes(buffer.readableBytes());
+            break;
+        }
+        case UPGRADED: {
+            int readableBytes = buffer.readableBytes();
+            if (readableBytes > 0) {
+                // Keep on consuming as otherwise we may trigger an DecoderException,
+                // other handler will replace this codec with the upgraded protocol codec to
+                // take the traffic over at some point then.
+                // See https://github.com/netty/netty/issues/2173
+                out.add(buffer.readBytes(readableBytes));
+            }
+            break;
+        }
         }
     }
 
     @Override
     protected void decodeLast(ChannelHandlerContext ctx, ByteBuf in, List<Object> out) throws Exception {
-        decode(ctx, in, out);
+        super.decodeLast(ctx, in, out);
 
+        if (resetRequested) {
+            // If a reset was requested by decodeLast() we need to do it now otherwise we may produce a
+            // LastHttpContent while there was already one.
+            resetNow();
+        }
         // Handle the last unfinished message.
         if (message != null) {
-            boolean chunked = HttpHeaders.isTransferEncodingChunked(message);
+            boolean chunked = HttpUtil.isTransferEncodingChunked(message);
             if (currentState == State.READ_VARIABLE_LENGTH_CONTENT && !in.isReadable() && !chunked) {
                 // End of connection.
                 out.add(LastHttpContent.EMPTY_LAST_CONTENT);
-                reset();
+                resetNow();
                 return;
             }
+
+            if (currentState == State.READ_HEADER) {
+                // If we are still in the state of reading headers we need to create a new invalid message that
+                // signals that the connection was closed before we received the headers.
+                out.add(invalidMessage(Unpooled.EMPTY_BUFFER,
+                        new PrematureChannelClosureException("Connection closed before received headers")));
+                resetNow();
+                return;
+            }
+
             // Check if the closure of the connection signifies the end of the content.
             boolean prematureClosure;
             if (isDecodingRequest() || chunked) {
@@ -418,22 +436,17 @@ public abstract class HttpObjectDecoder extends ByteToMessageDecoder {
                 // connection, so it is perfectly fine.
                 prematureClosure = contentLength() > 0;
             }
-            resetNow();
 
             if (!prematureClosure) {
                 out.add(LastHttpContent.EMPTY_LAST_CONTENT);
             }
+            resetNow();
         }
     }
 
     @Override
     public void userEventTriggered(ChannelHandlerContext ctx, Object evt) throws Exception {
-        if (evt instanceof ChannelInputShutdownEvent) {
-            // The decodeLast method is invoked when a channelInactive event is encountered.
-            // This method is responsible for ending requests in some situations and must be called
-            // when the input has been shutdown.
-            super.channelInactive(ctx);
-        } else if (evt instanceof HttpExpectationFailedEvent) {
+        if (evt instanceof HttpExpectationFailedEvent) {
             switch (currentState) {
             case READ_FIXED_LENGTH_CONTENT:
             case READ_VARIABLE_LENGTH_CONTENT:
@@ -450,7 +463,7 @@ public abstract class HttpObjectDecoder extends ByteToMessageDecoder {
     protected boolean isContentAlwaysEmpty(HttpMessage msg) {
         if (msg instanceof HttpResponse) {
             HttpResponse res = (HttpResponse) msg;
-            int code = res.getStatus().code();
+            int code = res.status().code();
 
             // Correctly handle return codes of 1xx.
             //
@@ -459,13 +472,13 @@ public abstract class HttpObjectDecoder extends ByteToMessageDecoder {
             //     - https://github.com/netty/netty/issues/222
             if (code >= 100 && code < 200) {
                 // One exception: Hixie 76 websocket handshake response
-                return !(code == 101 && !res.headers().contains(HttpHeaders.Names.SEC_WEBSOCKET_ACCEPT)
-                         && res.headers().contains(HttpHeaders.Names.UPGRADE, HttpHeaders.Values.WEBSOCKET, true));
+                return !(code == 101 && !res.headers().contains(HttpHeaderNames.SEC_WEBSOCKET_ACCEPT)
+                         && res.headers().contains(HttpHeaderNames.UPGRADE, HttpHeaderValues.WEBSOCKET, true));
             }
 
             switch (code) {
-                case 204: case 205: case 304:
-                    return true;
+            case 204: case 205: case 304:
+                return true;
             }
         }
         return false;
@@ -490,12 +503,13 @@ public abstract class HttpObjectDecoder extends ByteToMessageDecoder {
         trailer = null;
         if (!isDecodingRequest()) {
             HttpResponse res = (HttpResponse) message;
-            if (res != null && res.getStatus().code() == 101) {
+            if (res != null && res.status().code() == 101) {
                 currentState = State.UPGRADED;
                 return;
             }
         }
 
+        resetRequested = false;
         currentState = State.SKIP_CONTROL_CHARS;
     }
 
@@ -590,9 +604,9 @@ public abstract class HttpObjectDecoder extends ByteToMessageDecoder {
         State nextState;
 
         if (isContentAlwaysEmpty(message)) {
-            HttpHeaders.removeTransferEncodingChunked(message);
+            HttpUtil.setTransferEncodingChunked(message, false);
             nextState = State.SKIP_CONTROL_CHARS;
-        } else if (HttpHeaders.isTransferEncodingChunked(message)) {
+        } else if (HttpUtil.isTransferEncodingChunked(message)) {
             nextState = State.READ_CHUNK_SIZE;
         } else if (contentLength() >= 0) {
             nextState = State.READ_FIXED_LENGTH_CONTENT;
@@ -604,7 +618,7 @@ public abstract class HttpObjectDecoder extends ByteToMessageDecoder {
 
     private long contentLength() {
         if (contentLength == Long.MIN_VALUE) {
-            contentLength = HttpHeaders.getContentLength(message, -1);
+            contentLength = HttpUtil.getContentLength(message, -1L);
         }
         return contentLength;
     }
@@ -638,9 +652,9 @@ public abstract class HttpObjectDecoder extends ByteToMessageDecoder {
                 } else {
                     splitHeader(line);
                     CharSequence headerName = name;
-                    if (!HttpHeaders.equalsIgnoreCase(HttpHeaders.Names.CONTENT_LENGTH, headerName) &&
-                            !HttpHeaders.equalsIgnoreCase(HttpHeaders.Names.TRANSFER_ENCODING, headerName) &&
-                            !HttpHeaders.equalsIgnoreCase(HttpHeaders.Names.TRAILER, headerName)) {
+                    if (!HttpHeaderNames.CONTENT_LENGTH.contentEqualsIgnoreCase(headerName) &&
+                        !HttpHeaderNames.TRANSFER_ENCODING.contentEqualsIgnoreCase(headerName) &&
+                        !HttpHeaderNames.TRAILER.contentEqualsIgnoreCase(headerName)) {
                         trailer.trailingHeaders().add(headerName, value);
                     }
                     lastHeader = name;
@@ -762,7 +776,7 @@ public abstract class HttpObjectDecoder extends ByteToMessageDecoder {
         return 0;
     }
 
-    private static class HeaderParser implements ByteBufProcessor {
+    private static class HeaderParser implements ByteProcessor {
         private final AppendableCharSequence seq;
         private final int maxLength;
         private int size;
