@@ -15,10 +15,11 @@
  */
 package io.netty.handler.codec.http;
 
-import io.netty.buffer.ByteBuf;
-import io.netty.util.CharsetUtil;
-
 import static io.netty.handler.codec.http.HttpConstants.SP;
+import io.netty.buffer.ByteBuf;
+import io.netty.util.AsciiString;
+import io.netty.util.ByteProcessor;
+import io.netty.util.CharsetUtil;
 
 /**
  * The response code and its description of HTTP or its derived protocols, such as
@@ -30,292 +31,290 @@ public class HttpResponseStatus implements Comparable<HttpResponseStatus> {
     /**
      * 100 Continue
      */
-    public static final HttpResponseStatus CONTINUE = new HttpResponseStatus(100, "Continue", true);
+    public static final HttpResponseStatus CONTINUE = newStatus(100, "Continue");
 
     /**
      * 101 Switching Protocols
      */
-    public static final HttpResponseStatus SWITCHING_PROTOCOLS =
-            new HttpResponseStatus(101, "Switching Protocols", true);
+    public static final HttpResponseStatus SWITCHING_PROTOCOLS = newStatus(101, "Switching Protocols");
 
     /**
      * 102 Processing (WebDAV, RFC2518)
      */
-    public static final HttpResponseStatus PROCESSING = new HttpResponseStatus(102, "Processing", true);
+    public static final HttpResponseStatus PROCESSING = newStatus(102, "Processing");
 
     /**
      * 200 OK
      */
-    public static final HttpResponseStatus OK = new HttpResponseStatus(200, "OK", true);
+    public static final HttpResponseStatus OK = newStatus(200, "OK");
 
     /**
      * 201 Created
      */
-    public static final HttpResponseStatus CREATED = new HttpResponseStatus(201, "Created", true);
+    public static final HttpResponseStatus CREATED = newStatus(201, "Created");
 
     /**
      * 202 Accepted
      */
-    public static final HttpResponseStatus ACCEPTED = new HttpResponseStatus(202, "Accepted", true);
+    public static final HttpResponseStatus ACCEPTED = newStatus(202, "Accepted");
 
     /**
      * 203 Non-Authoritative Information (since HTTP/1.1)
      */
     public static final HttpResponseStatus NON_AUTHORITATIVE_INFORMATION =
-            new HttpResponseStatus(203, "Non-Authoritative Information", true);
+            newStatus(203, "Non-Authoritative Information");
 
     /**
      * 204 No Content
      */
-    public static final HttpResponseStatus NO_CONTENT = new HttpResponseStatus(204, "No Content", true);
+    public static final HttpResponseStatus NO_CONTENT = newStatus(204, "No Content");
 
     /**
      * 205 Reset Content
      */
-    public static final HttpResponseStatus RESET_CONTENT = new HttpResponseStatus(205, "Reset Content", true);
+    public static final HttpResponseStatus RESET_CONTENT = newStatus(205, "Reset Content");
 
     /**
      * 206 Partial Content
      */
-    public static final HttpResponseStatus PARTIAL_CONTENT = new HttpResponseStatus(206, "Partial Content", true);
+    public static final HttpResponseStatus PARTIAL_CONTENT = newStatus(206, "Partial Content");
 
     /**
      * 207 Multi-Status (WebDAV, RFC2518)
      */
-    public static final HttpResponseStatus MULTI_STATUS = new HttpResponseStatus(207, "Multi-Status", true);
+    public static final HttpResponseStatus MULTI_STATUS = newStatus(207, "Multi-Status");
 
     /**
      * 300 Multiple Choices
      */
-    public static final HttpResponseStatus MULTIPLE_CHOICES = new HttpResponseStatus(300, "Multiple Choices", true);
+    public static final HttpResponseStatus MULTIPLE_CHOICES = newStatus(300, "Multiple Choices");
 
     /**
      * 301 Moved Permanently
      */
-    public static final HttpResponseStatus MOVED_PERMANENTLY = new HttpResponseStatus(301, "Moved Permanently", true);
+    public static final HttpResponseStatus MOVED_PERMANENTLY = newStatus(301, "Moved Permanently");
 
     /**
      * 302 Found
      */
-    public static final HttpResponseStatus FOUND = new HttpResponseStatus(302, "Found", true);
+    public static final HttpResponseStatus FOUND = newStatus(302, "Found");
 
     /**
      * 303 See Other (since HTTP/1.1)
      */
-    public static final HttpResponseStatus SEE_OTHER = new HttpResponseStatus(303, "See Other", true);
+    public static final HttpResponseStatus SEE_OTHER = newStatus(303, "See Other");
 
     /**
      * 304 Not Modified
      */
-    public static final HttpResponseStatus NOT_MODIFIED = new HttpResponseStatus(304, "Not Modified", true);
+    public static final HttpResponseStatus NOT_MODIFIED = newStatus(304, "Not Modified");
 
     /**
      * 305 Use Proxy (since HTTP/1.1)
      */
-    public static final HttpResponseStatus USE_PROXY = new HttpResponseStatus(305, "Use Proxy", true);
+    public static final HttpResponseStatus USE_PROXY = newStatus(305, "Use Proxy");
 
     /**
      * 307 Temporary Redirect (since HTTP/1.1)
      */
-    public static final HttpResponseStatus TEMPORARY_REDIRECT = new HttpResponseStatus(307, "Temporary Redirect", true);
+    public static final HttpResponseStatus TEMPORARY_REDIRECT = newStatus(307, "Temporary Redirect");
 
     /**
      * 400 Bad Request
      */
-    public static final HttpResponseStatus BAD_REQUEST = new HttpResponseStatus(400, "Bad Request", true);
+    public static final HttpResponseStatus BAD_REQUEST = newStatus(400, "Bad Request");
 
     /**
      * 401 Unauthorized
      */
-    public static final HttpResponseStatus UNAUTHORIZED = new HttpResponseStatus(401, "Unauthorized", true);
+    public static final HttpResponseStatus UNAUTHORIZED = newStatus(401, "Unauthorized");
 
     /**
      * 402 Payment Required
      */
-    public static final HttpResponseStatus PAYMENT_REQUIRED = new HttpResponseStatus(402, "Payment Required", true);
+    public static final HttpResponseStatus PAYMENT_REQUIRED = newStatus(402, "Payment Required");
 
     /**
      * 403 Forbidden
      */
-    public static final HttpResponseStatus FORBIDDEN = new HttpResponseStatus(403, "Forbidden", true);
+    public static final HttpResponseStatus FORBIDDEN = newStatus(403, "Forbidden");
 
     /**
      * 404 Not Found
      */
-    public static final HttpResponseStatus NOT_FOUND = new HttpResponseStatus(404, "Not Found", true);
+    public static final HttpResponseStatus NOT_FOUND = newStatus(404, "Not Found");
 
     /**
      * 405 Method Not Allowed
      */
-    public static final HttpResponseStatus METHOD_NOT_ALLOWED = new HttpResponseStatus(405, "Method Not Allowed", true);
+    public static final HttpResponseStatus METHOD_NOT_ALLOWED = newStatus(405, "Method Not Allowed");
 
     /**
      * 406 Not Acceptable
      */
-    public static final HttpResponseStatus NOT_ACCEPTABLE = new HttpResponseStatus(406, "Not Acceptable", true);
+    public static final HttpResponseStatus NOT_ACCEPTABLE = newStatus(406, "Not Acceptable");
 
     /**
      * 407 Proxy Authentication Required
      */
     public static final HttpResponseStatus PROXY_AUTHENTICATION_REQUIRED =
-            new HttpResponseStatus(407, "Proxy Authentication Required", true);
+            newStatus(407, "Proxy Authentication Required");
 
     /**
      * 408 Request Timeout
      */
-    public static final HttpResponseStatus REQUEST_TIMEOUT = new HttpResponseStatus(408, "Request Timeout", true);
+    public static final HttpResponseStatus REQUEST_TIMEOUT = newStatus(408, "Request Timeout");
 
     /**
      * 409 Conflict
      */
-    public static final HttpResponseStatus CONFLICT = new HttpResponseStatus(409, "Conflict", true);
+    public static final HttpResponseStatus CONFLICT = newStatus(409, "Conflict");
 
     /**
      * 410 Gone
      */
-    public static final HttpResponseStatus GONE = new HttpResponseStatus(410, "Gone", true);
+    public static final HttpResponseStatus GONE = newStatus(410, "Gone");
 
     /**
      * 411 Length Required
      */
-    public static final HttpResponseStatus LENGTH_REQUIRED = new HttpResponseStatus(411, "Length Required", true);
+    public static final HttpResponseStatus LENGTH_REQUIRED = newStatus(411, "Length Required");
 
     /**
      * 412 Precondition Failed
      */
-    public static final HttpResponseStatus PRECONDITION_FAILED =
-            new HttpResponseStatus(412, "Precondition Failed", true);
+    public static final HttpResponseStatus PRECONDITION_FAILED = newStatus(412, "Precondition Failed");
 
     /**
      * 413 Request Entity Too Large
      */
     public static final HttpResponseStatus REQUEST_ENTITY_TOO_LARGE =
-            new HttpResponseStatus(413, "Request Entity Too Large", true);
+            newStatus(413, "Request Entity Too Large");
 
     /**
      * 414 Request-URI Too Long
      */
-    public static final HttpResponseStatus REQUEST_URI_TOO_LONG =
-            new HttpResponseStatus(414, "Request-URI Too Long", true);
+    public static final HttpResponseStatus REQUEST_URI_TOO_LONG = newStatus(414, "Request-URI Too Long");
 
     /**
      * 415 Unsupported Media Type
      */
-    public static final HttpResponseStatus UNSUPPORTED_MEDIA_TYPE =
-            new HttpResponseStatus(415, "Unsupported Media Type", true);
+    public static final HttpResponseStatus UNSUPPORTED_MEDIA_TYPE = newStatus(415, "Unsupported Media Type");
 
     /**
      * 416 Requested Range Not Satisfiable
      */
     public static final HttpResponseStatus REQUESTED_RANGE_NOT_SATISFIABLE =
-            new HttpResponseStatus(416, "Requested Range Not Satisfiable", true);
+            newStatus(416, "Requested Range Not Satisfiable");
 
     /**
      * 417 Expectation Failed
      */
-    public static final HttpResponseStatus EXPECTATION_FAILED =
-            new HttpResponseStatus(417, "Expectation Failed", true);
+    public static final HttpResponseStatus EXPECTATION_FAILED = newStatus(417, "Expectation Failed");
+
+    /**
+     * 421 Misdirected Request
+     *
+     * <a href="https://tools.ietf.org/html/draft-ietf-httpbis-http2-15#section-9.1.2">421 Status Code</a>
+     */
+    public static final HttpResponseStatus MISDIRECTED_REQUEST = newStatus(421, "Misdirected Request");
 
     /**
      * 422 Unprocessable Entity (WebDAV, RFC4918)
      */
-    public static final HttpResponseStatus UNPROCESSABLE_ENTITY =
-            new HttpResponseStatus(422, "Unprocessable Entity", true);
+    public static final HttpResponseStatus UNPROCESSABLE_ENTITY = newStatus(422, "Unprocessable Entity");
 
     /**
      * 423 Locked (WebDAV, RFC4918)
      */
-    public static final HttpResponseStatus LOCKED =
-            new HttpResponseStatus(423, "Locked", true);
+    public static final HttpResponseStatus LOCKED = newStatus(423, "Locked");
 
     /**
      * 424 Failed Dependency (WebDAV, RFC4918)
      */
-    public static final HttpResponseStatus FAILED_DEPENDENCY = new HttpResponseStatus(424, "Failed Dependency", true);
+    public static final HttpResponseStatus FAILED_DEPENDENCY = newStatus(424, "Failed Dependency");
 
     /**
      * 425 Unordered Collection (WebDAV, RFC3648)
      */
-    public static final HttpResponseStatus UNORDERED_COLLECTION =
-            new HttpResponseStatus(425, "Unordered Collection", true);
+    public static final HttpResponseStatus UNORDERED_COLLECTION = newStatus(425, "Unordered Collection");
 
     /**
      * 426 Upgrade Required (RFC2817)
      */
-    public static final HttpResponseStatus UPGRADE_REQUIRED = new HttpResponseStatus(426, "Upgrade Required", true);
+    public static final HttpResponseStatus UPGRADE_REQUIRED = newStatus(426, "Upgrade Required");
 
     /**
      * 428 Precondition Required (RFC6585)
      */
-    public static final HttpResponseStatus PRECONDITION_REQUIRED =
-            new HttpResponseStatus(428, "Precondition Required", true);
+    public static final HttpResponseStatus PRECONDITION_REQUIRED = newStatus(428, "Precondition Required");
 
     /**
      * 429 Too Many Requests (RFC6585)
      */
-    public static final HttpResponseStatus TOO_MANY_REQUESTS = new HttpResponseStatus(429, "Too Many Requests", true);
+    public static final HttpResponseStatus TOO_MANY_REQUESTS = newStatus(429, "Too Many Requests");
 
     /**
      * 431 Request Header Fields Too Large (RFC6585)
      */
     public static final HttpResponseStatus REQUEST_HEADER_FIELDS_TOO_LARGE =
-            new HttpResponseStatus(431, "Request Header Fields Too Large", true);
+            newStatus(431, "Request Header Fields Too Large");
 
     /**
      * 500 Internal Server Error
      */
-    public static final HttpResponseStatus INTERNAL_SERVER_ERROR =
-            new HttpResponseStatus(500, "Internal Server Error", true);
+    public static final HttpResponseStatus INTERNAL_SERVER_ERROR = newStatus(500, "Internal Server Error");
 
     /**
      * 501 Not Implemented
      */
-    public static final HttpResponseStatus NOT_IMPLEMENTED = new HttpResponseStatus(501, "Not Implemented", true);
+    public static final HttpResponseStatus NOT_IMPLEMENTED = newStatus(501, "Not Implemented");
 
     /**
      * 502 Bad Gateway
      */
-    public static final HttpResponseStatus BAD_GATEWAY = new HttpResponseStatus(502, "Bad Gateway", true);
+    public static final HttpResponseStatus BAD_GATEWAY = newStatus(502, "Bad Gateway");
 
     /**
      * 503 Service Unavailable
      */
-    public static final HttpResponseStatus SERVICE_UNAVAILABLE =
-            new HttpResponseStatus(503, "Service Unavailable", true);
+    public static final HttpResponseStatus SERVICE_UNAVAILABLE = newStatus(503, "Service Unavailable");
 
     /**
      * 504 Gateway Timeout
      */
-    public static final HttpResponseStatus GATEWAY_TIMEOUT = new HttpResponseStatus(504, "Gateway Timeout", true);
+    public static final HttpResponseStatus GATEWAY_TIMEOUT = newStatus(504, "Gateway Timeout");
 
     /**
      * 505 HTTP Version Not Supported
      */
     public static final HttpResponseStatus HTTP_VERSION_NOT_SUPPORTED =
-            new HttpResponseStatus(505, "HTTP Version Not Supported", true);
+            newStatus(505, "HTTP Version Not Supported");
 
     /**
      * 506 Variant Also Negotiates (RFC2295)
      */
-    public static final HttpResponseStatus VARIANT_ALSO_NEGOTIATES =
-            new HttpResponseStatus(506, "Variant Also Negotiates", true);
+    public static final HttpResponseStatus VARIANT_ALSO_NEGOTIATES = newStatus(506, "Variant Also Negotiates");
 
     /**
      * 507 Insufficient Storage (WebDAV, RFC4918)
      */
-    public static final HttpResponseStatus INSUFFICIENT_STORAGE =
-            new HttpResponseStatus(507, "Insufficient Storage", true);
+    public static final HttpResponseStatus INSUFFICIENT_STORAGE = newStatus(507, "Insufficient Storage");
 
     /**
      * 510 Not Extended (RFC2774)
      */
-    public static final HttpResponseStatus NOT_EXTENDED = new HttpResponseStatus(510, "Not Extended", true);
+    public static final HttpResponseStatus NOT_EXTENDED = newStatus(510, "Not Extended");
 
     /**
      * 511 Network Authentication Required (RFC6585)
      */
     public static final HttpResponseStatus NETWORK_AUTHENTICATION_REQUIRED =
-            new HttpResponseStatus(511, "Network Authentication Required", true);
+            newStatus(511, "Network Authentication Required");
+
+    private static HttpResponseStatus newStatus(int statusCode, String reasonPhrase) {
+        return new HttpResponseStatus(statusCode, reasonPhrase, true);
+    }
 
     /**
      * Returns the {@link HttpResponseStatus} represented by the specified code.
@@ -396,6 +395,8 @@ public class HttpResponseStatus implements Comparable<HttpResponseStatus> {
             return REQUESTED_RANGE_NOT_SATISFIABLE;
         case 417:
             return EXPECTATION_FAILED;
+        case 421:
+            return MISDIRECTED_REQUEST;
         case 422:
             return UNPROCESSABLE_ENTITY;
         case 423:
@@ -434,35 +435,134 @@ public class HttpResponseStatus implements Comparable<HttpResponseStatus> {
             return NETWORK_AUTHENTICATION_REQUIRED;
         }
 
-        final String reasonPhrase;
+        return new HttpResponseStatus(code);
+    }
 
-        if (code < 100) {
-            reasonPhrase = "Unknown Status";
-        } else if (code < 200) {
-            reasonPhrase = "Informational";
-        } else if (code < 300) {
-            reasonPhrase = "Successful";
-        } else if (code < 400) {
-            reasonPhrase = "Redirection";
-        } else if (code < 500) {
-            reasonPhrase = "Client Error";
-        } else if (code < 600) {
-            reasonPhrase = "Server Error";
-        } else {
-            reasonPhrase = "Unknown Status";
+    /**
+     * Parses the specified HTTP status line into a {@link HttpResponseStatus}.  The expected formats of the line are:
+     * <ul>
+     * <li>{@code statusCode} (e.g. 200)</li>
+     * <li>{@code statusCode} {@code reasonPhrase} (e.g. 404 Not Found)</li>
+     * </ul>
+     *
+     * @throws IllegalArgumentException if the specified status line is malformed
+     */
+    public static HttpResponseStatus parseLine(CharSequence line) {
+        String status = line.toString();
+        try {
+            int space = status.indexOf(' ');
+            if (space == -1) {
+                return valueOf(Integer.parseInt(status));
+            } else {
+                int code = Integer.parseInt(status.substring(0, space));
+                String reasonPhrase = status.substring(space + 1);
+                HttpResponseStatus responseStatus = valueOf(code);
+                if (responseStatus.reasonPhrase().contentEquals(reasonPhrase)) {
+                    return responseStatus;
+                } else {
+                    return new HttpResponseStatus(code, reasonPhrase);
+                }
+            }
+        } catch (Exception e) {
+            throw new IllegalArgumentException("malformed status line: " + status, e);
+        }
+    }
+
+    private static final class HttpStatusLineProcessor implements ByteProcessor {
+        private static final byte ASCII_SPACE = (byte) ' ';
+        private final AsciiString string;
+        private int i;
+        /**
+         * 0 = New or havn't seen {@link ASCII_SPACE}.
+         * 1 = Last byte was {@link ASCII_SPACE}.
+         * 2 = Terminal State. Processed the byte after {@link ASCII_SPACE}, and parsed the status line.
+         * 3 = Terminal State. There was no byte after {@link ASCII_SPACE} but status has been parsed with what we saw.
+         */
+        private int state;
+        private HttpResponseStatus status;
+
+        public HttpStatusLineProcessor(AsciiString string) {
+            this.string = string;
         }
 
-        return new HttpResponseStatus(code, reasonPhrase + " (" + code + ')');
+        @Override
+        public boolean process(byte value) {
+            switch (state) {
+            case 0:
+                if (value == ASCII_SPACE) {
+                    state = 1;
+                }
+                break;
+            case 1:
+                parseStatus(i);
+                state = 2;
+                return false;
+            default:
+                break;
+            }
+            ++i;
+            return true;
+        }
+
+        private void parseStatus(int codeEnd) {
+            int code = string.parseInt(0, codeEnd);
+            status = valueOf(code);
+            if (codeEnd < string.length()) {
+                String actualReason = string.toString(codeEnd + 1, string.length());
+                if (!status.reasonPhrase().contentEquals(actualReason)) {
+                    status = new HttpResponseStatus(code, actualReason);
+                }
+            }
+        }
+
+        public HttpResponseStatus status() {
+            if (state <= 1) {
+                parseStatus(string.length());
+                state = 3;
+            }
+            return status;
+        }
+    }
+
+    /**
+     * Parses the specified HTTP status line into a {@link HttpResponseStatus}. The expected formats of the line are:
+     * <ul>
+     * <li>{@code statusCode} (e.g. 200)</li>
+     * <li>{@code statusCode} {@code reasonPhrase} (e.g. 404 Not Found)</li>
+     * </ul>
+     *
+     * @throws IllegalArgumentException if the specified status line is malformed
+     */
+    public static HttpResponseStatus parseLine(AsciiString line) {
+        try {
+            HttpStatusLineProcessor processor = new HttpStatusLineProcessor(line);
+            line.forEachByte(processor);
+            HttpResponseStatus status = processor.status();
+            if (status == null) {
+                throw new IllegalArgumentException("unable to get status after parsing input");
+            }
+            return status;
+        } catch (Exception e) {
+            throw new IllegalArgumentException("malformed status line: " + line, e);
+        }
     }
 
     private final int code;
+    private final AsciiString codeAsText;
+    private HttpStatusClass codeClass;
 
     private final String reasonPhrase;
     private final byte[] bytes;
 
     /**
-     * Creates a new instance with the specified {@code code} and its
-     * {@code reasonPhrase}.
+     * Creates a new instance with the specified {@code code} and the auto-generated default reason phrase.
+     */
+    private HttpResponseStatus(int code) {
+        this(code, HttpStatusClass.valueOf(code).defaultReasonPhrase() + " (" + code + ')', false);
+    }
+
+    /**
+     * Creates a new instance with the specified {@code code} and its {@code reasonPhrase}.
      */
     public HttpResponseStatus(int code, String reasonPhrase) {
         this(code, reasonPhrase, false);
@@ -490,6 +590,7 @@ public class HttpResponseStatus implements Comparable<HttpResponseStatus> {
         }
 
         this.code = code;
+        codeAsText = new AsciiString(Integer.toString(code));
         this.reasonPhrase = reasonPhrase;
         if (bytes) {
             this.bytes = (code + " " + reasonPhrase).getBytes(CharsetUtil.US_ASCII);
@@ -505,6 +606,13 @@ public class HttpResponseStatus implements Comparable<HttpResponseStatus> {
         return code;
     }
 
+    /**
+     * Returns the status code as {@link AsciiString}.
+     */
+    public AsciiString codeAsText() {
+        return codeAsText;
+    }
+
     /**
      * Returns the reason phrase of this {@link HttpResponseStatus}.
      */
@@ -512,6 +620,17 @@ public class HttpResponseStatus implements Comparable<HttpResponseStatus> {
         return reasonPhrase;
     }
 
+    /**
+     * Returns the class of this {@link HttpResponseStatus}
+     */
+    public HttpStatusClass codeClass() {
+        HttpStatusClass type = this.codeClass;
+        if (type == null) {
+            this.codeClass = type = HttpStatusClass.valueOf(code);
+        }
+        return type;
+    }
+
     @Override
     public int hashCode() {
         return code();
@@ -550,9 +669,9 @@ public class HttpResponseStatus implements Comparable<HttpResponseStatus> {
 
     void encode(ByteBuf buf) {
         if (bytes == null) {
-            HttpHeaders.encodeAscii0(String.valueOf(code()), buf);
+            HttpUtil.encodeAscii0(String.valueOf(code()), buf);
             buf.writeByte(SP);
-            HttpHeaders.encodeAscii0(String.valueOf(reasonPhrase()), buf);
+            HttpUtil.encodeAscii0(String.valueOf(reasonPhrase()), buf);
         } else {
             buf.writeBytes(bytes);
         }
