@@ -18,6 +18,8 @@ package io.netty.handler.codec.http;
 
 import io.netty.buffer.ByteBuf;
 import io.netty.buffer.Unpooled;
+import io.netty.channel.ChannelHandlerContext;
+import io.netty.channel.ChannelInboundHandlerAdapter;
 import io.netty.channel.embedded.EmbeddedChannel;
 import io.netty.handler.codec.compression.ZlibCodecFactory;
 import io.netty.handler.codec.compression.ZlibDecoder;
@@ -30,9 +32,16 @@ import org.junit.Test;
 import java.util.ArrayList;
 import java.util.List;
 import java.util.Queue;
+import java.util.concurrent.atomic.AtomicReference;
 
-import static org.hamcrest.CoreMatchers.*;
-import static org.junit.Assert.*;
+import static org.hamcrest.CoreMatchers.instanceOf;
+import static org.hamcrest.CoreMatchers.is;
+import static org.junit.Assert.assertEquals;
+import static org.junit.Assert.assertFalse;
+import static org.junit.Assert.assertNotNull;
+import static org.junit.Assert.assertNull;
+import static org.junit.Assert.assertThat;
+import static org.junit.Assert.assertTrue;
 
 public class HttpContentDecoderTest {
     private static final String HELLO_WORLD = "hello, world";
@@ -73,8 +82,7 @@ public class HttpContentDecoderTest {
         Object o = channel.readInbound();
         assertThat(o, is(instanceOf(FullHttpRequest.class)));
         FullHttpRequest req = (FullHttpRequest) o;
-        assertEquals(HELLO_WORLD.length(),
-                Integer.valueOf(req.headers().get(HttpHeaders.Names.CONTENT_LENGTH)).intValue());
+        assertEquals(HELLO_WORLD.length(), req.headers().getInt(HttpHeaderNames.CONTENT_LENGTH).intValue());
         assertEquals(HELLO_WORLD, req.content().toString(CharsetUtil.US_ASCII));
         req.release();
 
@@ -101,8 +109,7 @@ public class HttpContentDecoderTest {
         Object o = channel.readInbound();
         assertThat(o, is(instanceOf(FullHttpResponse.class)));
         FullHttpResponse resp = (FullHttpResponse) o;
-        assertEquals(HELLO_WORLD.length(),
-                Integer.valueOf(resp.headers().get(HttpHeaders.Names.CONTENT_LENGTH)).intValue());
+        assertEquals(HELLO_WORLD.length(), resp.headers().getInt(HttpHeaderNames.CONTENT_LENGTH).intValue());
         assertEquals(HELLO_WORLD, resp.content().toString(CharsetUtil.US_ASCII));
         resp.release();
 
@@ -130,7 +137,7 @@ public class HttpContentDecoderTest {
         Object o = channel.readOutbound();
         assertThat(o, is(instanceOf(FullHttpResponse.class)));
         FullHttpResponse r = (FullHttpResponse) o;
-        assertEquals(100, r.getStatus().code());
+        assertEquals(100, r.status().code());
         assertTrue(channel.writeInbound(Unpooled.wrappedBuffer(GZ_HELLO_WORLD)));
         r.release();
 
@@ -156,7 +163,7 @@ public class HttpContentDecoderTest {
         Object o = channel.readOutbound();
         assertThat(o, is(instanceOf(FullHttpResponse.class)));
         FullHttpResponse r = (FullHttpResponse) o;
-        assertEquals(100, r.getStatus().code());
+        assertEquals(100, r.status().code());
         r.release();
         assertTrue(channel.writeInbound(Unpooled.wrappedBuffer(GZ_HELLO_WORLD)));
 
@@ -183,7 +190,7 @@ public class HttpContentDecoderTest {
         Object o = channel.readOutbound();
         assertThat(o, is(instanceOf(FullHttpResponse.class)));
         FullHttpResponse r = (FullHttpResponse) o;
-        assertEquals(100, r.getStatus().code());
+        assertEquals(100, r.status().code());
         r.release();
         assertTrue(channel.writeInbound(Unpooled.wrappedBuffer(GZ_HELLO_WORLD)));
 
@@ -210,7 +217,7 @@ public class HttpContentDecoderTest {
         Object o = channel.readOutbound();
         assertThat(o, is(instanceOf(FullHttpResponse.class)));
         FullHttpResponse r = (FullHttpResponse) o;
-        assertEquals(100, r.getStatus().code());
+        assertEquals(100, r.status().code());
         r.release();
         assertTrue(channel.writeInbound(Unpooled.wrappedBuffer(GZ_HELLO_WORLD)));
 
@@ -219,6 +226,60 @@ public class HttpContentDecoderTest {
         assertFalse(channel.finish());
     }
 
+    @Test
+    public void testExpectContinueResetHttpObjectDecoder() {
+        // request with header "Expect: 100-continue" must be replied with one "100 Continue" response
+        // case 5: Test that HttpObjectDecoder correctly resets its internal state after a failed expectation.
+        HttpRequestDecoder decoder = new HttpRequestDecoder();
+        final int maxBytes = 10;
+        HttpObjectAggregator aggregator = new HttpObjectAggregator(maxBytes);
+        final AtomicReference<FullHttpRequest> secondRequestRef = new AtomicReference<FullHttpRequest>();
+        EmbeddedChannel channel = new EmbeddedChannel(decoder, aggregator, new ChannelInboundHandlerAdapter() {
+            @Override
+            public void channelRead(ChannelHandlerContext ctx, Object msg) throws Exception {
+                if (msg instanceof FullHttpRequest) {
+                    if (!secondRequestRef.compareAndSet(null, (FullHttpRequest) msg)) {
+                        ((FullHttpRequest) msg).release();
+                    }
+                } else {
+                    ReferenceCountUtil.release(msg);
+                }
+            }
+        });
+        String req1 = "POST /1 HTTP/1.1\r\n" +
+                "Content-Length: " + (maxBytes + 1) + "\r\n" +
+                "Expect: 100-continue\r\n" +
+                "\r\n";
+        assertFalse(channel.writeInbound(Unpooled.wrappedBuffer(req1.getBytes(CharsetUtil.US_ASCII))));
+
+        FullHttpResponse resp = channel.readOutbound();
+        assertEquals(HttpStatusClass.CLIENT_ERROR, resp.status().codeClass());
+        resp.release();
+
+        String req2 = "POST /2 HTTP/1.1\r\n" +
+                "Content-Length: " + maxBytes + "\r\n" +
+                "Expect: 100-continue\r\n" +
+                "\r\n";
+        assertFalse(channel.writeInbound(Unpooled.wrappedBuffer(req2.getBytes(CharsetUtil.US_ASCII))));
+
+        resp = channel.readOutbound();
+        assertEquals(100, resp.status().code());
+        resp.release();
+
+        byte[] content = new byte[maxBytes];
+        assertFalse(channel.writeInbound(Unpooled.wrappedBuffer(content)));
+
+        FullHttpRequest req = secondRequestRef.get();
+        assertNotNull(req);
+        assertEquals("/2", req.uri());
+        assertEquals(10, req.content().readableBytes());
+        req.release();
+
+        assertHasInboundMessages(channel, false);
+        assertHasOutboundMessages(channel, false);
+        assertFalse(channel.finish());
+    }
+
     @Test
     public void testRequestContentLength1() {
         // case 1: test that ContentDecompressor either sets the correct Content-Length header
@@ -240,7 +301,7 @@ public class HttpContentDecoderTest {
         Object o = req.peek();
         assertThat(o, is(instanceOf(HttpRequest.class)));
         HttpRequest r = (HttpRequest) o;
-        String v = r.headers().get(HttpHeaders.Names.CONTENT_LENGTH);
+        String v = r.headers().get(HttpHeaderNames.CONTENT_LENGTH);
         Long value = v == null ? null : Long.parseLong(v);
         assertTrue(value == null || value.longValue() == HELLO_WORLD.length());
 
@@ -268,7 +329,7 @@ public class HttpContentDecoderTest {
         Object o = channel.readInbound();
         assertThat(o, is(instanceOf(FullHttpRequest.class)));
         FullHttpRequest r = (FullHttpRequest) o;
-        String v = r.headers().get(HttpHeaders.Names.CONTENT_LENGTH);
+        String v = r.headers().get(HttpHeaderNames.CONTENT_LENGTH);
         Long value = v == null ? null : Long.parseLong(v);
 
         r.release();
@@ -302,11 +363,11 @@ public class HttpContentDecoderTest {
         assertThat(o, is(instanceOf(HttpResponse.class)));
         HttpResponse r = (HttpResponse) o;
 
-        assertFalse("Content-Length header not removed.", r.headers().contains(HttpHeaders.Names.CONTENT_LENGTH));
+        assertFalse("Content-Length header not removed.", r.headers().contains(HttpHeaderNames.CONTENT_LENGTH));
 
-        String transferEncoding = r.headers().get(HttpHeaders.Names.TRANSFER_ENCODING);
+        String transferEncoding = r.headers().get(HttpHeaderNames.TRANSFER_ENCODING);
         assertNotNull("Content-length as well as transfer-encoding not set.", transferEncoding);
-        assertEquals("Unexpected transfer-encoding value.", HttpHeaders.Values.CHUNKED.toString(), transferEncoding);
+        assertEquals("Unexpected transfer-encoding value.", HttpHeaderValues.CHUNKED.toString(), transferEncoding);
 
         assertHasInboundMessages(channel, true);
         assertHasOutboundMessages(channel, false);
@@ -332,7 +393,7 @@ public class HttpContentDecoderTest {
         Object o = channel.readInbound();
         assertThat(o, is(instanceOf(FullHttpResponse.class)));
         FullHttpResponse r = (FullHttpResponse) o;
-        String v = r.headers().get(HttpHeaders.Names.CONTENT_LENGTH);
+        String v = r.headers().get(HttpHeaderNames.CONTENT_LENGTH);
         Long value = v == null ? null : Long.parseLong(v);
         assertNotNull(value);
         assertEquals(HELLO_WORLD.length(), value.longValue());
@@ -361,7 +422,7 @@ public class HttpContentDecoderTest {
         int contentLength = 0;
         contentLength = calculateContentLength(req, contentLength);
 
-        byte[] receivedContent = readContent(req, contentLength);
+        byte[] receivedContent = readContent(req, contentLength, true);
 
         assertEquals(HELLO_WORLD, new String(receivedContent, CharsetUtil.US_ASCII));
 
@@ -388,7 +449,35 @@ public class HttpContentDecoderTest {
         int contentLength = 0;
         contentLength = calculateContentLength(resp, contentLength);
 
-        byte[] receivedContent = readContent(resp, contentLength);
+        byte[] receivedContent = readContent(resp, contentLength, true);
+
+        assertEquals(HELLO_WORLD, new String(receivedContent, CharsetUtil.US_ASCII));
+
+        assertHasInboundMessages(channel, true);
+        assertHasOutboundMessages(channel, false);
+        assertFalse(channel.finish());
+    }
+
+    // See https://github.com/netty/netty/issues/5892
+    @Test
+    public void testFullHttpResponseEOF() {
+        // test that ContentDecoder can be used after the ObjectAggregator
+        HttpResponseDecoder decoder = new HttpResponseDecoder(4096, 4096, 5);
+        HttpContentDecoder decompressor = new HttpContentDecompressor();
+        EmbeddedChannel channel = new EmbeddedChannel(decoder, decompressor);
+        String headers = "HTTP/1.1 200 OK\r\n" +
+                "Content-Encoding: gzip\r\n" +
+                "\r\n";
+        assertTrue(channel.writeInbound(Unpooled.copiedBuffer(headers.getBytes(), GZ_HELLO_WORLD)));
+        // This should terminate it.
+        assertTrue(channel.finish());
+
+        Queue<Object> resp = channel.inboundMessages();
+        assertTrue(resp.size() > 1);
+        int contentLength = 0;
+        contentLength = calculateContentLength(resp, contentLength);
+
+        byte[] receivedContent = readContent(resp, contentLength, false);
 
         assertEquals(HELLO_WORLD, new String(receivedContent, CharsetUtil.US_ASCII));
 
@@ -406,7 +495,7 @@ public class HttpContentDecoderTest {
         int outputSize = 0;
         ByteBuf o;
         List<ByteBuf> inbound = new ArrayList<ByteBuf>();
-        while ((o = (ByteBuf) channel.readInbound()) != null) {
+        while ((o = channel.readInbound()) != null) {
             inbound.add(o);
             outputSize += o.readableBytes();
         }
@@ -423,7 +512,7 @@ public class HttpContentDecoderTest {
         return output;
     }
 
-    private static byte[] readContent(Queue<Object> req, int contentLength) {
+    private static byte[] readContent(Queue<Object> req, int contentLength, boolean hasTransferEncoding) {
         byte[] receivedContent = new byte[contentLength];
         int readCount = 0;
         for (Object o : req) {
@@ -433,6 +522,10 @@ public class HttpContentDecoderTest {
                 b.readBytes(receivedContent, readCount, readableBytes);
                 readCount += readableBytes;
             }
+            if (o instanceof HttpMessage) {
+                assertEquals(hasTransferEncoding,
+                        ((HttpMessage) o).headers().contains(HttpHeaderNames.TRANSFER_ENCODING));
+            }
         }
         return receivedContent;
     }
@@ -457,7 +550,7 @@ public class HttpContentDecoderTest {
         int outputSize = 0;
         ByteBuf o;
         List<ByteBuf> outbound = new ArrayList<ByteBuf>();
-        while ((o = (ByteBuf) channel.readOutbound()) != null) {
+        while ((o = channel.readOutbound()) != null) {
             outbound.add(o);
             outputSize += o.readableBytes();
         }
