@@ -20,16 +20,25 @@ import io.netty.buffer.CompositeByteBuf;
 import io.netty.buffer.Unpooled;
 import io.netty.channel.ChannelHandlerContext;
 import io.netty.channel.embedded.EmbeddedChannel;
+import io.netty.handler.codec.DecoderResultProvider;
 import io.netty.handler.codec.TooLongFrameException;
 import io.netty.util.CharsetUtil;
 import org.easymock.EasyMock;
 import org.junit.Test;
 
+import java.nio.channels.ClosedChannelException;
 import java.util.List;
 
-import static io.netty.util.ReferenceCountUtil.*;
-import static org.hamcrest.CoreMatchers.*;
-import static org.junit.Assert.*;
+import static io.netty.handler.codec.http.HttpHeadersTestUtils.of;
+import static org.hamcrest.CoreMatchers.instanceOf;
+import static org.hamcrest.CoreMatchers.is;
+import static org.junit.Assert.assertEquals;
+import static org.junit.Assert.assertFalse;
+import static org.junit.Assert.assertNotNull;
+import static org.junit.Assert.assertNull;
+import static org.junit.Assert.assertThat;
+import static org.junit.Assert.assertTrue;
+import static org.junit.Assert.fail;
 
 public class HttpObjectAggregatorTest {
 
@@ -38,9 +47,8 @@ public class HttpObjectAggregatorTest {
         HttpObjectAggregator aggr = new HttpObjectAggregator(1024 * 1024);
         EmbeddedChannel embedder = new EmbeddedChannel(aggr);
 
-        HttpRequest message = new DefaultHttpRequest(HttpVersion.HTTP_1_1,
-                HttpMethod.GET, "http://localhost");
-        HttpHeaders.setHeader(message, "X-Test", true);
+        HttpRequest message = new DefaultHttpRequest(HttpVersion.HTTP_1_1, HttpMethod.GET, "http://localhost");
+        message.headers().set(of("X-Test"), true);
         HttpContent chunk1 = new DefaultHttpContent(Unpooled.copiedBuffer("test", CharsetUtil.US_ASCII));
         HttpContent chunk2 = new DefaultHttpContent(Unpooled.copiedBuffer("test2", CharsetUtil.US_ASCII));
         HttpContent chunk3 = new DefaultLastHttpContent(Unpooled.EMPTY_BUFFER);
@@ -51,12 +59,12 @@ public class HttpObjectAggregatorTest {
         // this should trigger a channelRead event so return true
         assertTrue(embedder.writeInbound(chunk3));
         assertTrue(embedder.finish());
-        FullHttpRequest aggratedMessage = (FullHttpRequest) embedder.readInbound();
+        FullHttpRequest aggratedMessage = embedder.readInbound();
         assertNotNull(aggratedMessage);
 
         assertEquals(chunk1.content().readableBytes() + chunk2.content().readableBytes(),
-                HttpHeaders.getContentLength(aggratedMessage));
-        assertEquals(aggratedMessage.headers().get("X-Test"), Boolean.TRUE.toString());
+                HttpUtil.getContentLength(aggratedMessage));
+        assertEquals(aggratedMessage.headers().get(of("X-Test")), Boolean.TRUE.toString());
         checkContentBuffer(aggratedMessage);
         assertNull(embedder.readInbound());
     }
@@ -77,14 +85,13 @@ public class HttpObjectAggregatorTest {
     public void testAggregateWithTrailer() {
         HttpObjectAggregator aggr = new HttpObjectAggregator(1024 * 1024);
         EmbeddedChannel embedder = new EmbeddedChannel(aggr);
-        HttpRequest message = new DefaultHttpRequest(HttpVersion.HTTP_1_1,
-                HttpMethod.GET, "http://localhost");
-        HttpHeaders.setHeader(message, "X-Test", true);
-        HttpHeaders.setTransferEncodingChunked(message);
+        HttpRequest message = new DefaultHttpRequest(HttpVersion.HTTP_1_1, HttpMethod.GET, "http://localhost");
+        message.headers().set(of("X-Test"), true);
+        HttpUtil.setTransferEncodingChunked(message, true);
         HttpContent chunk1 = new DefaultHttpContent(Unpooled.copiedBuffer("test", CharsetUtil.US_ASCII));
         HttpContent chunk2 = new DefaultHttpContent(Unpooled.copiedBuffer("test2", CharsetUtil.US_ASCII));
         LastHttpContent trailer = new DefaultLastHttpContent();
-        trailer.trailingHeaders().set("X-Trailer", true);
+        trailer.trailingHeaders().set(of("X-Trailer"), true);
 
         assertFalse(embedder.writeInbound(message));
         assertFalse(embedder.writeInbound(chunk1));
@@ -93,55 +100,115 @@ public class HttpObjectAggregatorTest {
         // this should trigger a channelRead event so return true
         assertTrue(embedder.writeInbound(trailer));
         assertTrue(embedder.finish());
-        FullHttpRequest aggratedMessage = (FullHttpRequest) embedder.readInbound();
+        FullHttpRequest aggratedMessage = embedder.readInbound();
         assertNotNull(aggratedMessage);
 
         assertEquals(chunk1.content().readableBytes() + chunk2.content().readableBytes(),
-                HttpHeaders.getContentLength(aggratedMessage));
-        assertEquals(aggratedMessage.headers().get("X-Test"), Boolean.TRUE.toString());
-        assertEquals(aggratedMessage.trailingHeaders().get("X-Trailer"), Boolean.TRUE.toString());
+                HttpUtil.getContentLength(aggratedMessage));
+        assertEquals(aggratedMessage.headers().get(of("X-Test")), Boolean.TRUE.toString());
+        assertEquals(aggratedMessage.trailingHeaders().get(of("X-Trailer")), Boolean.TRUE.toString());
         checkContentBuffer(aggratedMessage);
         assertNull(embedder.readInbound());
     }
 
     @Test
-    public void testTooLongFrameException() {
-        HttpObjectAggregator aggr = new HttpObjectAggregator(4);
-        EmbeddedChannel embedder = new EmbeddedChannel(aggr);
-        HttpRequest message = new DefaultHttpRequest(HttpVersion.HTTP_1_1,
-                HttpMethod.GET, "http://localhost");
-        HttpContent chunk1 = releaseLater(new DefaultHttpContent(Unpooled.copiedBuffer("test", CharsetUtil.US_ASCII)));
-        HttpContent chunk2 = releaseLater(new DefaultHttpContent(Unpooled.copiedBuffer("test2", CharsetUtil.US_ASCII)));
-        HttpContent chunk3 = releaseLater(new DefaultHttpContent(Unpooled.copiedBuffer("test3", CharsetUtil.US_ASCII)));
-        HttpContent chunk4 = LastHttpContent.EMPTY_LAST_CONTENT;
+    public void testOversizedRequest() {
+        EmbeddedChannel embedder = new EmbeddedChannel(new HttpObjectAggregator(4));
+        HttpRequest message = new DefaultHttpRequest(HttpVersion.HTTP_1_1, HttpMethod.PUT, "http://localhost");
+        HttpContent chunk1 = new DefaultHttpContent(Unpooled.copiedBuffer("test", CharsetUtil.US_ASCII));
+        HttpContent chunk2 = new DefaultHttpContent(Unpooled.copiedBuffer("test2", CharsetUtil.US_ASCII));
+        HttpContent chunk3 = LastHttpContent.EMPTY_LAST_CONTENT;
 
         assertFalse(embedder.writeInbound(message));
-        assertFalse(embedder.writeInbound(chunk1.copy()));
+        assertFalse(embedder.writeInbound(chunk1));
+        assertFalse(embedder.writeInbound(chunk2));
+
+        FullHttpResponse response = embedder.readOutbound();
+        assertEquals(HttpResponseStatus.REQUEST_ENTITY_TOO_LARGE, response.status());
+        assertEquals("0", response.headers().get(HttpHeaderNames.CONTENT_LENGTH));
+        assertFalse(embedder.isOpen());
+
         try {
-            embedder.writeInbound(chunk2.copy());
+            assertFalse(embedder.writeInbound(chunk3));
             fail();
-        } catch (TooLongFrameException e) {
-            // expected
+        } catch (Exception e) {
+            assertTrue(e instanceof ClosedChannelException);
         }
-        assertFalse(embedder.writeInbound(chunk3.copy()));
-        assertFalse(embedder.writeInbound(chunk4.copy()));
+
+        assertFalse(embedder.finish());
+    }
+
+    @Test
+    public void testOversizedRequestWithoutKeepAlive() {
+        // send a HTTP/1.0 request with no keep-alive header
+        HttpRequest message = new DefaultHttpRequest(HttpVersion.HTTP_1_0, HttpMethod.PUT, "http://localhost");
+        HttpUtil.setContentLength(message, 5);
+        checkOversizedRequest(message);
+    }
+
+    @Test
+    public void testOversizedRequestWithContentLength() {
+        HttpRequest message = new DefaultHttpRequest(HttpVersion.HTTP_1_1, HttpMethod.PUT, "http://localhost");
+        HttpUtil.setContentLength(message, 5);
+        checkOversizedRequest(message);
+    }
+
+    private static void checkOversizedRequest(HttpRequest message) {
+        EmbeddedChannel embedder = new EmbeddedChannel(new HttpObjectAggregator(4));
 
         assertFalse(embedder.writeInbound(message));
-        assertFalse(embedder.writeInbound(chunk1.copy()));
-        try {
-            embedder.writeInbound(chunk2.copy());
-            fail();
-        } catch (TooLongFrameException e) {
-            // expected
+        HttpResponse response = embedder.readOutbound();
+        assertEquals(HttpResponseStatus.REQUEST_ENTITY_TOO_LARGE, response.status());
+        assertEquals("0", response.headers().get(HttpHeaderNames.CONTENT_LENGTH));
+
+        if (serverShouldCloseConnection(message, response)) {
+            assertFalse(embedder.isOpen());
+            assertFalse(embedder.finish());
+        } else {
+            assertTrue(embedder.isOpen());
         }
-        assertFalse(embedder.writeInbound(chunk3.copy()));
-        assertFalse(embedder.writeInbound(chunk4.copy()));
-        embedder.finish();
+    }
+
+    private static boolean serverShouldCloseConnection(HttpRequest message, HttpResponse response) {
+        // If the response wasn't keep-alive, the server should close the connection.
+        if (!HttpUtil.isKeepAlive(response)) {
+            return true;
+        }
+        // The connection should only be kept open if Expect: 100-continue is set,
+        // or if keep-alive is on.
+        if (HttpUtil.is100ContinueExpected(message)) {
+            return false;
+        }
+        if (HttpUtil.isKeepAlive(message)) {
+            return false;
+        }
+        return true;
+    }
+
+    @Test
+    public void testOversizedResponse() {
+        EmbeddedChannel embedder = new EmbeddedChannel(new HttpObjectAggregator(4));
+        HttpResponse message = new DefaultHttpResponse(HttpVersion.HTTP_1_1, HttpResponseStatus.OK);
+        HttpContent chunk1 = new DefaultHttpContent(Unpooled.copiedBuffer("test", CharsetUtil.US_ASCII));
+        HttpContent chunk2 = new DefaultHttpContent(Unpooled.copiedBuffer("test2", CharsetUtil.US_ASCII));
+
+        assertFalse(embedder.writeInbound(message));
+        assertFalse(embedder.writeInbound(chunk1));
+
+        try {
+            embedder.writeInbound(chunk2);
+            fail();
+        } catch (TooLongFrameException expected) {
+            // Expected
+        }
+
+        assertFalse(embedder.isOpen());
+        assertFalse(embedder.finish());
     }
 
     @Test(expected = IllegalArgumentException.class)
     public void testInvalidConstructorUsage() {
-        new HttpObjectAggregator(0);
+        new HttpObjectAggregator(-1);
     }
 
     @Test(expected = IllegalArgumentException.class)
@@ -164,10 +231,9 @@ public class HttpObjectAggregatorTest {
         HttpObjectAggregator aggr = new HttpObjectAggregator(1024 * 1024);
         EmbeddedChannel embedder = new EmbeddedChannel(aggr);
 
-        HttpRequest message = new DefaultHttpRequest(HttpVersion.HTTP_1_1,
-                HttpMethod.GET, "http://localhost");
-        HttpHeaders.setHeader(message, "X-Test", true);
-        HttpHeaders.setHeader(message, "Transfer-Encoding", "Chunked");
+        HttpRequest message = new DefaultHttpRequest(HttpVersion.HTTP_1_1, HttpMethod.PUT, "http://localhost");
+        message.headers().set(of("X-Test"), true);
+        message.headers().set(of("Transfer-Encoding"), of("Chunked"));
         HttpContent chunk1 = new DefaultHttpContent(Unpooled.copiedBuffer("test", CharsetUtil.US_ASCII));
         HttpContent chunk2 = new DefaultHttpContent(Unpooled.copiedBuffer("test2", CharsetUtil.US_ASCII));
         HttpContent chunk3 = LastHttpContent.EMPTY_LAST_CONTENT;
@@ -178,12 +244,12 @@ public class HttpObjectAggregatorTest {
         // this should trigger a channelRead event so return true
         assertTrue(embedder.writeInbound(chunk3));
         assertTrue(embedder.finish());
-        FullHttpRequest aggratedMessage = (FullHttpRequest) embedder.readInbound();
+        FullHttpRequest aggratedMessage = embedder.readInbound();
         assertNotNull(aggratedMessage);
 
         assertEquals(chunk1.content().readableBytes() + chunk2.content().readableBytes(),
-                HttpHeaders.getContentLength(aggratedMessage));
-        assertEquals(aggratedMessage.headers().get("X-Test"), Boolean.TRUE.toString());
+                HttpUtil.getContentLength(aggratedMessage));
+        assertEquals(aggratedMessage.headers().get(of("X-Test")), Boolean.TRUE.toString());
         checkContentBuffer(aggratedMessage);
         assertNull(embedder.readInbound());
     }
@@ -194,7 +260,7 @@ public class HttpObjectAggregatorTest {
         ch.writeInbound(Unpooled.copiedBuffer("GET / HTTP/1.0 with extra\r\n", CharsetUtil.UTF_8));
         Object inbound = ch.readInbound();
         assertThat(inbound, is(instanceOf(FullHttpRequest.class)));
-        assertTrue(((HttpObject) inbound).getDecoderResult().isFailure());
+        assertTrue(((DecoderResultProvider) inbound).decoderResult().isFailure());
         assertNull(ch.readInbound());
         ch.finish();
     }
@@ -205,7 +271,7 @@ public class HttpObjectAggregatorTest {
         ch.writeInbound(Unpooled.copiedBuffer("HTTP/1.0 BAD_CODE Bad Server\r\n", CharsetUtil.UTF_8));
         Object inbound = ch.readInbound();
         assertThat(inbound, is(instanceOf(FullHttpResponse.class)));
-        assertTrue(((HttpObject) inbound).getDecoderResult().isFailure());
+        assertTrue(((DecoderResultProvider) inbound).decoderResult().isFailure());
         assertNull(ch.readInbound());
         ch.finish();
     }
@@ -214,22 +280,22 @@ public class HttpObjectAggregatorTest {
     public void testOversizedRequestWith100Continue() {
         EmbeddedChannel embedder = new EmbeddedChannel(new HttpObjectAggregator(8));
 
-        // send an oversized request with 100 continue
+        // Send an oversized request with 100 continue.
         HttpRequest message = new DefaultHttpRequest(HttpVersion.HTTP_1_1, HttpMethod.PUT, "http://localhost");
-        HttpHeaders.set100ContinueExpected(message, true);
-        HttpHeaders.setContentLength(message, 16);
+        HttpUtil.set100ContinueExpected(message, true);
+        HttpUtil.setContentLength(message, 16);
 
-        HttpContent chunk1 = releaseLater(new DefaultHttpContent(Unpooled.copiedBuffer("some", CharsetUtil.US_ASCII)));
-        HttpContent chunk2 = releaseLater(new DefaultHttpContent(Unpooled.copiedBuffer("test", CharsetUtil.US_ASCII)));
+        HttpContent chunk1 = new DefaultHttpContent(Unpooled.copiedBuffer("some", CharsetUtil.US_ASCII));
+        HttpContent chunk2 = new DefaultHttpContent(Unpooled.copiedBuffer("test", CharsetUtil.US_ASCII));
         HttpContent chunk3 = LastHttpContent.EMPTY_LAST_CONTENT;
 
         // Send a request with 100-continue + large Content-Length header value.
         assertFalse(embedder.writeInbound(message));
 
         // The aggregator should respond with '417.'
-        FullHttpResponse response = (FullHttpResponse) embedder.readOutbound();
-        assertEquals(HttpResponseStatus.EXPECTATION_FAILED, response.getStatus());
-        assertEquals("0", response.headers().get(HttpHeaders.Names.CONTENT_LENGTH));
+        FullHttpResponse response = embedder.readOutbound();
+        assertEquals(HttpResponseStatus.EXPECTATION_FAILED, response.status());
+        assertEquals("0", response.headers().get(HttpHeaderNames.CONTENT_LENGTH));
 
         // An ill-behaving client could continue to send data without a respect, and such data should be discarded.
         assertFalse(embedder.writeInbound(chunk1));
@@ -249,9 +315,9 @@ public class HttpObjectAggregatorTest {
 
         assertEquals(
                 chunk2.content().readableBytes() + chunk3.content().readableBytes(),
-                HttpHeaders.getContentLength(fullMsg));
+                HttpUtil.getContentLength(fullMsg));
 
-        assertEquals(HttpHeaders.getContentLength(fullMsg), fullMsg.content().readableBytes());
+        assertEquals(HttpUtil.getContentLength(fullMsg), fullMsg.content().readableBytes());
 
         fullMsg.release();
         assertFalse(embedder.finish());
@@ -268,8 +334,8 @@ public class HttpObjectAggregatorTest {
         assertNull(embedder.readInbound());
 
         FullHttpResponse response = (FullHttpResponse) embedder.readOutbound();
-        assertEquals(HttpResponseStatus.EXPECTATION_FAILED, response.getStatus());
-        assertEquals("0", response.headers().get(HttpHeaders.Names.CONTENT_LENGTH));
+        assertEquals(HttpResponseStatus.EXPECTATION_FAILED, response.status());
+        assertEquals("0", response.headers().get(HttpHeaderNames.CONTENT_LENGTH));
 
         // Keep-alive is on by default in HTTP/1.1, so the connection should be still alive.
         assertTrue(embedder.isOpen());
@@ -278,8 +344,8 @@ public class HttpObjectAggregatorTest {
         embedder.writeInbound(Unpooled.copiedBuffer("GET /max-upload-size HTTP/1.1\r\n\r\n", CharsetUtil.US_ASCII));
 
         FullHttpRequest request = (FullHttpRequest) embedder.readInbound();
-        assertThat(request.getMethod(), is(HttpMethod.GET));
-        assertThat(request.getUri(), is("/max-upload-size"));
+        assertThat(request.method(), is(HttpMethod.GET));
+        assertThat(request.uri(), is("/max-upload-size"));
         assertThat(request.content().readableBytes(), is(0));
         request.release();
 
@@ -297,8 +363,8 @@ public class HttpObjectAggregatorTest {
         assertNull(embedder.readInbound());
 
         FullHttpResponse response = (FullHttpResponse) embedder.readOutbound();
-        assertEquals(HttpResponseStatus.EXPECTATION_FAILED, response.getStatus());
-        assertEquals("0", response.headers().get(HttpHeaders.Names.CONTENT_LENGTH));
+        assertEquals(HttpResponseStatus.EXPECTATION_FAILED, response.status());
+        assertEquals("0", response.headers().get(HttpHeaderNames.CONTENT_LENGTH));
 
         // We are forcing the connection closed if an expectation is exceeded.
         assertFalse(embedder.isOpen());
@@ -309,22 +375,22 @@ public class HttpObjectAggregatorTest {
     public void testRequestAfterOversized100ContinueAndDecoder() {
         EmbeddedChannel embedder = new EmbeddedChannel(new HttpRequestDecoder(), new HttpObjectAggregator(15));
 
-        // Write first request with Expect: 100-continue
+        // Write first request with Expect: 100-continue.
         HttpRequest message = new DefaultHttpRequest(HttpVersion.HTTP_1_1, HttpMethod.PUT, "http://localhost");
-        HttpHeaders.set100ContinueExpected(message, true);
-        HttpHeaders.setContentLength(message, 16);
+        HttpUtil.set100ContinueExpected(message, true);
+        HttpUtil.setContentLength(message, 16);
 
-        HttpContent chunk1 = releaseLater(new DefaultHttpContent(Unpooled.copiedBuffer("some", CharsetUtil.US_ASCII)));
-        HttpContent chunk2 = releaseLater(new DefaultHttpContent(Unpooled.copiedBuffer("test", CharsetUtil.US_ASCII)));
+        HttpContent chunk1 = new DefaultHttpContent(Unpooled.copiedBuffer("some", CharsetUtil.US_ASCII));
+        HttpContent chunk2 = new DefaultHttpContent(Unpooled.copiedBuffer("test", CharsetUtil.US_ASCII));
         HttpContent chunk3 = LastHttpContent.EMPTY_LAST_CONTENT;
 
         // Send a request with 100-continue + large Content-Length header value.
         assertFalse(embedder.writeInbound(message));
 
-        // The aggregator should respond with '417'
-        FullHttpResponse response = (FullHttpResponse) embedder.readOutbound();
-        assertEquals(HttpResponseStatus.EXPECTATION_FAILED, response.getStatus());
-        assertEquals("0", response.headers().get(HttpHeaders.Names.CONTENT_LENGTH));
+        // The aggregator should respond with '417'.
+        FullHttpResponse response = embedder.readOutbound();
+        assertEquals(HttpResponseStatus.EXPECTATION_FAILED, response.status());
+        assertEquals("0", response.headers().get(HttpHeaderNames.CONTENT_LENGTH));
 
         // An ill-behaving client could continue to send data without a respect, and such data should be discarded.
         assertFalse(embedder.writeInbound(chunk1));
@@ -344,9 +410,9 @@ public class HttpObjectAggregatorTest {
 
         assertEquals(
                 chunk2.content().readableBytes() + chunk3.content().readableBytes(),
-                HttpHeaders.getContentLength(fullMsg));
+                HttpUtil.getContentLength(fullMsg));
 
-        assertEquals(HttpHeaders.getContentLength(fullMsg), fullMsg.content().readableBytes());
+        assertEquals(HttpUtil.getContentLength(fullMsg), fullMsg.content().readableBytes());
 
         fullMsg.release();
         assertFalse(embedder.finish());
