@@ -15,18 +15,29 @@
  */
 package io.netty.handler.codec.http;
 
+import io.netty.buffer.ByteBuf;
 import io.netty.buffer.Unpooled;
 import io.netty.channel.embedded.EmbeddedChannel;
+import io.netty.handler.codec.PrematureChannelClosureException;
 import io.netty.handler.codec.TooLongFrameException;
-import io.netty.handler.codec.http.HttpHeaders.Names;
 import io.netty.util.CharsetUtil;
 import org.junit.Test;
 
 import java.util.Arrays;
 import java.util.List;
 
-import static org.hamcrest.CoreMatchers.*;
-import static org.junit.Assert.*;
+import static io.netty.handler.codec.http.HttpHeadersTestUtils.of;
+import static org.hamcrest.CoreMatchers.instanceOf;
+import static org.hamcrest.CoreMatchers.is;
+import static org.hamcrest.CoreMatchers.not;
+import static org.hamcrest.CoreMatchers.nullValue;
+import static org.hamcrest.CoreMatchers.sameInstance;
+import static org.junit.Assert.assertArrayEquals;
+import static org.junit.Assert.assertEquals;
+import static org.junit.Assert.assertFalse;
+import static org.junit.Assert.assertNull;
+import static org.junit.Assert.assertThat;
+import static org.junit.Assert.assertTrue;
 
 public class HttpResponseDecoderTest {
 
@@ -54,9 +65,9 @@ public class HttpResponseDecoderTest {
         ch.writeInbound(Unpooled.copiedBuffer("\r\n", CharsetUtil.US_ASCII));
         ch.writeInbound(Unpooled.copiedBuffer("\r\n", CharsetUtil.US_ASCII));
 
-        HttpResponse res = (HttpResponse) ch.readInbound();
-        assertNull(res.getDecoderResult().cause());
-        assertTrue(res.getDecoderResult().isSuccess());
+        HttpResponse res = ch.readInbound();
+        assertNull(res.decoderResult().cause());
+        assertTrue(res.decoderResult().isSuccess());
 
         assertNull(ch.readInbound());
         assertTrue(ch.finish());
@@ -64,7 +75,7 @@ public class HttpResponseDecoderTest {
     }
 
     /**
-     * Complementary test case of {@link #testMaxHeaderSize1()}. When it actually exceeds the maximum, it should fail.
+     * Complementary test case of {@link #testMaxHeaderSize1()} When it actually exceeds the maximum, it should fail.
      */
     @Test
     public void testMaxHeaderSize2() {
@@ -86,8 +97,8 @@ public class HttpResponseDecoderTest {
         ch.writeInbound(Unpooled.copiedBuffer("\r\n", CharsetUtil.US_ASCII));
         ch.writeInbound(Unpooled.copiedBuffer("\r\n", CharsetUtil.US_ASCII));
 
-        HttpResponse res = (HttpResponse) ch.readInbound();
-        assertTrue(res.getDecoderResult().cause() instanceof TooLongFrameException);
+        HttpResponse res = ch.readInbound();
+        assertTrue(res.decoderResult().cause() instanceof TooLongFrameException);
 
         assertFalse(ch.finish());
         assertNull(ch.readInbound());
@@ -99,9 +110,9 @@ public class HttpResponseDecoderTest {
         ch.writeInbound(Unpooled.copiedBuffer("HTTP/1.1 200 OK\r\nTransfer-Encoding: chunked\r\n\r\n",
                 CharsetUtil.US_ASCII));
 
-        HttpResponse res = (HttpResponse) ch.readInbound();
-        assertThat(res.getProtocolVersion(), sameInstance(HttpVersion.HTTP_1_1));
-        assertThat(res.getStatus(), is(HttpResponseStatus.OK));
+        HttpResponse res = ch.readInbound();
+        assertThat(res.protocolVersion(), sameInstance(HttpVersion.HTTP_1_1));
+        assertThat(res.status(), is(HttpResponseStatus.OK));
 
         byte[] data = new byte[64];
         for (int i = 0; i < data.length; i++) {
@@ -112,7 +123,7 @@ public class HttpResponseDecoderTest {
             assertFalse(ch.writeInbound(Unpooled.copiedBuffer(Integer.toHexString(data.length) + "\r\n",
                     CharsetUtil.US_ASCII)));
             assertTrue(ch.writeInbound(Unpooled.wrappedBuffer(data)));
-            HttpContent content = (HttpContent) ch.readInbound();
+            HttpContent content = ch.readInbound();
             assertEquals(data.length, content.content().readableBytes());
 
             byte[] decodedData = new byte[data.length];
@@ -127,7 +138,7 @@ public class HttpResponseDecoderTest {
         ch.writeInbound(Unpooled.copiedBuffer("0\r\n\r\n", CharsetUtil.US_ASCII));
 
         // Ensure the last chunk was decoded.
-        LastHttpContent content = (LastHttpContent) ch.readInbound();
+        LastHttpContent content = ch.readInbound();
         assertFalse(content.content().isReadable());
         content.release();
 
@@ -141,9 +152,9 @@ public class HttpResponseDecoderTest {
         ch.writeInbound(
                 Unpooled.copiedBuffer("HTTP/1.1 200 OK\r\nTransfer-Encoding: chunked\r\n\r\n", CharsetUtil.US_ASCII));
 
-        HttpResponse res = (HttpResponse) ch.readInbound();
-        assertThat(res.getProtocolVersion(), sameInstance(HttpVersion.HTTP_1_1));
-        assertThat(res.getStatus(), is(HttpResponseStatus.OK));
+        HttpResponse res = ch.readInbound();
+        assertThat(res.protocolVersion(), sameInstance(HttpVersion.HTTP_1_1));
+        assertThat(res.status(), is(HttpResponseStatus.OK));
 
         byte[] data = new byte[64];
         for (int i = 0; i < data.length; i++) {
@@ -156,12 +167,12 @@ public class HttpResponseDecoderTest {
             assertTrue(ch.writeInbound(Unpooled.wrappedBuffer(data)));
 
             byte[] decodedData = new byte[data.length];
-            HttpContent content = (HttpContent) ch.readInbound();
+            HttpContent content = ch.readInbound();
             assertEquals(32, content.content().readableBytes());
             content.content().readBytes(decodedData, 0, 32);
             content.release();
 
-            content = (HttpContent) ch.readInbound();
+            content = ch.readInbound();
             assertEquals(32, content.content().readableBytes());
 
             content.content().readBytes(decodedData, 32, 32);
@@ -176,7 +187,7 @@ public class HttpResponseDecoderTest {
         ch.writeInbound(Unpooled.copiedBuffer("0\r\n\r\n", CharsetUtil.US_ASCII));
 
         // Ensure the last chunk was decoded.
-        LastHttpContent content = (LastHttpContent) ch.readInbound();
+        LastHttpContent content = ch.readInbound();
         assertFalse(content.content().isReadable());
         content.release();
 
@@ -190,16 +201,16 @@ public class HttpResponseDecoderTest {
         ch.writeInbound(Unpooled.copiedBuffer("HTTP/1.1 200 OK\r\n\r\n", CharsetUtil.US_ASCII));
 
         // Read the response headers.
-        HttpResponse res = (HttpResponse) ch.readInbound();
-        assertThat(res.getProtocolVersion(), sameInstance(HttpVersion.HTTP_1_1));
-        assertThat(res.getStatus(), is(HttpResponseStatus.OK));
+        HttpResponse res = ch.readInbound();
+        assertThat(res.protocolVersion(), sameInstance(HttpVersion.HTTP_1_1));
+        assertThat(res.status(), is(HttpResponseStatus.OK));
         assertThat(ch.readInbound(), is(nullValue()));
 
         // Close the connection without sending anything.
         assertTrue(ch.finish());
 
         // The decoder should still produce the last content.
-        LastHttpContent content = (LastHttpContent) ch.readInbound();
+        LastHttpContent content = ch.readInbound();
         assertThat(content.content().isReadable(), is(false));
         content.release();
 
@@ -215,12 +226,12 @@ public class HttpResponseDecoderTest {
         ch.writeInbound(Unpooled.copiedBuffer("HTTP/1.1 200 OK\r\n\r\n12345678", CharsetUtil.US_ASCII));
 
         // Read the response headers.
-        HttpResponse res = (HttpResponse) ch.readInbound();
-        assertThat(res.getProtocolVersion(), sameInstance(HttpVersion.HTTP_1_1));
-        assertThat(res.getStatus(), is(HttpResponseStatus.OK));
+        HttpResponse res = ch.readInbound();
+        assertThat(res.protocolVersion(), sameInstance(HttpVersion.HTTP_1_1));
+        assertThat(res.status(), is(HttpResponseStatus.OK));
 
         // Read the partial content.
-        HttpContent content = (HttpContent) ch.readInbound();
+        HttpContent content = ch.readInbound();
         assertThat(content.content().toString(CharsetUtil.US_ASCII), is("12345678"));
         assertThat(content, is(not(instanceOf(LastHttpContent.class))));
         content.release();
@@ -231,7 +242,7 @@ public class HttpResponseDecoderTest {
         assertTrue(ch.finish());
 
         // The decoder should still produce the last content.
-        LastHttpContent lastContent = (LastHttpContent) ch.readInbound();
+        LastHttpContent lastContent = ch.readInbound();
         assertThat(lastContent.content().isReadable(), is(false));
         lastContent.release();
 
@@ -246,10 +257,10 @@ public class HttpResponseDecoderTest {
                 Unpooled.copiedBuffer("HTTP/1.1 200 OK\r\nTransfer-Encoding: chunked\r\n\r\n", CharsetUtil.US_ASCII));
 
         // Read the response headers.
-        HttpResponse res = (HttpResponse) ch.readInbound();
-        assertThat(res.getProtocolVersion(), sameInstance(HttpVersion.HTTP_1_1));
-        assertThat(res.getStatus(), is(HttpResponseStatus.OK));
-        assertThat(res.headers().get(Names.TRANSFER_ENCODING), is("chunked"));
+        HttpResponse res = ch.readInbound();
+        assertThat(res.protocolVersion(), sameInstance(HttpVersion.HTTP_1_1));
+        assertThat(res.status(), is(HttpResponseStatus.OK));
+        assertThat(res.headers().get(HttpHeaderNames.TRANSFER_ENCODING), is("chunked"));
         assertThat(ch.readInbound(), is(nullValue()));
 
         // Close the connection without sending anything.
@@ -267,13 +278,13 @@ public class HttpResponseDecoderTest {
                 "HTTP/1.1 200 OK\r\nTransfer-Encoding: chunked\r\n\r\n8\r\n12345678", CharsetUtil.US_ASCII));
 
         // Read the response headers.
-        HttpResponse res = (HttpResponse) ch.readInbound();
-        assertThat(res.getProtocolVersion(), sameInstance(HttpVersion.HTTP_1_1));
-        assertThat(res.getStatus(), is(HttpResponseStatus.OK));
-        assertThat(res.headers().get(Names.TRANSFER_ENCODING), is("chunked"));
+        HttpResponse res = ch.readInbound();
+        assertThat(res.protocolVersion(), sameInstance(HttpVersion.HTTP_1_1));
+        assertThat(res.status(), is(HttpResponseStatus.OK));
+        assertThat(res.headers().get(HttpHeaderNames.TRANSFER_ENCODING), is("chunked"));
 
         // Read the partial content.
-        HttpContent content = (HttpContent) ch.readInbound();
+        HttpContent content = ch.readInbound();
         assertThat(content.content().toString(CharsetUtil.US_ASCII), is("12345678"));
         assertThat(content, is(not(instanceOf(LastHttpContent.class))));
         content.release();
@@ -292,14 +303,14 @@ public class HttpResponseDecoderTest {
         EmbeddedChannel ch = new EmbeddedChannel(new HttpResponseDecoder());
         ch.writeInbound(Unpooled.copiedBuffer("HTTP/1.1 200 OK\r\n\r\n", CharsetUtil.US_ASCII));
 
-        HttpResponse res = (HttpResponse) ch.readInbound();
-        assertThat(res.getProtocolVersion(), sameInstance(HttpVersion.HTTP_1_1));
-        assertThat(res.getStatus(), is(HttpResponseStatus.OK));
+        HttpResponse res = ch.readInbound();
+        assertThat(res.protocolVersion(), sameInstance(HttpVersion.HTTP_1_1));
+        assertThat(res.status(), is(HttpResponseStatus.OK));
         assertThat(ch.readInbound(), is(nullValue()));
 
         assertThat(ch.finish(), is(true));
 
-        LastHttpContent content = (LastHttpContent) ch.readInbound();
+        LastHttpContent content = ch.readInbound();
         assertThat(content.content().isReadable(), is(false));
         content.release();
 
@@ -311,19 +322,46 @@ public class HttpResponseDecoderTest {
         EmbeddedChannel ch = new EmbeddedChannel(new HttpResponseDecoder());
         ch.writeInbound(Unpooled.copiedBuffer("HTTP/1.1 200 OK\r\n\r\n", CharsetUtil.US_ASCII));
 
-        HttpResponse res = (HttpResponse) ch.readInbound();
-        assertThat(res.getProtocolVersion(), sameInstance(HttpVersion.HTTP_1_1));
-        assertThat(res.getStatus(), is(HttpResponseStatus.OK));
+        HttpResponse res = ch.readInbound();
+        assertThat(res.protocolVersion(), sameInstance(HttpVersion.HTTP_1_1));
+        assertThat(res.status(), is(HttpResponseStatus.OK));
         assertThat(ch.readInbound(), is(nullValue()));
 
         ch.writeInbound(Unpooled.wrappedBuffer(new byte[1024]));
-        HttpContent content = (HttpContent) ch.readInbound();
+        HttpContent content = ch.readInbound();
         assertThat(content.content().readableBytes(), is(1024));
         content.release();
 
         assertThat(ch.finish(), is(true));
 
-        LastHttpContent lastContent = (LastHttpContent) ch.readInbound();
+        LastHttpContent lastContent = ch.readInbound();
+        assertThat(lastContent.content().isReadable(), is(false));
+        lastContent.release();
+
+        assertThat(ch.readInbound(), is(nullValue()));
+    }
+
+    @Test
+    public void testLastResponseWithHeaderRemoveTrailingSpaces() {
+        EmbeddedChannel ch = new EmbeddedChannel(new HttpResponseDecoder());
+        ch.writeInbound(Unpooled.copiedBuffer(
+                "HTTP/1.1 200 OK\r\nX-Header: h2=h2v2; Expires=Wed, 09-Jun-2021 10:18:14 GMT       \r\n\r\n",
+                CharsetUtil.US_ASCII));
+
+        HttpResponse res = ch.readInbound();
+        assertThat(res.protocolVersion(), sameInstance(HttpVersion.HTTP_1_1));
+        assertThat(res.status(), is(HttpResponseStatus.OK));
+        assertThat(res.headers().get(of("X-Header")), is("h2=h2v2; Expires=Wed, 09-Jun-2021 10:18:14 GMT"));
+        assertThat(ch.readInbound(), is(nullValue()));
+
+        ch.writeInbound(Unpooled.wrappedBuffer(new byte[1024]));
+        HttpContent content = ch.readInbound();
+        assertThat(content.content().readableBytes(), is(1024));
+        content.release();
+
+        assertThat(ch.finish(), is(true));
+
+        LastHttpContent lastContent = ch.readInbound();
         assertThat(lastContent.content().isReadable(), is(false));
         lastContent.release();
 
@@ -343,15 +381,15 @@ public class HttpResponseDecoderTest {
                         "\r\n",
                 CharsetUtil.US_ASCII));
 
-        HttpResponse res = (HttpResponse) ch.readInbound();
-        assertThat(res.getProtocolVersion(), sameInstance(HttpVersion.HTTP_1_1));
-        assertThat(res.getStatus(), is(HttpResponseStatus.OK));
+        HttpResponse res = ch.readInbound();
+        assertThat(res.protocolVersion(), sameInstance(HttpVersion.HTTP_1_1));
+        assertThat(res.status(), is(HttpResponseStatus.OK));
 
-        LastHttpContent lastContent = (LastHttpContent) ch.readInbound();
+        LastHttpContent lastContent = ch.readInbound();
         assertThat(lastContent.content().isReadable(), is(false));
         HttpHeaders headers = lastContent.trailingHeaders();
         assertEquals(1, headers.names().size());
-        List<String> values = headers.getAll("Set-Cookie");
+        List<String> values = headers.getAll(of("Set-Cookie"));
         assertEquals(2, values.size());
         assertTrue(values.contains("t1=t1v1"));
         assertTrue(values.contains("t2=t2v2; Expires=Wed, 09-Jun-2021 10:18:14 GMT"));
@@ -393,15 +431,15 @@ public class HttpResponseDecoderTest {
         }
 
         ch.writeInbound(Unpooled.wrappedBuffer(content, headerLength, content.length - headerLength));
-        HttpResponse res = (HttpResponse) ch.readInbound();
-        assertThat(res.getProtocolVersion(), sameInstance(HttpVersion.HTTP_1_1));
-        assertThat(res.getStatus(), is(HttpResponseStatus.OK));
+        HttpResponse res = ch.readInbound();
+        assertThat(res.protocolVersion(), sameInstance(HttpVersion.HTTP_1_1));
+        assertThat(res.status(), is(HttpResponseStatus.OK));
 
-        LastHttpContent lastContent = (LastHttpContent) ch.readInbound();
+        LastHttpContent lastContent = ch.readInbound();
         assertThat(lastContent.content().isReadable(), is(false));
         HttpHeaders headers = lastContent.trailingHeaders();
         assertEquals(1, headers.names().size());
-        List<String> values = headers.getAll("Set-Cookie");
+        List<String> values = headers.getAll(of("Set-Cookie"));
         assertEquals(2, values.size());
         assertTrue(values.contains("t1=t1v1"));
         assertTrue(values.contains("t2=t2v2; Expires=Wed, 09-Jun-2021 10:18:14 GMT"));
@@ -426,16 +464,16 @@ public class HttpResponseDecoderTest {
         ch.writeInbound(Unpooled.wrappedBuffer(data, 0, data.length / 2));
         ch.writeInbound(Unpooled.wrappedBuffer(data, 5, data.length / 2));
 
-        HttpResponse res = (HttpResponse) ch.readInbound();
-        assertThat(res.getProtocolVersion(), sameInstance(HttpVersion.HTTP_1_1));
-        assertThat(res.getStatus(), is(HttpResponseStatus.OK));
+        HttpResponse res = ch.readInbound();
+        assertThat(res.protocolVersion(), sameInstance(HttpVersion.HTTP_1_1));
+        assertThat(res.status(), is(HttpResponseStatus.OK));
 
-        HttpContent firstContent = (HttpContent) ch.readInbound();
+        HttpContent firstContent = ch.readInbound();
         assertThat(firstContent.content().readableBytes(), is(5));
         assertEquals(Unpooled.wrappedBuffer(data, 0, 5), firstContent.content());
         firstContent.release();
 
-        LastHttpContent lastContent = (LastHttpContent) ch.readInbound();
+        LastHttpContent lastContent = ch.readInbound();
         assertEquals(5, lastContent.content().readableBytes());
         assertEquals(Unpooled.wrappedBuffer(data, 5, 5), lastContent.content());
         lastContent.release();
@@ -474,16 +512,16 @@ public class HttpResponseDecoderTest {
         ch.writeInbound(Unpooled.wrappedBuffer(data, 0, data.length / 2));
         ch.writeInbound(Unpooled.wrappedBuffer(data, 5, data.length / 2));
 
-        HttpResponse res = (HttpResponse) ch.readInbound();
-        assertThat(res.getProtocolVersion(), sameInstance(HttpVersion.HTTP_1_1));
-        assertThat(res.getStatus(), is(HttpResponseStatus.OK));
+        HttpResponse res = ch.readInbound();
+        assertThat(res.protocolVersion(), sameInstance(HttpVersion.HTTP_1_1));
+        assertThat(res.status(), is(HttpResponseStatus.OK));
 
-        HttpContent firstContent = (HttpContent) ch.readInbound();
+        HttpContent firstContent = ch.readInbound();
         assertThat(firstContent.content().readableBytes(), is(5));
         assertEquals(Unpooled.wrappedBuffer(data, 0, 5), firstContent.content());
         firstContent.release();
 
-        LastHttpContent lastContent = (LastHttpContent) ch.readInbound();
+        LastHttpContent lastContent = ch.readInbound();
         assertEquals(5, lastContent.content().readableBytes());
         assertEquals(Unpooled.wrappedBuffer(data, 5, 5), lastContent.content());
         lastContent.release();
@@ -504,10 +542,10 @@ public class HttpResponseDecoderTest {
         EmbeddedChannel ch = new EmbeddedChannel(new HttpResponseDecoder());
         ch.writeInbound(Unpooled.wrappedBuffer(data));
 
-        HttpResponse res = (HttpResponse) ch.readInbound();
-        assertThat(res.getProtocolVersion(), sameInstance(HttpVersion.HTTP_1_1));
-        assertThat(res.getStatus(), is(HttpResponseStatus.SWITCHING_PROTOCOLS));
-        HttpContent content = (HttpContent) ch.readInbound();
+        HttpResponse res = ch.readInbound();
+        assertThat(res.protocolVersion(), sameInstance(HttpVersion.HTTP_1_1));
+        assertThat(res.status(), is(HttpResponseStatus.SWITCHING_PROTOCOLS));
+        HttpContent content = ch.readInbound();
         assertThat(content.content().readableBytes(), is(16));
         content.release();
 
@@ -531,16 +569,25 @@ public class HttpResponseDecoderTest {
         EmbeddedChannel ch = new EmbeddedChannel(new HttpResponseDecoder());
         ch.writeInbound(Unpooled.wrappedBuffer(data, otherData));
 
-        HttpResponse res = (HttpResponse) ch.readInbound();
-        assertThat(res.getProtocolVersion(), sameInstance(HttpVersion.HTTP_1_1));
-        assertThat(res.getStatus(), is(HttpResponseStatus.SWITCHING_PROTOCOLS));
-        HttpContent content = (HttpContent) ch.readInbound();
+        HttpResponse res = ch.readInbound();
+        assertThat(res.protocolVersion(), sameInstance(HttpVersion.HTTP_1_1));
+        assertThat(res.status(), is(HttpResponseStatus.SWITCHING_PROTOCOLS));
+        HttpContent content = ch.readInbound();
         assertThat(content.content().readableBytes(), is(16));
         content.release();
 
         assertThat(ch.finish(), is(true));
 
-        assertEquals(ch.readInbound(), Unpooled.wrappedBuffer(otherData));
+        ByteBuf expected = Unpooled.wrappedBuffer(otherData);
+        ByteBuf buffer = ch.readInbound();
+        try {
+            assertEquals(expected, buffer);
+        } finally {
+            expected.release();
+            if (buffer != null) {
+                buffer.release();
+            }
+        }
     }
 
     @Test
@@ -559,11 +606,11 @@ public class HttpResponseDecoderTest {
         ch.writeInbound(Unpooled.wrappedBuffer(data));
 
         // Garbage input should generate the 999 Unknown response.
-        HttpResponse res = (HttpResponse) ch.readInbound();
-        assertThat(res.getProtocolVersion(), sameInstance(HttpVersion.HTTP_1_0));
-        assertThat(res.getStatus().code(), is(999));
-        assertThat(res.getDecoderResult().isFailure(), is(true));
-        assertThat(res.getDecoderResult().isFinished(), is(true));
+        HttpResponse res = ch.readInbound();
+        assertThat(res.protocolVersion(), sameInstance(HttpVersion.HTTP_1_0));
+        assertThat(res.status().code(), is(999));
+        assertThat(res.decoderResult().isFailure(), is(true));
+        assertThat(res.decoderResult().isFinished(), is(true));
         assertThat(ch.readInbound(), is(nullValue()));
 
         // More garbage should not generate anything (i.e. the decoder discards anything beyond this point.)
@@ -591,8 +638,8 @@ public class HttpResponseDecoderTest {
         assertThat(channel.readInbound(), is(instanceOf(HttpResponse.class)));
 
         // Ensure that the decoder generates the last chunk with correct decoder result.
-        LastHttpContent invalidChunk = (LastHttpContent) channel.readInbound();
-        assertThat(invalidChunk.getDecoderResult().isFailure(), is(true));
+        LastHttpContent invalidChunk = channel.readInbound();
+        assertThat(invalidChunk.decoderResult().isFailure(), is(true));
         invalidChunk.release();
 
         // And no more messages should be produced by the decoder.
@@ -601,4 +648,17 @@ public class HttpResponseDecoderTest {
         // .. even after the connection is closed.
         assertThat(channel.finish(), is(false));
     }
+
+    @Test
+    public void testConnectionClosedBeforeHeadersReceived() {
+        EmbeddedChannel channel = new EmbeddedChannel(new HttpResponseDecoder());
+        String responseInitialLine =
+                "HTTP/1.1 200 OK\r\n";
+        assertFalse(channel.writeInbound(Unpooled.copiedBuffer(responseInitialLine, CharsetUtil.US_ASCII)));
+        assertTrue(channel.finish());
+        HttpMessage message = channel.readInbound();
+        assertTrue(message.decoderResult().isFailure());
+        assertThat(message.decoderResult().cause(), instanceOf(PrematureChannelClosureException.class));
+        assertNull(channel.readInbound());
+    }
 }
