@@ -23,18 +23,24 @@ import io.netty.handler.codec.DecoderResult;
 import io.netty.handler.codec.http.DefaultFullHttpRequest;
 import io.netty.handler.codec.http.DefaultHttpContent;
 import io.netty.handler.codec.http.DefaultHttpRequest;
+import io.netty.handler.codec.http.DefaultLastHttpContent;
 import io.netty.handler.codec.http.HttpContent;
-import io.netty.handler.codec.http.HttpHeaders;
+import io.netty.handler.codec.http.HttpHeaderNames;
+import io.netty.handler.codec.http.HttpHeaderValues;
 import io.netty.handler.codec.http.HttpMethod;
 import io.netty.handler.codec.http.HttpVersion;
 import io.netty.handler.codec.http.LastHttpContent;
 import io.netty.util.CharsetUtil;
 import org.junit.Test;
 
+import java.nio.charset.UnsupportedCharsetException;
 import java.util.Arrays;
 
-import static io.netty.util.ReferenceCountUtil.*;
-import static org.junit.Assert.*;
+import static org.junit.Assert.assertEquals;
+import static org.junit.Assert.assertFalse;
+import static org.junit.Assert.assertNotNull;
+import static org.junit.Assert.assertTrue;
+import static org.junit.Assert.fail;
 
 /** {@link HttpPostRequestDecoder} test case. */
 public class HttpPostRequestDecoderTest {
@@ -51,8 +57,7 @@ public class HttpPostRequestDecoderTest {
     }
 
     private static void testBinaryStreamUpload(boolean withSpace) throws Exception {
-        // Boundary starts here with '=' to check against issue https://github.com/netty/netty/issues/3004
-        final String boundary = "=dLV9Wyq26L_-JQxk6ferf-RT153LhOO";
+        final String boundary = "dLV9Wyq26L_-JQxk6ferf-RT153LhOO";
         final String contentTypeValue;
         if (withSpace) {
             contentTypeValue = "multipart/form-data; boundary=" + boundary;
@@ -63,8 +68,8 @@ public class HttpPostRequestDecoderTest {
                 "http://localhost");
 
         req.setDecoderResult(DecoderResult.SUCCESS);
-        req.headers().add(HttpHeaders.Names.CONTENT_TYPE, contentTypeValue);
-        req.headers().add(HttpHeaders.Names.TRANSFER_ENCODING, HttpHeaders.Values.CHUNKED);
+        req.headers().add(HttpHeaderNames.CONTENT_TYPE, contentTypeValue);
+        req.headers().add(HttpHeaderNames.TRANSFER_ENCODING, HttpHeaderValues.CHUNKED);
 
         // Force to use memory-based data.
         final DefaultHttpDataFactory inMemoryFactory = new DefaultHttpDataFactory(false);
@@ -81,8 +86,8 @@ public class HttpPostRequestDecoderTest {
             // Create decoder instance to test.
             final HttpPostRequestDecoder decoder = new HttpPostRequestDecoder(inMemoryFactory, req);
 
-            decoder.offer(releaseLater(new DefaultHttpContent(Unpooled.copiedBuffer(body, CharsetUtil.UTF_8))));
-            decoder.offer(releaseLater(new DefaultHttpContent(Unpooled.EMPTY_BUFFER)));
+            decoder.offer(new DefaultHttpContent(Unpooled.copiedBuffer(body, CharsetUtil.UTF_8)));
+            decoder.offer(new DefaultHttpContent(Unpooled.EMPTY_BUFFER));
 
             // Validate it's enough chunks to decode upload.
             assertTrue(decoder.hasNext());
@@ -107,8 +112,8 @@ public class HttpPostRequestDecoderTest {
                 "http://localhost");
 
         req.setDecoderResult(DecoderResult.SUCCESS);
-        req.headers().add(HttpHeaders.Names.CONTENT_TYPE, "multipart/form-data; boundary=" + boundary);
-        req.headers().add(HttpHeaders.Names.TRANSFER_ENCODING, HttpHeaders.Values.CHUNKED);
+        req.headers().add(HttpHeaderNames.CONTENT_TYPE, "multipart/form-data; boundary=" + boundary);
+        req.headers().add(HttpHeaderNames.TRANSFER_ENCODING, HttpHeaderValues.CHUNKED);
 
         // Force to use memory-based data.
         final DefaultHttpDataFactory inMemoryFactory = new DefaultHttpDataFactory(false);
@@ -139,10 +144,10 @@ public class HttpPostRequestDecoderTest {
         final DefaultHttpDataFactory inMemoryFactory = new DefaultHttpDataFactory(false);
         // Build test case
         String extradata = "aaaa";
-        String [] datas = new String[5];
+        String[] datas = new String[5];
         for (int i = 0; i < 4; i++) {
             datas[i] = extradata;
-            for (int j = 0; j < i ; j++) {
+            for (int j = 0; j < i; j++) {
                 datas[i] += '\r';
             }
         }
@@ -151,8 +156,8 @@ public class HttpPostRequestDecoderTest {
             final DefaultFullHttpRequest req = new DefaultFullHttpRequest(HttpVersion.HTTP_1_1, HttpMethod.POST,
                     "http://localhost");
             req.setDecoderResult(DecoderResult.SUCCESS);
-            req.headers().add(HttpHeaders.Names.CONTENT_TYPE, "multipart/form-data; boundary=" + boundary);
-            req.headers().add(HttpHeaders.Names.TRANSFER_ENCODING, HttpHeaders.Values.CHUNKED);
+            req.headers().add(HttpHeaderNames.CONTENT_TYPE, "multipart/form-data; boundary=" + boundary);
+            req.headers().add(HttpHeaderNames.TRANSFER_ENCODING, HttpHeaderValues.CHUNKED);
             final String body =
                     "--" + boundary + "\r\n" +
                             "Content-Disposition: form-data; name=\"file" + i + "\"\r\n" +
@@ -169,7 +174,7 @@ public class HttpPostRequestDecoderTest {
             InterfaceHttpData httpdata = decoder.getBodyHttpData("file" + i);
             assertNotNull(httpdata);
             Attribute attribute = (Attribute) httpdata;
-            byte []datar = attribute.get();
+            byte[] datar = attribute.get();
             assertNotNull(datar);
             assertEquals(datas[i].getBytes(CharsetUtil.UTF_8).length, datar.length);
 
@@ -186,8 +191,8 @@ public class HttpPostRequestDecoderTest {
                 "http://localhost");
 
         req.setDecoderResult(DecoderResult.SUCCESS);
-        req.headers().add(HttpHeaders.Names.CONTENT_TYPE, "multipart/form-data; boundary=\"" + boundary + '"');
-        req.headers().add(HttpHeaders.Names.TRANSFER_ENCODING, HttpHeaders.Values.CHUNKED);
+        req.headers().add(HttpHeaderNames.CONTENT_TYPE, "multipart/form-data; boundary=\"" + boundary + '"');
+        req.headers().add(HttpHeaderNames.TRANSFER_ENCODING, HttpHeaderValues.CHUNKED);
 
         // Force to use memory-based data.
         final DefaultHttpDataFactory inMemoryFactory = new DefaultHttpDataFactory(false);
@@ -219,10 +224,10 @@ public class HttpPostRequestDecoderTest {
         DefaultHttpRequest aRequest = new DefaultHttpRequest(HttpVersion.HTTP_1_1,
                                                              HttpMethod.POST,
                                                              "http://localhost");
-        aRequest.headers().set(HttpHeaders.Names.CONTENT_TYPE,
+        aRequest.headers().set(HttpHeaderNames.CONTENT_TYPE,
                                "multipart/form-data; boundary=" + boundary);
-        aRequest.headers().set(HttpHeaders.Names.TRANSFER_ENCODING,
-                               HttpHeaders.Values.CHUNKED);
+        aRequest.headers().set(HttpHeaderNames.TRANSFER_ENCODING,
+                               HttpHeaderValues.CHUNKED);
 
         HttpPostRequestDecoder aDecoder = new HttpPostRequestDecoder(aMemFactory, aRequest);
 
@@ -251,8 +256,8 @@ public class HttpPostRequestDecoderTest {
         aSmallBuf.writeBytes(aBytes, 0, split);
         aLargeBuf.writeBytes(aBytes, split, aBytes.length - split);
 
-        aDecoder.offer(releaseLater(new DefaultHttpContent(aSmallBuf)));
-        aDecoder.offer(releaseLater(new DefaultHttpContent(aLargeBuf)));
+        aDecoder.offer(new DefaultHttpContent(aSmallBuf));
+        aDecoder.offer(new DefaultHttpContent(aLargeBuf));
 
         aDecoder.offer(LastHttpContent.EMPTY_LAST_CONTENT);
 
@@ -328,7 +333,7 @@ public class HttpPostRequestDecoderTest {
         final String boundary = "dLV9Wyq26L_-JQxk6ferf-RT153LhOO";
         final DefaultFullHttpRequest req = new DefaultFullHttpRequest(HttpVersion.HTTP_1_1, HttpMethod.POST,
                 "http://localhost");
-        req.headers().add(HttpHeaders.Names.CONTENT_TYPE, "multipart/form-data; boundary=" + boundary);
+        req.headers().add(HttpHeaderNames.CONTENT_TYPE, "multipart/form-data; boundary=" + boundary);
         // Force to use memory-based data.
         final DefaultHttpDataFactory inMemoryFactory = new DefaultHttpDataFactory(false);
         final String data = "asdf";
@@ -353,7 +358,7 @@ public class HttpPostRequestDecoderTest {
         final String boundary = "dLV9Wyq26L_-JQxk6ferf-RT153LhOO";
         final DefaultFullHttpRequest req = new DefaultFullHttpRequest(HttpVersion.HTTP_1_1, HttpMethod.POST,
                                                                       "http://localhost");
-        req.headers().add(HttpHeaders.Names.CONTENT_TYPE, "multipart/form-data; boundary=" + boundary);
+        req.headers().add(HttpHeaderNames.CONTENT_TYPE, "multipart/form-data; boundary=" + boundary);
         // Force to use memory-based data.
         final DefaultHttpDataFactory inMemoryFactory = new DefaultHttpDataFactory(false);
         final String data = "asdf";
@@ -385,8 +390,8 @@ public class HttpPostRequestDecoderTest {
                 "http://localhost");
 
         req.setDecoderResult(DecoderResult.SUCCESS);
-        req.headers().add(HttpHeaders.Names.CONTENT_TYPE, "multipart/form-data; boundary=" + boundary);
-        req.headers().add(HttpHeaders.Names.TRANSFER_ENCODING, HttpHeaders.Values.CHUNKED);
+        req.headers().add(HttpHeaderNames.CONTENT_TYPE, "multipart/form-data; boundary=" + boundary);
+        req.headers().add(HttpHeaderNames.TRANSFER_ENCODING, HttpHeaderValues.CHUNKED);
 
         // Force to use memory-based data.
         final DefaultHttpDataFactory inMemoryFactory = new DefaultHttpDataFactory(false);
@@ -406,4 +411,84 @@ public class HttpPostRequestDecoderTest {
         assertFalse(decoder.getBodyHttpDatas().isEmpty());
         decoder.destroy();
     }
+
+    @Test
+    public void testMultipartRequestWithFileInvalidCharset() throws Exception {
+        final String boundary = "dLV9Wyq26L_-JQxk6ferf-RT153LhOO";
+        final DefaultFullHttpRequest req = new DefaultFullHttpRequest(HttpVersion.HTTP_1_1, HttpMethod.POST,
+            "http://localhost");
+        req.headers().add(HttpHeaderNames.CONTENT_TYPE, "multipart/form-data; boundary=" + boundary);
+        // Force to use memory-based data.
+        final DefaultHttpDataFactory inMemoryFactory = new DefaultHttpDataFactory(false);
+        final String data = "asdf";
+        final String filename = "tmp;0.txt";
+        final String body =
+            "--" + boundary + "\r\n" +
+                "Content-Disposition: form-data; name=\"file\"; filename=\"" + filename + "\"\r\n" +
+                "Content-Type: image/gif; charset=ABCD\r\n" +
+                "\r\n" +
+                data + "\r\n" +
+                "--" + boundary + "--\r\n";
+
+        req.content().writeBytes(body.getBytes(CharsetUtil.UTF_8));
+        // Create decoder instance to test.
+        try {
+            new HttpPostRequestDecoder(inMemoryFactory, req);
+            fail("Was expecting an ErrorDataDecoderException");
+        } catch (HttpPostRequestDecoder.ErrorDataDecoderException e) {
+            assertTrue(e.getCause() instanceof UnsupportedCharsetException);
+        } finally {
+            req.release();
+        }
+    }
+
+    @Test
+    public void testMultipartRequestWithFieldInvalidCharset() throws Exception {
+        final String boundary = "dLV9Wyq26L_-JQxk6ferf-RT153LhOO";
+        final DefaultFullHttpRequest req = new DefaultFullHttpRequest(HttpVersion.HTTP_1_1, HttpMethod.POST,
+            "http://localhost");
+        req.headers().add(HttpHeaderNames.CONTENT_TYPE, "multipart/form-data; boundary=" + boundary);
+        // Force to use memory-based data.
+        final DefaultHttpDataFactory inMemoryFactory = new DefaultHttpDataFactory(false);
+        final String aData = "some data would be here. the data should be long enough that it " +
+            "will be longer than the original buffer length of 256 bytes in " +
+            "the HttpPostRequestDecoder in order to trigger the issue. Some more " +
+            "data just to be on the safe side.";
+        final String body =
+            "--" + boundary + "\r\n" +
+                "Content-Disposition: form-data; name=\"root\"\r\n" +
+                "Content-Type: text/plain; charset=ABCD\r\n" +
+                "\r\n" +
+                aData +
+                "\r\n" +
+                "--" + boundary + "--\r\n";
+
+        req.content().writeBytes(body.getBytes(CharsetUtil.UTF_8));
+        // Create decoder instance to test.
+        try {
+            new HttpPostRequestDecoder(inMemoryFactory, req);
+            fail("Was expecting an ErrorDataDecoderException");
+        } catch (HttpPostRequestDecoder.ErrorDataDecoderException e) {
+            assertTrue(e.getCause() instanceof UnsupportedCharsetException);
+        } finally {
+            req.release();
+        }
+    }
+
+    @Test
+    public void testFormEncodeIncorrect() throws Exception {
+        LastHttpContent content = new DefaultLastHttpContent(
+                Unpooled.copiedBuffer("project=netty&&project=netty", CharsetUtil.US_ASCII));
+        DefaultHttpRequest req = new DefaultHttpRequest(HttpVersion.HTTP_1_1, HttpMethod.POST, "/");
+        HttpPostRequestDecoder decoder = new HttpPostRequestDecoder(req);
+        try {
+            decoder.offer(content);
+            fail();
+        } catch (HttpPostRequestDecoder.ErrorDataDecoderException e) {
+            assertTrue(e.getCause() instanceof IllegalArgumentException);
+        } finally {
+            decoder.destroy();
+            content.release();
+        }
+    }
 }
