@@ -17,7 +17,6 @@ package io.netty.handler.codec.spdy;
 
 import io.netty.buffer.ByteBuf;
 import io.netty.buffer.Unpooled;
-import io.netty.util.ReferenceCountUtil;
 import org.junit.Before;
 import org.junit.Test;
 
@@ -143,7 +142,7 @@ public class SpdyFrameDecoderTest {
         byte flags = 0;
         int length = 1024;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeDataFrameHeader(buf, streamId, flags, length);
         for (int i = 0; i < 256; i ++) {
             buf.writeInt(RANDOM.nextInt());
@@ -153,6 +152,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -161,7 +161,7 @@ public class SpdyFrameDecoderTest {
         byte flags = 0;
         int length = 0;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeDataFrameHeader(buf, streamId, flags, length);
 
         delegate.readDataFrame(streamId, false, Unpooled.EMPTY_BUFFER);
@@ -169,6 +169,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -177,7 +178,7 @@ public class SpdyFrameDecoderTest {
         byte flags = 0x01; // FLAG_FIN
         int length = 0;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeDataFrameHeader(buf, streamId, flags, length);
 
         delegate.readDataFrame(streamId, true, Unpooled.EMPTY_BUFFER);
@@ -185,6 +186,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -193,7 +195,7 @@ public class SpdyFrameDecoderTest {
         byte flags = (byte) 0xFE; // should ignore any unknown flags
         int length = 0;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeDataFrameHeader(buf, streamId, flags, length);
 
         delegate.readDataFrame(streamId, false, Unpooled.EMPTY_BUFFER);
@@ -201,6 +203,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -209,7 +212,7 @@ public class SpdyFrameDecoderTest {
         byte flags = 0;
         int length = 0;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeDataFrameHeader(buf, streamId, flags, length);
 
         delegate.readFrameError((String) anyObject());
@@ -217,6 +220,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -226,7 +230,7 @@ public class SpdyFrameDecoderTest {
         byte flags = 0;
         int length = 0;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(2 * (SPDY_HEADER_SIZE + length)));
+        ByteBuf buf = Unpooled.buffer(2 * (SPDY_HEADER_SIZE + length));
         encodeDataFrameHeader(buf, streamId1, flags, length);
         encodeDataFrameHeader(buf, streamId2, flags, length);
 
@@ -236,6 +240,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -247,7 +252,7 @@ public class SpdyFrameDecoderTest {
         int associatedToStreamId = RANDOM.nextInt() & 0x7FFFFFFF;
         byte priority = (byte) (RANDOM.nextInt() & 0x07);
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(streamId);
         buf.writeInt(associatedToStreamId);
@@ -260,6 +265,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -271,7 +277,7 @@ public class SpdyFrameDecoderTest {
         int associatedToStreamId = RANDOM.nextInt() & 0x7FFFFFFF;
         byte priority = (byte) (RANDOM.nextInt() & 0x07);
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(streamId);
         buf.writeInt(associatedToStreamId);
@@ -284,6 +290,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -295,7 +302,7 @@ public class SpdyFrameDecoderTest {
         int associatedToStreamId = RANDOM.nextInt() & 0x7FFFFFFF;
         byte priority = (byte) (RANDOM.nextInt() & 0x07);
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(streamId);
         buf.writeInt(associatedToStreamId);
@@ -308,6 +315,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -319,7 +327,7 @@ public class SpdyFrameDecoderTest {
         int associatedToStreamId = 0; // independent of all other streams
         byte priority = (byte) (RANDOM.nextInt() & 0x07);
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(streamId);
         buf.writeInt(associatedToStreamId);
@@ -332,6 +340,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -343,7 +352,7 @@ public class SpdyFrameDecoderTest {
         int associatedToStreamId = RANDOM.nextInt() & 0x7FFFFFFF;
         byte priority = (byte) (RANDOM.nextInt() & 0x07);
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(streamId);
         buf.writeInt(associatedToStreamId);
@@ -356,6 +365,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -367,7 +377,7 @@ public class SpdyFrameDecoderTest {
         int associatedToStreamId = RANDOM.nextInt() & 0x7FFFFFFF;
         byte priority = (byte) (RANDOM.nextInt() & 0x07);
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(streamId | 0x80000000); // should ignore reserved bit
         buf.writeInt(associatedToStreamId | 0x80000000); // should ignore reserved bit
@@ -380,6 +390,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -390,7 +401,7 @@ public class SpdyFrameDecoderTest {
         int streamId = RANDOM.nextInt() & 0x7FFFFFFF | 0x01;
         int associatedToStreamId = RANDOM.nextInt() & 0x7FFFFFFF;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(streamId);
         buf.writeInt(associatedToStreamId);
@@ -400,6 +411,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -411,7 +423,7 @@ public class SpdyFrameDecoderTest {
         int associatedToStreamId = RANDOM.nextInt() & 0x7FFFFFFF;
         byte priority = (byte) (RANDOM.nextInt() & 0x07);
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(streamId);
         buf.writeInt(associatedToStreamId);
@@ -423,6 +435,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -435,14 +448,14 @@ public class SpdyFrameDecoderTest {
         int associatedToStreamId = RANDOM.nextInt() & 0x7FFFFFFF;
         byte priority = (byte) (RANDOM.nextInt() & 0x07);
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length + headerBlockLength));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length + headerBlockLength);
         encodeControlFrameHeader(buf, type, flags, length + headerBlockLength);
         buf.writeInt(streamId);
         buf.writeInt(associatedToStreamId);
         buf.writeByte(priority << 5);
         buf.writeByte(0);
 
-        ByteBuf headerBlock = ReferenceCountUtil.releaseLater(Unpooled.buffer(headerBlockLength));
+        ByteBuf headerBlock = Unpooled.buffer(headerBlockLength);
         for (int i = 0; i < 256; i ++) {
             headerBlock.writeInt(RANDOM.nextInt());
         }
@@ -456,6 +469,8 @@ public class SpdyFrameDecoderTest {
         verify(delegate);
         assertFalse(buf.isReadable());
         assertFalse(headerBlock.isReadable());
+        buf.release();
+        headerBlock.release();
     }
 
     @Test
@@ -465,7 +480,7 @@ public class SpdyFrameDecoderTest {
         int length = 4;
         int streamId = RANDOM.nextInt() & 0x7FFFFFFF | 0x01;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(streamId);
 
@@ -475,6 +490,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -484,7 +500,7 @@ public class SpdyFrameDecoderTest {
         int length = 4;
         int streamId = RANDOM.nextInt() & 0x7FFFFFFF | 0x01;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(streamId);
 
@@ -494,6 +510,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -503,7 +520,7 @@ public class SpdyFrameDecoderTest {
         int length = 4;
         int streamId = RANDOM.nextInt() & 0x7FFFFFFF | 0x01;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(streamId);
 
@@ -513,6 +530,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -522,7 +540,7 @@ public class SpdyFrameDecoderTest {
         int length = 4;
         int streamId = RANDOM.nextInt() & 0x7FFFFFFF | 0x01;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(streamId | 0x80000000); // should ignore reserved bit
 
@@ -532,6 +550,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -540,7 +559,7 @@ public class SpdyFrameDecoderTest {
         byte flags = 0;
         int length = 0; // invalid length
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
 
         delegate.readFrameError((String) anyObject());
@@ -548,6 +567,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -557,7 +577,7 @@ public class SpdyFrameDecoderTest {
         int length = 4;
         int streamId = 0; // invalid stream identifier
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(streamId);
 
@@ -566,6 +586,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -576,7 +597,7 @@ public class SpdyFrameDecoderTest {
         int headerBlockLength = 1024;
         int streamId = RANDOM.nextInt() & 0x7FFFFFFF | 0x01;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length + headerBlockLength));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length + headerBlockLength);
         encodeControlFrameHeader(buf, type, flags, length + headerBlockLength);
         buf.writeInt(streamId);
 
@@ -594,6 +615,8 @@ public class SpdyFrameDecoderTest {
         verify(delegate);
         assertFalse(buf.isReadable());
         assertFalse(headerBlock.isReadable());
+        buf.release();
+        headerBlock.release();
     }
 
     @Test
@@ -604,7 +627,7 @@ public class SpdyFrameDecoderTest {
         int streamId = RANDOM.nextInt() & 0x7FFFFFFF | 0x01;
         int statusCode = RANDOM.nextInt() | 0x01;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(streamId);
         buf.writeInt(statusCode);
@@ -614,6 +637,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -624,7 +648,7 @@ public class SpdyFrameDecoderTest {
         int streamId = RANDOM.nextInt() & 0x7FFFFFFF | 0x01;
         int statusCode = RANDOM.nextInt() | 0x01;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(streamId | 0x80000000); // should ignore reserved bit
         buf.writeInt(statusCode);
@@ -634,6 +658,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -644,7 +669,7 @@ public class SpdyFrameDecoderTest {
         int streamId = RANDOM.nextInt() & 0x7FFFFFFF | 0x01;
         int statusCode = RANDOM.nextInt() | 0x01;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(streamId);
         buf.writeInt(statusCode);
@@ -654,6 +679,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -664,7 +690,7 @@ public class SpdyFrameDecoderTest {
         int streamId = RANDOM.nextInt() & 0x7FFFFFFF | 0x01;
         int statusCode = RANDOM.nextInt() | 0x01;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(streamId);
         buf.writeInt(statusCode);
@@ -674,6 +700,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -684,7 +711,7 @@ public class SpdyFrameDecoderTest {
         int streamId = 0; // invalid stream identifier
         int statusCode = RANDOM.nextInt() | 0x01;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(streamId);
         buf.writeInt(statusCode);
@@ -694,6 +721,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -704,7 +732,7 @@ public class SpdyFrameDecoderTest {
         int streamId = RANDOM.nextInt() & 0x7FFFFFFF | 0x01;
         int statusCode = 0; // invalid status code
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(streamId);
         buf.writeInt(statusCode);
@@ -714,6 +742,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -726,7 +755,7 @@ public class SpdyFrameDecoderTest {
         int id = RANDOM.nextInt() & 0x00FFFFFF;
         int value = RANDOM.nextInt();
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(numSettings);
         for (int i = 0; i < numSettings; i++) {
@@ -743,6 +772,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -752,7 +782,7 @@ public class SpdyFrameDecoderTest {
         int numSettings = 0;
         int length = 8 * numSettings + 4;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(numSettings);
 
@@ -762,6 +792,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -771,7 +802,7 @@ public class SpdyFrameDecoderTest {
         int numSettings = 0;
         int length = 8 * numSettings + 4;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(numSettings);
 
@@ -781,6 +812,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -793,7 +825,7 @@ public class SpdyFrameDecoderTest {
         int id = RANDOM.nextInt() & 0x00FFFFFF;
         int value = RANDOM.nextInt();
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(numSettings);
         for (int i = 0; i < numSettings; i++) {
@@ -810,6 +842,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -822,7 +855,7 @@ public class SpdyFrameDecoderTest {
         int id = RANDOM.nextInt() & 0x00FFFFFF;
         int value = RANDOM.nextInt();
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(numSettings);
         for (int i = 0; i < numSettings; i++) {
@@ -839,6 +872,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -848,7 +882,7 @@ public class SpdyFrameDecoderTest {
         int numSettings = 0;
         int length = 8 * numSettings + 4;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(numSettings);
 
@@ -858,6 +892,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -870,7 +905,7 @@ public class SpdyFrameDecoderTest {
         int id = RANDOM.nextInt() & 0x00FFFFFF;
         int value = RANDOM.nextInt();
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(numSettings);
         for (int i = 0; i < numSettings; i++) {
@@ -887,6 +922,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -899,7 +935,7 @@ public class SpdyFrameDecoderTest {
         int id = RANDOM.nextInt() & 0x00FFFFFF;
         int value = RANDOM.nextInt();
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(numSettings);
         for (int i = 0; i < numSettings; i++) {
@@ -913,6 +949,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -925,7 +962,7 @@ public class SpdyFrameDecoderTest {
         int id = RANDOM.nextInt() & 0x00FFFFFF;
         int value = RANDOM.nextInt();
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(0); // invalid num_settings
         for (int i = 0; i < numSettings; i++) {
@@ -939,6 +976,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -947,7 +985,7 @@ public class SpdyFrameDecoderTest {
         byte flags = (byte) 0xFF;
         int length = 8;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeLong(RANDOM.nextLong());
 
@@ -955,6 +993,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -963,13 +1002,14 @@ public class SpdyFrameDecoderTest {
         byte flags = (byte) 0xFF;
         int length = 0;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
 
         replay(delegate);
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -979,7 +1019,7 @@ public class SpdyFrameDecoderTest {
         int segment = 4;
         int length = 2 * segment;
 
-        ByteBuf header = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE));
+        ByteBuf header = Unpooled.buffer(SPDY_HEADER_SIZE);
         ByteBuf segment1 = Unpooled.buffer(segment);
         ByteBuf segment2 = Unpooled.buffer(segment);
         encodeControlFrameHeader(header, type, flags, length);
@@ -994,6 +1034,9 @@ public class SpdyFrameDecoderTest {
         assertFalse(header.isReadable());
         assertFalse(segment1.isReadable());
         assertFalse(segment2.isReadable());
+        header.release();
+        segment1.release();
+        segment2.release();
     }
 
     @Test
@@ -1003,7 +1046,7 @@ public class SpdyFrameDecoderTest {
         int length = 4;
         int id = RANDOM.nextInt();
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(id);
 
@@ -1012,6 +1055,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -1021,7 +1065,7 @@ public class SpdyFrameDecoderTest {
         int length = 4;
         int id = RANDOM.nextInt();
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(id);
 
@@ -1030,6 +1074,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -1039,7 +1084,7 @@ public class SpdyFrameDecoderTest {
         int length = 8; // invalid length
         int id = RANDOM.nextInt();
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(id);
 
@@ -1048,6 +1093,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -1058,7 +1104,7 @@ public class SpdyFrameDecoderTest {
         int lastGoodStreamId = RANDOM.nextInt() & 0x7FFFFFFF;
         int statusCode = RANDOM.nextInt() | 0x01;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(lastGoodStreamId);
         buf.writeInt(statusCode);
@@ -1068,6 +1114,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -1078,7 +1125,7 @@ public class SpdyFrameDecoderTest {
         int lastGoodStreamId = RANDOM.nextInt() & 0x7FFFFFFF;
         int statusCode = RANDOM.nextInt() | 0x01;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(lastGoodStreamId);
         buf.writeInt(statusCode);
@@ -1088,6 +1135,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -1098,7 +1146,7 @@ public class SpdyFrameDecoderTest {
         int lastGoodStreamId = RANDOM.nextInt() & 0x7FFFFFFF;
         int statusCode = RANDOM.nextInt() | 0x01;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(lastGoodStreamId | 0x80000000); // should ignore reserved bit
         buf.writeInt(statusCode);
@@ -1108,6 +1156,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -1118,7 +1167,7 @@ public class SpdyFrameDecoderTest {
         int lastGoodStreamId = RANDOM.nextInt() & 0x7FFFFFFF;
         int statusCode = RANDOM.nextInt() | 0x01;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(lastGoodStreamId);
         buf.writeInt(statusCode);
@@ -1128,6 +1177,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -1137,7 +1187,7 @@ public class SpdyFrameDecoderTest {
         int length = 4;
         int streamId = RANDOM.nextInt() & 0x7FFFFFFF | 0x01;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(streamId);
 
@@ -1147,6 +1197,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -1156,7 +1207,7 @@ public class SpdyFrameDecoderTest {
         int length = 4;
         int streamId = RANDOM.nextInt() & 0x7FFFFFFF | 0x01;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(streamId);
 
@@ -1166,6 +1217,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -1175,7 +1227,7 @@ public class SpdyFrameDecoderTest {
         int length = 4;
         int streamId = RANDOM.nextInt() & 0x7FFFFFFF | 0x01;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(streamId);
 
@@ -1185,6 +1237,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -1194,7 +1247,7 @@ public class SpdyFrameDecoderTest {
         int length = 4;
         int streamId = RANDOM.nextInt() & 0x7FFFFFFF | 0x01;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(streamId | 0x80000000); // should ignore reserved bit
 
@@ -1204,6 +1257,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -1212,7 +1266,7 @@ public class SpdyFrameDecoderTest {
         byte flags = 0;
         int length = 0; // invalid length
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
 
         delegate.readFrameError((String) anyObject());
@@ -1220,6 +1274,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -1229,7 +1284,7 @@ public class SpdyFrameDecoderTest {
         int length = 4;
         int streamId = 0; // invalid stream identifier
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(streamId);
 
@@ -1238,6 +1293,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -1248,11 +1304,11 @@ public class SpdyFrameDecoderTest {
         int headerBlockLength = 1024;
         int streamId = RANDOM.nextInt() & 0x7FFFFFFF | 0x01;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length + headerBlockLength);
         buf.writeInt(streamId);
 
-        ByteBuf headerBlock = ReferenceCountUtil.releaseLater(Unpooled.buffer(headerBlockLength));
+        ByteBuf headerBlock = Unpooled.buffer(headerBlockLength);
         for (int i = 0; i < 256; i ++) {
             headerBlock.writeInt(RANDOM.nextInt());
         }
@@ -1266,6 +1322,8 @@ public class SpdyFrameDecoderTest {
         verify(delegate);
         assertFalse(buf.isReadable());
         assertFalse(headerBlock.isReadable());
+        buf.release();
+        headerBlock.release();
     }
 
     @Test
@@ -1296,7 +1354,7 @@ public class SpdyFrameDecoderTest {
         int streamId = RANDOM.nextInt() & 0x7FFFFFFF;
         int deltaWindowSize = RANDOM.nextInt() & 0x7FFFFFFF | 0x01;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(streamId);
         buf.writeInt(deltaWindowSize);
@@ -1306,6 +1364,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -1316,7 +1375,7 @@ public class SpdyFrameDecoderTest {
         int streamId = RANDOM.nextInt() & 0x7FFFFFFF;
         int deltaWindowSize = RANDOM.nextInt() & 0x7FFFFFFF | 0x01;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(streamId | 0x80000000); // should ignore reserved bit
         buf.writeInt(deltaWindowSize | 0x80000000); // should ignore reserved bit
@@ -1326,6 +1385,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -1336,7 +1396,7 @@ public class SpdyFrameDecoderTest {
         int streamId = RANDOM.nextInt() & 0x7FFFFFFF;
         int deltaWindowSize = RANDOM.nextInt() & 0x7FFFFFFF | 0x01;
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(streamId);
         buf.writeInt(deltaWindowSize);
@@ -1346,6 +1406,7 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 
     @Test
@@ -1356,7 +1417,7 @@ public class SpdyFrameDecoderTest {
         int streamId = RANDOM.nextInt() & 0x7FFFFFFF;
         int deltaWindowSize = 0; // invalid delta window size
 
-        ByteBuf buf = ReferenceCountUtil.releaseLater(Unpooled.buffer(SPDY_HEADER_SIZE + length));
+        ByteBuf buf = Unpooled.buffer(SPDY_HEADER_SIZE + length);
         encodeControlFrameHeader(buf, type, flags, length);
         buf.writeInt(streamId);
         buf.writeInt(deltaWindowSize);
@@ -1366,5 +1427,6 @@ public class SpdyFrameDecoderTest {
         decoder.decode(buf);
         verify(delegate);
         assertFalse(buf.isReadable());
+        buf.release();
     }
 }
