@@ -27,15 +27,6 @@ import java.util.List;
  * Before returning SocksRequest decoder removes itself from pipeline.
  */
 public class SocksCmdRequestDecoder extends ReplayingDecoder<State> {
-    private static final String name = "SOCKS_CMD_REQUEST_DECODER";
-
-    /**
-     * @deprecated Will be removed at the next minor version bump.
-     */
-    @Deprecated
-    public static String getName() {
-        return name;
-    }
 
     private SocksProtocolVersion version;
     private int fieldLength;
