@@ -17,9 +17,9 @@ package io.netty.handler.codec;
 
 import io.netty.buffer.ByteBuf;
 import io.netty.buffer.ByteBufAllocator;
-import io.netty.buffer.ByteBufProcessor;
 import io.netty.buffer.SwappedByteBuf;
 import io.netty.buffer.Unpooled;
+import io.netty.util.ByteProcessor;
 import io.netty.util.Signal;
 import io.netty.util.internal.StringUtil;
 
@@ -27,6 +27,7 @@ import java.io.InputStream;
 import java.io.OutputStream;
 import java.nio.ByteBuffer;
 import java.nio.ByteOrder;
+import java.nio.channels.FileChannel;
 import java.nio.channels.GatheringByteChannel;
 import java.nio.channels.ScatteringByteChannel;
 import java.nio.charset.Charset;
@@ -73,8 +74,7 @@ final class ReplayingDecoderByteBuf extends ByteBuf {
 
     @Override
     public ByteBuf capacity(int newCapacity) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
@@ -87,6 +87,17 @@ final class ReplayingDecoderByteBuf extends ByteBuf {
         return buffer.alloc();
     }
 
+    @Override
+    public boolean isReadOnly() {
+        return false;
+    }
+
+    @SuppressWarnings("deprecation")
+    @Override
+    public ByteBuf asReadOnly() {
+        return Unpooled.unmodifiableBuffer(this);
+    }
+
     @Override
     public boolean isDirect() {
         return buffer.isDirect();
@@ -119,8 +130,7 @@ final class ReplayingDecoderByteBuf extends ByteBuf {
 
     @Override
     public ByteBuf clear() {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
@@ -130,14 +140,12 @@ final class ReplayingDecoderByteBuf extends ByteBuf {
 
     @Override
     public int compareTo(ByteBuf buffer) {
-        reject();
-        return 0;
+        throw reject();
     }
 
     @Override
     public ByteBuf copy() {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
@@ -148,26 +156,27 @@ final class ReplayingDecoderByteBuf extends ByteBuf {
 
     @Override
     public ByteBuf discardReadBytes() {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
     public ByteBuf ensureWritable(int writableBytes) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
     public int ensureWritable(int minWritableBytes, boolean force) {
-        reject();
-        return 0;
+        throw reject();
     }
 
     @Override
     public ByteBuf duplicate() {
-        reject();
-        return this;
+        throw reject();
+    }
+
+    @Override
+    public ByteBuf retainedDuplicate() {
+        throw reject();
     }
 
     @Override
@@ -204,8 +213,7 @@ final class ReplayingDecoderByteBuf extends ByteBuf {
 
     @Override
     public ByteBuf getBytes(int index, ByteBuffer dst) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
@@ -217,26 +225,27 @@ final class ReplayingDecoderByteBuf extends ByteBuf {
 
     @Override
     public ByteBuf getBytes(int index, ByteBuf dst, int length) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
     public ByteBuf getBytes(int index, ByteBuf dst) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
     public int getBytes(int index, GatheringByteChannel out, int length) {
-        reject();
-        return 0;
+        throw reject();
+    }
+
+    @Override
+    public int getBytes(int index, FileChannel out, long position, int length) {
+        throw reject();
     }
 
     @Override
     public ByteBuf getBytes(int index, OutputStream out, int length) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
@@ -245,42 +254,84 @@ final class ReplayingDecoderByteBuf extends ByteBuf {
         return buffer.getInt(index);
     }
 
+    @Override
+    public int getIntLE(int index) {
+        checkIndex(index, 4);
+        return buffer.getIntLE(index);
+    }
+
     @Override
     public long getUnsignedInt(int index) {
         checkIndex(index, 4);
         return buffer.getUnsignedInt(index);
     }
 
+    @Override
+    public long getUnsignedIntLE(int index) {
+        checkIndex(index, 4);
+        return buffer.getUnsignedIntLE(index);
+    }
+
     @Override
     public long getLong(int index) {
         checkIndex(index, 8);
         return buffer.getLong(index);
     }
 
+    @Override
+    public long getLongLE(int index) {
+        checkIndex(index, 8);
+        return buffer.getLongLE(index);
+    }
+
     @Override
     public int getMedium(int index) {
         checkIndex(index, 3);
         return buffer.getMedium(index);
     }
 
+    @Override
+    public int getMediumLE(int index) {
+        checkIndex(index, 3);
+        return buffer.getMediumLE(index);
+    }
+
     @Override
     public int getUnsignedMedium(int index) {
         checkIndex(index, 3);
         return buffer.getUnsignedMedium(index);
     }
 
+    @Override
+    public int getUnsignedMediumLE(int index) {
+        checkIndex(index, 3);
+        return buffer.getUnsignedMediumLE(index);
+    }
+
     @Override
     public short getShort(int index) {
         checkIndex(index, 2);
         return buffer.getShort(index);
     }
 
+    @Override
+    public short getShortLE(int index) {
+        checkIndex(index, 2);
+        return buffer.getShortLE(index);
+    }
+
     @Override
     public int getUnsignedShort(int index) {
         checkIndex(index, 2);
         return buffer.getUnsignedShort(index);
     }
 
+    @Override
+    public int getUnsignedShortLE(int index) {
+        checkIndex(index, 2);
+        return buffer.getUnsignedShortLE(index);
+    }
+
     @Override
     public char getChar(int index) {
         checkIndex(index, 2);
@@ -299,10 +350,15 @@ final class ReplayingDecoderByteBuf extends ByteBuf {
         return buffer.getDouble(index);
     }
 
+    @Override
+    public CharSequence getCharSequence(int index, int length, Charset charset) {
+        checkIndex(index, length);
+        return buffer.getCharSequence(index, length, charset);
+    }
+
     @Override
     public int hashCode() {
-        reject();
-        return 0;
+        throw reject();
     }
 
     @Override
@@ -329,8 +385,7 @@ final class ReplayingDecoderByteBuf extends ByteBuf {
 
     @Override
     public int bytesBefore(int length, byte value) {
-        final int readerIndex = buffer.readerIndex();
-        return bytesBefore(readerIndex, buffer.writerIndex() - readerIndex, value);
+        return bytesBefore(buffer.readerIndex(), length, value);
     }
 
     @Override
@@ -353,7 +408,7 @@ final class ReplayingDecoderByteBuf extends ByteBuf {
     }
 
     @Override
-    public int forEachByte(ByteBufProcessor processor) {
+    public int forEachByte(ByteProcessor processor) {
         int ret = buffer.forEachByte(processor);
         if (ret < 0) {
             throw REPLAY;
@@ -363,7 +418,7 @@ final class ReplayingDecoderByteBuf extends ByteBuf {
     }
 
     @Override
-    public int forEachByte(int index, int length, ByteBufProcessor processor) {
+    public int forEachByte(int index, int length, ByteProcessor processor) {
         final int writerIndex = buffer.writerIndex();
         if (index >= writerIndex) {
             throw REPLAY;
@@ -382,17 +437,16 @@ final class ReplayingDecoderByteBuf extends ByteBuf {
     }
 
     @Override
-    public int forEachByteDesc(ByteBufProcessor processor) {
+    public int forEachByteDesc(ByteProcessor processor) {
         if (terminated) {
             return buffer.forEachByteDesc(processor);
         } else {
-            reject();
-            return 0;
+            throw reject();
         }
     }
 
     @Override
-    public int forEachByteDesc(int index, int length, ByteBufProcessor processor) {
+    public int forEachByteDesc(int index, int length, ByteProcessor processor) {
         if (index + length > buffer.writerIndex()) {
             throw REPLAY;
         }
@@ -408,8 +462,7 @@ final class ReplayingDecoderByteBuf extends ByteBuf {
 
     @Override
     public ByteBuf markWriterIndex() {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
@@ -486,8 +539,7 @@ final class ReplayingDecoderByteBuf extends ByteBuf {
 
     @Override
     public ByteBuf readBytes(ByteBuffer dst) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
@@ -499,8 +551,7 @@ final class ReplayingDecoderByteBuf extends ByteBuf {
 
     @Override
     public ByteBuf readBytes(ByteBuf dst, int length) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
@@ -512,8 +563,12 @@ final class ReplayingDecoderByteBuf extends ByteBuf {
 
     @Override
     public int readBytes(GatheringByteChannel out, int length) {
-        reject();
-        return 0;
+        throw reject();
+    }
+
+    @Override
+    public int readBytes(FileChannel out, long position, int length) {
+        throw reject();
     }
 
     @Override
@@ -528,10 +583,15 @@ final class ReplayingDecoderByteBuf extends ByteBuf {
         return buffer.readSlice(length);
     }
 
+    @Override
+    public ByteBuf readRetainedSlice(int length) {
+        checkReadableBytes(length);
+        return buffer.readRetainedSlice(length);
+    }
+
     @Override
     public ByteBuf readBytes(OutputStream out, int length) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
@@ -551,42 +611,84 @@ final class ReplayingDecoderByteBuf extends ByteBuf {
         return buffer.readInt();
     }
 
+    @Override
+    public int readIntLE() {
+        checkReadableBytes(4);
+        return buffer.readIntLE();
+    }
+
     @Override
     public long readUnsignedInt() {
         checkReadableBytes(4);
         return buffer.readUnsignedInt();
     }
 
+    @Override
+    public long readUnsignedIntLE() {
+        checkReadableBytes(4);
+        return buffer.readUnsignedIntLE();
+    }
+
     @Override
     public long readLong() {
         checkReadableBytes(8);
         return buffer.readLong();
     }
 
+    @Override
+    public long readLongLE() {
+        checkReadableBytes(8);
+        return buffer.readLongLE();
+    }
+
     @Override
     public int readMedium() {
         checkReadableBytes(3);
         return buffer.readMedium();
     }
 
+    @Override
+    public int readMediumLE() {
+        checkReadableBytes(3);
+        return buffer.readMediumLE();
+    }
+
     @Override
     public int readUnsignedMedium() {
         checkReadableBytes(3);
         return buffer.readUnsignedMedium();
     }
 
+    @Override
+    public int readUnsignedMediumLE() {
+        checkReadableBytes(3);
+        return buffer.readUnsignedMediumLE();
+    }
+
     @Override
     public short readShort() {
         checkReadableBytes(2);
         return buffer.readShort();
     }
 
+    @Override
+    public short readShortLE() {
+        checkReadableBytes(2);
+        return buffer.readShortLE();
+    }
+
     @Override
     public int readUnsignedShort() {
         checkReadableBytes(2);
         return buffer.readUnsignedShort();
     }
 
+    @Override
+    public int readUnsignedShortLE() {
+        checkReadableBytes(2);
+        return buffer.readUnsignedShortLE();
+    }
+
     @Override
     public char readChar() {
         checkReadableBytes(2);
@@ -605,6 +707,12 @@ final class ReplayingDecoderByteBuf extends ByteBuf {
         return buffer.readDouble();
     }
 
+    @Override
+    public CharSequence readCharSequence(int length, Charset charset) {
+        checkReadableBytes(length);
+        return buffer.readCharSequence(length, charset);
+    }
+
     @Override
     public ByteBuf resetReaderIndex() {
         buffer.resetReaderIndex();
@@ -613,122 +721,127 @@ final class ReplayingDecoderByteBuf extends ByteBuf {
 
     @Override
     public ByteBuf resetWriterIndex() {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
     public ByteBuf setBoolean(int index, boolean value) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
     public ByteBuf setByte(int index, int value) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
     public ByteBuf setBytes(int index, byte[] src, int srcIndex, int length) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
     public ByteBuf setBytes(int index, byte[] src) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
     public ByteBuf setBytes(int index, ByteBuffer src) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
     public ByteBuf setBytes(int index, ByteBuf src, int srcIndex, int length) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
     public ByteBuf setBytes(int index, ByteBuf src, int length) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
     public ByteBuf setBytes(int index, ByteBuf src) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
     public int setBytes(int index, InputStream in, int length) {
-        reject();
-        return 0;
+        throw reject();
     }
 
     @Override
     public ByteBuf setZero(int index, int length) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
     public int setBytes(int index, ScatteringByteChannel in, int length) {
-        reject();
-        return 0;
+        throw reject();
+    }
+
+    @Override
+    public int setBytes(int index, FileChannel in, long position, int length) {
+        throw reject();
     }
 
     @Override
     public ByteBuf setIndex(int readerIndex, int writerIndex) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
     public ByteBuf setInt(int index, int value) {
-        reject();
-        return this;
+        throw reject();
+    }
+
+    @Override
+    public ByteBuf setIntLE(int index, int value) {
+        throw reject();
     }
 
     @Override
     public ByteBuf setLong(int index, long value) {
-        reject();
-        return this;
+        throw reject();
+    }
+
+    @Override
+    public ByteBuf setLongLE(int index, long value) {
+        throw reject();
     }
 
     @Override
     public ByteBuf setMedium(int index, int value) {
-        reject();
-        return this;
+        throw reject();
+    }
+
+    @Override
+    public ByteBuf setMediumLE(int index, int value) {
+        throw reject();
     }
 
     @Override
     public ByteBuf setShort(int index, int value) {
-        reject();
-        return this;
+        throw reject();
+    }
+
+    @Override
+    public ByteBuf setShortLE(int index, int value) {
+        throw reject();
     }
 
     @Override
     public ByteBuf setChar(int index, int value) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
     public ByteBuf setFloat(int index, float value) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
     public ByteBuf setDouble(int index, double value) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
@@ -740,8 +853,12 @@ final class ReplayingDecoderByteBuf extends ByteBuf {
 
     @Override
     public ByteBuf slice() {
-        reject();
-        return this;
+        throw reject();
+    }
+
+    @Override
+    public ByteBuf retainedSlice() {
+        throw reject();
     }
 
     @Override
@@ -750,6 +867,12 @@ final class ReplayingDecoderByteBuf extends ByteBuf {
         return buffer.slice(index, length);
     }
 
+    @Override
+    public ByteBuf retainedSlice(int index, int length) {
+        checkIndex(index, length);
+        return buffer.slice(index, length);
+    }
+
     @Override
     public int nioBufferCount() {
         return buffer.nioBufferCount();
@@ -757,8 +880,7 @@ final class ReplayingDecoderByteBuf extends ByteBuf {
 
     @Override
     public ByteBuffer nioBuffer() {
-        reject();
-        return null;
+        throw reject();
     }
 
     @Override
@@ -769,8 +891,7 @@ final class ReplayingDecoderByteBuf extends ByteBuf {
 
     @Override
     public ByteBuffer[] nioBuffers() {
-        reject();
-        return null;
+        throw reject();
     }
 
     @Override
@@ -793,8 +914,7 @@ final class ReplayingDecoderByteBuf extends ByteBuf {
 
     @Override
     public String toString(Charset charsetName) {
-        reject();
-        return null;
+        throw reject();
     }
 
     @Override
@@ -830,86 +950,92 @@ final class ReplayingDecoderByteBuf extends ByteBuf {
 
     @Override
     public ByteBuf writeBoolean(boolean value) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
     public ByteBuf writeByte(int value) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
     public ByteBuf writeBytes(byte[] src, int srcIndex, int length) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
     public ByteBuf writeBytes(byte[] src) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
     public ByteBuf writeBytes(ByteBuffer src) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
     public ByteBuf writeBytes(ByteBuf src, int srcIndex, int length) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
     public ByteBuf writeBytes(ByteBuf src, int length) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
     public ByteBuf writeBytes(ByteBuf src) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
     public int writeBytes(InputStream in, int length) {
-        reject();
-        return 0;
+        throw reject();
     }
 
     @Override
     public int writeBytes(ScatteringByteChannel in, int length) {
-        reject();
-        return 0;
+        throw reject();
+    }
+
+    @Override
+    public int writeBytes(FileChannel in, long position, int length) {
+        throw reject();
     }
 
     @Override
     public ByteBuf writeInt(int value) {
-        reject();
-        return this;
+        throw reject();
+    }
+
+    @Override
+    public ByteBuf writeIntLE(int value) {
+        throw reject();
     }
 
     @Override
     public ByteBuf writeLong(long value) {
-        reject();
-        return this;
+        throw reject();
+    }
+
+    @Override
+    public ByteBuf writeLongLE(long value) {
+        throw reject();
     }
 
     @Override
     public ByteBuf writeMedium(int value) {
-        reject();
-        return this;
+        throw reject();
+    }
+
+    @Override
+    public ByteBuf writeMediumLE(int value) {
+        throw reject();
     }
 
     @Override
     public ByteBuf writeZero(int length) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
@@ -919,32 +1045,42 @@ final class ReplayingDecoderByteBuf extends ByteBuf {
 
     @Override
     public ByteBuf writerIndex(int writerIndex) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
     public ByteBuf writeShort(int value) {
-        reject();
-        return this;
+        throw reject();
+    }
+
+    @Override
+    public ByteBuf writeShortLE(int value) {
+        throw reject();
     }
 
     @Override
     public ByteBuf writeChar(int value) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
     public ByteBuf writeFloat(float value) {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
     public ByteBuf writeDouble(double value) {
-        reject();
-        return this;
+        throw reject();
+    }
+
+    @Override
+    public int setCharSequence(int index, CharSequence sequence, Charset charset) {
+        throw reject();
+    }
+
+    @Override
+    public int writeCharSequence(CharSequence sequence, Charset charset) {
+        throw reject();
     }
 
     private void checkIndex(int index, int length) {
@@ -961,8 +1097,7 @@ final class ReplayingDecoderByteBuf extends ByteBuf {
 
     @Override
     public ByteBuf discardSomeReadBytes() {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
@@ -972,35 +1107,42 @@ final class ReplayingDecoderByteBuf extends ByteBuf {
 
     @Override
     public ByteBuf retain() {
-        reject();
-        return this;
+        throw reject();
     }
 
     @Override
     public ByteBuf retain(int increment) {
-        reject();
+        throw reject();
+    }
+
+    @Override
+    public ByteBuf touch() {
+        buffer.touch();
+        return this;
+    }
+
+    @Override
+    public ByteBuf touch(Object hint) {
+        buffer.touch(hint);
         return this;
     }
 
     @Override
     public boolean release() {
-        reject();
-        return false;
+        throw reject();
     }
 
     @Override
     public boolean release(int decrement) {
-        reject();
-        return false;
+        throw reject();
     }
 
     @Override
     public ByteBuf unwrap() {
-        reject();
-        return this;
+        throw reject();
     }
 
-    private static void reject() {
-        throw new UnsupportedOperationException("not a replayable operation");
+    private static UnsupportedOperationException reject() {
+        return new UnsupportedOperationException("not a replayable operation");
     }
 }
