@@ -15,7 +15,6 @@
  */
 package io.netty.util;
 
-import io.netty.util.internal.MpscLinkedQueueNode;
 import io.netty.util.internal.PlatformDependent;
 import io.netty.util.internal.StringUtil;
 import io.netty.util.internal.logging.InternalLogger;
@@ -27,9 +26,11 @@ import java.util.Queue;
 import java.util.Set;
 import java.util.concurrent.CountDownLatch;
 import java.util.concurrent.Executors;
+import java.util.concurrent.RejectedExecutionException;
 import java.util.concurrent.ThreadFactory;
 import java.util.concurrent.TimeUnit;
 import java.util.concurrent.atomic.AtomicIntegerFieldUpdater;
+import java.util.concurrent.atomic.AtomicLong;
 
 /**
  * A {@link Timer} optimized for approximated I/O timeout scheduling.
@@ -77,21 +78,13 @@ public class HashedWheelTimer implements Timer {
     static final InternalLogger logger =
             InternalLoggerFactory.getInstance(HashedWheelTimer.class);
 
-    private static final ResourceLeakDetector<HashedWheelTimer> leakDetector =
-            new ResourceLeakDetector<HashedWheelTimer>(
-                    HashedWheelTimer.class, 1, Runtime.getRuntime().availableProcessors() * 4);
+    private static final ResourceLeakDetector<HashedWheelTimer> leakDetector = ResourceLeakDetectorFactory.instance()
+            .newResourceLeakDetector(HashedWheelTimer.class, 1, Runtime.getRuntime().availableProcessors() * 4L);
 
-    private static final AtomicIntegerFieldUpdater<HashedWheelTimer> WORKER_STATE_UPDATER;
-    static {
-        AtomicIntegerFieldUpdater<HashedWheelTimer> workerStateUpdater =
-                PlatformDependent.newAtomicIntegerFieldUpdater(HashedWheelTimer.class, "workerState");
-        if (workerStateUpdater == null) {
-            workerStateUpdater = AtomicIntegerFieldUpdater.newUpdater(HashedWheelTimer.class, "workerState");
-        }
-        WORKER_STATE_UPDATER = workerStateUpdater;
-    }
+    private static final AtomicIntegerFieldUpdater<HashedWheelTimer> WORKER_STATE_UPDATER =
+            AtomicIntegerFieldUpdater.newUpdater(HashedWheelTimer.class, "workerState");
 
-    private final ResourceLeak leak;
+    private final ResourceLeakTracker<HashedWheelTimer> leak;
     private final Worker worker = new Worker();
     private final Thread workerThread;
 
@@ -106,7 +99,9 @@ public class HashedWheelTimer implements Timer {
     private final int mask;
     private final CountDownLatch startTimeInitialized = new CountDownLatch(1);
     private final Queue<HashedWheelTimeout> timeouts = PlatformDependent.newMpscQueue();
-    private final Queue<Runnable> cancelledTimeouts = PlatformDependent.newMpscQueue();
+    private final Queue<HashedWheelTimeout> cancelledTimeouts = PlatformDependent.newMpscQueue();
+    private final AtomicLong pendingTimeouts = new AtomicLong(0);
+    private final long maxPendingTimeouts;
 
     private volatile long startTime;
 
@@ -127,7 +122,7 @@ public class HashedWheelTimer implements Timer {
      * @param tickDuration   the duration between tick
      * @param unit           the time unit of the {@code tickDuration}
      * @throws NullPointerException     if {@code unit} is {@code null}
-     * @throws IllegalArgumentException if {@code tickDuration} is <= 0
+     * @throws IllegalArgumentException if {@code tickDuration} is &lt;= 0
      */
     public HashedWheelTimer(long tickDuration, TimeUnit unit) {
         this(Executors.defaultThreadFactory(), tickDuration, unit);
@@ -141,7 +136,7 @@ public class HashedWheelTimer implements Timer {
      * @param unit           the time unit of the {@code tickDuration}
      * @param ticksPerWheel  the size of the wheel
      * @throws NullPointerException     if {@code unit} is {@code null}
-     * @throws IllegalArgumentException if either of {@code tickDuration} and {@code ticksPerWheel} is <= 0
+     * @throws IllegalArgumentException if either of {@code tickDuration} and {@code ticksPerWheel} is &lt;= 0
      */
     public HashedWheelTimer(long tickDuration, TimeUnit unit, int ticksPerWheel) {
         this(Executors.defaultThreadFactory(), tickDuration, unit, ticksPerWheel);
@@ -169,7 +164,7 @@ public class HashedWheelTimer implements Timer {
      * @param tickDuration   the duration between tick
      * @param unit           the time unit of the {@code tickDuration}
      * @throws NullPointerException     if either of {@code threadFactory} and {@code unit} is {@code null}
-     * @throws IllegalArgumentException if {@code tickDuration} is <= 0
+     * @throws IllegalArgumentException if {@code tickDuration} is &lt;= 0
      */
     public HashedWheelTimer(
             ThreadFactory threadFactory, long tickDuration, TimeUnit unit) {
@@ -186,11 +181,59 @@ public class HashedWheelTimer implements Timer {
      * @param unit           the time unit of the {@code tickDuration}
      * @param ticksPerWheel  the size of the wheel
      * @throws NullPointerException     if either of {@code threadFactory} and {@code unit} is {@code null}
-     * @throws IllegalArgumentException if either of {@code tickDuration} and {@code ticksPerWheel} is <= 0
+     * @throws IllegalArgumentException if either of {@code tickDuration} and {@code ticksPerWheel} is &lt;= 0
      */
     public HashedWheelTimer(
             ThreadFactory threadFactory,
             long tickDuration, TimeUnit unit, int ticksPerWheel) {
+        this(threadFactory, tickDuration, unit, ticksPerWheel, true);
+    }
+
+    /**
+     * Creates a new timer.
+     *
+     * @param threadFactory        a {@link ThreadFactory} that creates a
+     *                             background {@link Thread} which is dedicated to
+     *                             {@link TimerTask} execution.
+     * @param tickDuration         the duration between tick
+     * @param unit                 the time unit of the {@code tickDuration}
+     * @param ticksPerWheel        the size of the wheel
+     * @param leakDetection        {@code true} if leak detection should be enabled always,
+     *                             if false it will only be enabled if the worker thread is not
+     *                             a daemon thread.
+     * @throws NullPointerException     if either of {@code threadFactory} and {@code unit} is {@code null}
+     * @throws IllegalArgumentException if either of {@code tickDuration} and {@code ticksPerWheel} is &lt;= 0
+     */
+    public HashedWheelTimer(
+        ThreadFactory threadFactory,
+        long tickDuration, TimeUnit unit, int ticksPerWheel, boolean leakDetection) {
+        this(threadFactory, tickDuration, unit, ticksPerWheel, leakDetection, -1);
+    }
+
+    /**
+     * Creates a new timer.
+     *
+     * @param threadFactory        a {@link ThreadFactory} that creates a
+     *                             background {@link Thread} which is dedicated to
+     *                             {@link TimerTask} execution.
+     * @param tickDuration         the duration between tick
+     * @param unit                 the time unit of the {@code tickDuration}
+     * @param ticksPerWheel        the size of the wheel
+     * @param leakDetection        {@code true} if leak detection should be enabled always,
+     *                             if false it will only be enabled if the worker thread is not
+     *                             a daemon thread.
+     * @param  maxPendingTimeouts  The maximum number of pending timeouts after which call to
+     *                             {@code newTimeout} will result in
+     *                             {@link java.util.concurrent.RejectedExecutionException}
+     *                             being thrown. No maximum pending timeouts limit is assumed if
+     *                             this value is 0 or negative.
+     * @throws NullPointerException     if either of {@code threadFactory} and {@code unit} is {@code null}
+     * @throws IllegalArgumentException if either of {@code tickDuration} and {@code ticksPerWheel} is &lt;= 0
+     */
+    public HashedWheelTimer(
+            ThreadFactory threadFactory,
+            long tickDuration, TimeUnit unit, int ticksPerWheel, boolean leakDetection,
+            long maxPendingTimeouts) {
 
         if (threadFactory == null) {
             throw new NullPointerException("threadFactory");
@@ -220,7 +263,9 @@ public class HashedWheelTimer implements Timer {
         }
         workerThread = threadFactory.newThread(worker);
 
-        leak = leakDetector.open(this);
+        leak = leakDetection || !workerThread.isDaemon() ? leakDetector.track(this) : null;
+
+        this.maxPendingTimeouts = maxPendingTimeouts;
     }
 
     private static HashedWheelBucket[] createWheel(int ticksPerWheel) {
@@ -295,7 +340,8 @@ public class HashedWheelTimer implements Timer {
             WORKER_STATE_UPDATER.set(this, WORKER_STATE_SHUTDOWN);
 
             if (leak != null) {
-                leak.close();
+                boolean closed = leak.close(this);
+                assert closed;
             }
 
             return Collections.emptySet();
@@ -316,7 +362,8 @@ public class HashedWheelTimer implements Timer {
         }
 
         if (leak != null) {
-            leak.close();
+            boolean closed = leak.close(this);
+            assert closed;
         }
         return worker.unprocessedTimeouts();
     }
@@ -329,6 +376,16 @@ public class HashedWheelTimer implements Timer {
         if (unit == null) {
             throw new NullPointerException("unit");
         }
+        if (shouldLimitTimeouts()) {
+            long pendingTimeoutsCount = pendingTimeouts.incrementAndGet();
+            if (pendingTimeoutsCount > maxPendingTimeouts) {
+                pendingTimeouts.decrementAndGet();
+                throw new RejectedExecutionException("Number of pending timeouts ("
+                    + pendingTimeoutsCount + ") is greater than or equal to maximum allowed pending "
+                    + "timeouts (" + maxPendingTimeouts + ")");
+            }
+        }
+
         start();
 
         // Add the timeout to the timeout queue which will be processed on the next tick.
@@ -339,6 +396,10 @@ public class HashedWheelTimer implements Timer {
         return timeout;
     }
 
+    private boolean shouldLimitTimeouts() {
+        return maxPendingTimeouts > 0;
+    }
+
     private final class Worker implements Runnable {
         private final Set<Timeout> unprocessedTimeouts = new HashSet<Timeout>();
 
@@ -412,13 +473,13 @@ public class HashedWheelTimer implements Timer {
 
         private void processCancelledTasks() {
             for (;;) {
-                Runnable task = cancelledTimeouts.poll();
-                if (task == null) {
+                HashedWheelTimeout timeout = cancelledTimeouts.poll();
+                if (timeout == null) {
                     // all processed
                     break;
                 }
                 try {
-                    task.run();
+                    timeout.remove();
                 } catch (Throwable t) {
                     if (logger.isWarnEnabled()) {
                         logger.warn("An exception was thrown while process a cancellation task", t);
@@ -472,22 +533,13 @@ public class HashedWheelTimer implements Timer {
         }
     }
 
-    private static final class HashedWheelTimeout extends MpscLinkedQueueNode<Timeout>
-            implements Timeout {
+    private static final class HashedWheelTimeout implements Timeout {
 
         private static final int ST_INIT = 0;
         private static final int ST_CANCELLED = 1;
         private static final int ST_EXPIRED = 2;
-        private static final AtomicIntegerFieldUpdater<HashedWheelTimeout> STATE_UPDATER;
-
-        static {
-            AtomicIntegerFieldUpdater<HashedWheelTimeout> updater =
-                    PlatformDependent.newAtomicIntegerFieldUpdater(HashedWheelTimeout.class, "state");
-            if (updater == null) {
-                updater = AtomicIntegerFieldUpdater.newUpdater(HashedWheelTimeout.class, "state");
-            }
-            STATE_UPDATER = updater;
-        }
+        private static final AtomicIntegerFieldUpdater<HashedWheelTimeout> STATE_UPDATER =
+                AtomicIntegerFieldUpdater.newUpdater(HashedWheelTimeout.class, "state");
 
         private final HashedWheelTimer timer;
         private final TimerTask task;
@@ -530,25 +582,22 @@ public class HashedWheelTimer implements Timer {
             if (!compareAndSetState(ST_INIT, ST_CANCELLED)) {
                 return false;
             }
-            // If a task should be canceled we create a new Runnable for this to another queue which will
-            // be processed on each tick. So this means that we will have a GC latency of max. 1 tick duration
-            // which is good enough. This way we can make again use of our MpscLinkedQueue and so minimize the
-            // locking / overhead as much as possible.
-            //
-            // It is important that we not just add the HashedWheelTimeout itself again as it extends
-            // MpscLinkedQueueNode and so may still be used as tombstone.
-            timer.cancelledTimeouts.add(new Runnable() {
-                @Override
-                public void run() {
-                    HashedWheelBucket bucket = HashedWheelTimeout.this.bucket;
-                    if (bucket != null) {
-                        bucket.remove(HashedWheelTimeout.this);
-                    }
-                }
-            });
+            // If a task should be canceled we put this to another queue which will be processed on each tick.
+            // So this means that we will have a GC latency of max. 1 tick duration which is good enough. This way
+            // we can make again use of our MpscLinkedQueue and so minimize the locking / overhead as much as possible.
+            timer.cancelledTimeouts.add(this);
             return true;
         }
 
+        void remove() {
+            HashedWheelBucket bucket = this.bucket;
+            if (bucket != null) {
+                bucket.remove(this);
+            } else if (timer.shouldLimitTimeouts()) {
+                timer.pendingTimeouts.decrementAndGet();
+            }
+        }
+
         public boolean compareAndSetState(int expected, int state) {
             return STATE_UPDATER.compareAndSet(this, expected, state);
         }
@@ -567,11 +616,6 @@ public class HashedWheelTimer implements Timer {
             return state() == ST_EXPIRED;
         }
 
-        @Override
-        public HashedWheelTimeout value() {
-            return this;
-        }
-
         public void expire() {
             if (!compareAndSetState(ST_INIT, ST_EXPIRED)) {
                 return;
@@ -649,8 +693,9 @@ public class HashedWheelTimer implements Timer {
 
             // process all timeouts
             while (timeout != null) {
-                boolean remove = false;
+                HashedWheelTimeout next = timeout.next;
                 if (timeout.remainingRounds <= 0) {
+                    next = remove(timeout);
                     if (timeout.deadline <= deadline) {
                         timeout.expire();
                     } else {
@@ -658,22 +703,16 @@ public class HashedWheelTimer implements Timer {
                         throw new IllegalStateException(String.format(
                                 "timeout.deadline (%d) > deadline (%d)", timeout.deadline, deadline));
                     }
-                    remove = true;
                 } else if (timeout.isCancelled()) {
-                    remove = true;
+                    next = remove(timeout);
                 } else {
                     timeout.remainingRounds --;
                 }
-                // store reference to next as we may null out timeout.next in the remove block.
-                HashedWheelTimeout next = timeout.next;
-                if (remove) {
-                    remove(timeout);
-                }
                 timeout = next;
             }
         }
 
-        public void remove(HashedWheelTimeout timeout) {
+        public HashedWheelTimeout remove(HashedWheelTimeout timeout) {
             HashedWheelTimeout next = timeout.next;
             // remove timeout that was either processed or cancelled by updating the linked-list
             if (timeout.prev != null) {
@@ -699,6 +738,10 @@ public class HashedWheelTimer implements Timer {
             timeout.prev = null;
             timeout.next = null;
             timeout.bucket = null;
+            if (timeout.timer.shouldLimitTimeouts()) {
+                timeout.timer.pendingTimeouts.decrementAndGet();
+            }
+            return next;
         }
 
         /**
