@@ -25,6 +25,7 @@ import java.io.FileReader;
 import java.net.Inet4Address;
 import java.net.Inet6Address;
 import java.net.InetAddress;
+import java.net.InetSocketAddress;
 import java.net.NetworkInterface;
 import java.net.SocketException;
 import java.net.UnknownHostException;
@@ -117,23 +118,31 @@ public final class NetUtil {
     private static final int IPV4_SEPARATORS = 3;
 
     /**
-     * {@code true} if ipv4 should be used on a system that supports ipv4 and ipv6.
+     * {@code true} if IPv4 should be used even if the system supports both IPv4 and IPv6.
      */
     private static final boolean IPV4_PREFERRED = Boolean.getBoolean("java.net.preferIPv4Stack");
 
+    /**
+     * {@code true} if an IPv6 address should be preferred when a host has both an IPv4 address and an IPv6 address.
+     */
+    private static final boolean IPV6_ADDRESSES_PREFERRED = Boolean.getBoolean("java.net.preferIPv6Addresses");
+
     /**
      * The logger being used by this class
      */
     private static final InternalLogger logger = InternalLoggerFactory.getInstance(NetUtil.class);
 
     static {
+        logger.debug("-Djava.net.preferIPv4Stack: {}", IPV4_PREFERRED);
+        logger.debug("-Djava.net.preferIPv6Addresses: {}", IPV6_ADDRESSES_PREFERRED);
+
         byte[] LOCALHOST4_BYTES = {127, 0, 0, 1};
         byte[] LOCALHOST6_BYTES = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1};
 
         // Create IPv4 loopback address.
         Inet4Address localhost4 = null;
         try {
-            localhost4 = (Inet4Address) InetAddress.getByAddress(LOCALHOST4_BYTES);
+            localhost4 = (Inet4Address) InetAddress.getByAddress("localhost", LOCALHOST4_BYTES);
         } catch (Exception e) {
             // We should not get here as long as the length of the address is correct.
             PlatformDependent.throwException(e);
@@ -143,7 +152,7 @@ public final class NetUtil {
         // Create IPv6 loopback address.
         Inet6Address localhost6 = null;
         try {
-            localhost6 = (Inet6Address) InetAddress.getByAddress(LOCALHOST6_BYTES);
+            localhost6 = (Inet6Address) InetAddress.getByAddress("localhost", LOCALHOST6_BYTES);
         } catch (Exception e) {
             // We should not get here as long as the length of the address is correct.
             PlatformDependent.throwException(e);
@@ -244,28 +253,31 @@ public final class NetUtil {
                 // - Linux and Mac OS X: 128
                 int somaxconn = PlatformDependent.isWindows() ? 200 : 128;
                 File file = new File("/proc/sys/net/core/somaxconn");
-                if (file.exists()) {
-                    BufferedReader in = null;
-                    try {
+                BufferedReader in = null;
+                try {
+                    // file.exists() may throw a SecurityException if a SecurityManager is used, so execute it in the
+                    // try / catch block.
+                    // See https://github.com/netty/netty/issues/4936
+                    if (file.exists()) {
                         in = new BufferedReader(new FileReader(file));
                         somaxconn = Integer.parseInt(in.readLine());
                         if (logger.isDebugEnabled()) {
                             logger.debug("{}: {}", file, somaxconn);
                         }
-                    } catch (Exception e) {
-                        logger.debug("Failed to get SOMAXCONN from: {}", file, e);
-                    } finally {
-                        if (in != null) {
-                            try {
-                                in.close();
-                            } catch (Exception e) {
-                                // Ignored.
-                            }
+                    } else {
+                        if (logger.isDebugEnabled()) {
+                            logger.debug("{}: {} (non-existent)", file, somaxconn);
                         }
                     }
-                } else {
-                    if (logger.isDebugEnabled()) {
-                        logger.debug("{}: {} (non-existent)", file, somaxconn);
+                } catch (Exception e) {
+                    logger.debug("Failed to get SOMAXCONN from: {}", file, e);
+                } finally {
+                    if (in != null) {
+                        try {
+                            in.close();
+                        } catch (Exception e) {
+                            // Ignored.
+                        }
                     }
                 }
                 return somaxconn;
@@ -274,12 +286,27 @@ public final class NetUtil {
     }
 
     /**
-     * Returns {@code true} if ipv4 should be prefered on a system that supports ipv4 and ipv6.
+     * Returns {@code true} if IPv4 should be used even if the system supports both IPv4 and IPv6. Setting this
+     * property to {@code true} will disable IPv6 support. The default value of this property is {@code false}.
+     *
+     * @see <a href="https://docs.oracle.com/javase/8/docs/api/java/net/doc-files/net-properties.html">Java SE
+     *      networking properties</a>
      */
     public static boolean isIpV4StackPreferred() {
         return IPV4_PREFERRED;
     }
 
+    /**
+     * Returns {@code true} if an IPv6 address should be preferred when a host has both an IPv4 address and an IPv6
+     * address. The default value of this property is {@code false}.
+     *
+     * @see <a href="https://docs.oracle.com/javase/8/docs/api/java/net/doc-files/net-properties.html">Java SE
+     *      networking properties</a>
+     */
+    public static boolean isIpV6AddressesPreferred() {
+        return IPV6_ADDRESSES_PREFERRED;
+    }
+
     /**
      * Creates an byte[] based on an ipAddressString. No error handling is
      * performed here.
@@ -416,49 +443,98 @@ public final class NetUtil {
         ipByteArray[byteIndex + 1] |= charValue & 15;
     }
 
-    static int getIntValue(char c) {
-
+    private static int getIntValue(char c) {
         switch (c) {
-        case '0':
-            return 0;
-        case '1':
-            return 1;
-        case '2':
-            return 2;
-        case '3':
-            return 3;
-        case '4':
-            return 4;
-        case '5':
-            return 5;
-        case '6':
-            return 6;
-        case '7':
-            return 7;
-        case '8':
-            return 8;
-        case '9':
-            return 9;
+            case '0':
+                return 0;
+            case '1':
+                return 1;
+            case '2':
+                return 2;
+            case '3':
+                return 3;
+            case '4':
+                return 4;
+            case '5':
+                return 5;
+            case '6':
+                return 6;
+            case '7':
+                return 7;
+            case '8':
+                return 8;
+            case '9':
+                return 9;
         }
 
         c = Character.toLowerCase(c);
         switch (c) {
-        case 'a':
-            return 10;
-        case 'b':
-            return 11;
-        case 'c':
-            return 12;
-        case 'd':
-            return 13;
-        case 'e':
-            return 14;
-        case 'f':
-            return 15;
+            case 'a':
+                return 10;
+            case 'b':
+                return 11;
+            case 'c':
+                return 12;
+            case 'd':
+                return 13;
+            case 'e':
+                return 14;
+            case 'f':
+                return 15;
         }
         return 0;
     }
 
+    /**
+     * Converts a 32-bit integer into an IPv4 address.
+     */
+    public static String intToIpAddress(int i) {
+        StringBuilder buf = new StringBuilder(15);
+        buf.append(i >> 24 & 0xff);
+        buf.append('.');
+        buf.append(i >> 16 & 0xff);
+        buf.append('.');
+        buf.append(i >> 8 & 0xff);
+        buf.append('.');
+        buf.append(i & 0xff);
+        return buf.toString();
+    }
+
+    /**
+     * Converts 4-byte or 16-byte data into an IPv4 or IPv6 string respectively.
+     *
+     * @throws IllegalArgumentException
+     *         if {@code length} is not {@code 4} nor {@code 16}
+     */
+    public static String bytesToIpAddress(byte[] bytes) {
+        return bytesToIpAddress(bytes, 0, bytes.length);
+    }
+
+    /**
+     * Converts 4-byte or 16-byte data into an IPv4 or IPv6 string respectively.
+     *
+     * @throws IllegalArgumentException
+     *         if {@code length} is not {@code 4} nor {@code 16}
+     */
+    public static String bytesToIpAddress(byte[] bytes, int offset, int length) {
+        switch (length) {
+            case 4: {
+                return new StringBuilder(15)
+                        .append(bytes[offset] & 0xff)
+                        .append('.')
+                        .append(bytes[offset + 1] & 0xff)
+                        .append('.')
+                        .append(bytes[offset + 2] & 0xff)
+                        .append('.')
+                        .append(bytes[offset + 3] & 0xff).toString();
+            }
+            case 16:
+                return toAddressString(bytes, offset, false);
+            default:
+                throw new IllegalArgumentException("length: " + length + " (expected: 4 or 16)");
+        }
+    }
+
     public static boolean isValidIpV6Address(String ipAddress) {
         int length = ipAddress.length();
         boolean doubleColon = false;
@@ -494,59 +570,59 @@ public final class NetUtil {
             prevChar = c;
             c = ipAddress.charAt(i);
             switch (c) {
-            // case for the last 32-bits represented as IPv4 x:x:x:x:x:x:d.d.d.d
-            case '.':
-                numberOfPeriods ++;
-                if (numberOfPeriods > 3) {
-                    return false;
-                }
-                if (!isValidIp4Word(word.toString())) {
-                    return false;
-                }
-                if (numberOfColons != 6 && !doubleColon) {
-                    return false;
-                }
-                // a special case ::1:2:3:4:5:d.d.d.d allows 7 colons with an
-                // IPv4 ending, otherwise 7 :'s is bad
-                if (numberOfColons == 7 && ipAddress.charAt(startOffset) != ':' &&
-                    ipAddress.charAt(1 + startOffset) != ':') {
-                    return false;
-                }
-                word.delete(0, word.length());
-                break;
-
-            case ':':
-                // FIX "IP6 mechanism syntax #ip6-bad1"
-                // An IPV6 address cannot start with a single ":".
-                // Either it can starti with "::" or with a number.
-                if (i == startOffset && (ipAddress.length() <= i || ipAddress.charAt(i + 1) != ':')) {
-                    return false;
-                }
-                // END FIX "IP6 mechanism syntax #ip6-bad1"
-                numberOfColons ++;
-                if (numberOfColons > 7) {
-                    return false;
-                }
-                if (numberOfPeriods > 0) {
-                    return false;
-                }
-                if (prevChar == ':') {
-                    if (doubleColon) {
+                // case for the last 32-bits represented as IPv4 x:x:x:x:x:x:d.d.d.d
+                case '.':
+                    numberOfPeriods ++;
+                    if (numberOfPeriods > 3) {
                         return false;
                     }
-                    doubleColon = true;
-                }
-                word.delete(0, word.length());
-                break;
+                    if (!isValidIp4Word(word.toString())) {
+                        return false;
+                    }
+                    if (numberOfColons != 6 && !doubleColon) {
+                        return false;
+                    }
+                    // a special case ::1:2:3:4:5:d.d.d.d allows 7 colons with an
+                    // IPv4 ending, otherwise 7 :'s is bad
+                    if (numberOfColons == 7 && ipAddress.charAt(startOffset) != ':' &&
+                        ipAddress.charAt(1 + startOffset) != ':') {
+                        return false;
+                    }
+                    word.delete(0, word.length());
+                    break;
 
-            default:
-                if (word != null && word.length() > 3) {
-                    return false;
-                }
-                if (!isValidHexChar(c)) {
-                    return false;
-                }
-                word.append(c);
+                case ':':
+                    // FIX "IP6 mechanism syntax #ip6-bad1"
+                    // An IPV6 address cannot start with a single ":".
+                    // Either it can starti with "::" or with a number.
+                    if (i == startOffset && (ipAddress.length() <= i || ipAddress.charAt(i + 1) != ':')) {
+                        return false;
+                    }
+                    // END FIX "IP6 mechanism syntax #ip6-bad1"
+                    numberOfColons ++;
+                    if (numberOfColons > 7) {
+                        return false;
+                    }
+                    if (numberOfPeriods > 0) {
+                        return false;
+                    }
+                    if (prevChar == ':') {
+                        if (doubleColon) {
+                            return false;
+                        }
+                        doubleColon = true;
+                    }
+                    word.delete(0, word.length());
+                    break;
+
+                default:
+                    if (word != null && word.length() > 3) {
+                        return false;
+                    }
+                    if (!isValidHexChar(c)) {
+                        return false;
+                    }
+                    word.append(c);
             }
         }
 
@@ -575,11 +651,7 @@ public final class NetUtil {
         return true;
     }
 
-    /**
-     * @deprecated Do not use; published by mistake.
-     */
-    @Deprecated
-    public static boolean isValidIp4Word(String word) {
+    private static boolean isValidIp4Word(String word) {
         char c;
         if (word.length() < 1 || word.length() > 3) {
             return false;
@@ -693,8 +765,8 @@ public final class NetUtil {
             case ':':
                 ++ipv6Seperators;
                 if (i - begin > IPV6_MAX_CHAR_BETWEEN_SEPARATOR ||
-                    ipv4Seperators > 0 || ipv6Seperators > IPV6_MAX_SEPARATORS ||
-                    currentIndex + 1 >= bytes.length) {
+                        ipv4Seperators > 0 || ipv6Seperators > IPV6_MAX_SEPARATORS ||
+                        currentIndex + 1 >= bytes.length) {
                     return null;
                 }
                 value <<= (IPV6_MAX_CHAR_BETWEEN_SEPARATOR - (i - begin)) << 2;
@@ -726,17 +798,17 @@ public final class NetUtil {
             case '.':
                 ++ipv4Seperators;
                 if (i - begin > IPV4_MAX_CHAR_BETWEEN_SEPARATOR
-                    || ipv4Seperators > IPV4_SEPARATORS
-                    || (ipv6Seperators > 0 && (currentIndex + compressLength < 12))
-                    || i + 1 >= ipLength
-                    || currentIndex >= bytes.length
-                    || begin < 0
-                    || (begin == 0 && (i == 3 && (!isValidNumericChar(ip.charAt(2)) ||
-                                                  !isValidNumericChar(ip.charAt(1)) ||
-                                                  !isValidNumericChar(ip.charAt(0))) ||
-                                       i == 2 && (!isValidNumericChar(ip.charAt(1)) ||
-                                                  !isValidNumericChar(ip.charAt(0))) ||
-                                       i == 1 && !isValidNumericChar(ip.charAt(0))))) {
+                        || ipv4Seperators > IPV4_SEPARATORS
+                        || (ipv6Seperators > 0 && (currentIndex + compressLength < 12))
+                        || i + 1 >= ipLength
+                        || currentIndex >= bytes.length
+                        || begin < 0
+                        || (begin == 0 && (i == 3 && (!isValidNumericChar(ip.charAt(2)) ||
+                                                      !isValidNumericChar(ip.charAt(1)) ||
+                                                      !isValidNumericChar(ip.charAt(0))) ||
+                                           i == 2 && (!isValidNumericChar(ip.charAt(1)) ||
+                                                      !isValidNumericChar(ip.charAt(0))) ||
+                                           i == 1 && !isValidNumericChar(ip.charAt(0))))) {
                     return null;
                 }
                 value <<= (IPV4_MAX_CHAR_BETWEEN_SEPARATOR - (i - begin)) << 2;
@@ -774,17 +846,17 @@ public final class NetUtil {
         // Finish up last set of data that was accumulated in the loop (or before the loop)
         if (ipv4Seperators > 0) {
             if (begin > 0 && i - begin > IPV4_MAX_CHAR_BETWEEN_SEPARATOR ||
-                ipv4Seperators != IPV4_SEPARATORS ||
-                currentIndex >= bytes.length) {
+                    ipv4Seperators != IPV4_SEPARATORS ||
+                    currentIndex >= bytes.length) {
                 return null;
             }
             if (ipv6Seperators == 0) {
                 compressLength = 12;
             } else if (ipv6Seperators >= IPV6_MIN_SEPARATORS &&
-                       ip.charAt(ipLength - 1) != ':' &&
-                       (!isCompressed && (ipv6Seperators == 6 && ip.charAt(0) != ':') ||
-                        isCompressed && (ipv6Seperators + 1 < IPV6_MAX_SEPARATORS &&
-                                         (ip.charAt(0) != ':' || compressBegin <= 2)))) {
+                           ip.charAt(ipLength - 1) != ':' &&
+                           (!isCompressed && (ipv6Seperators == 6 && ip.charAt(0) != ':') ||
+                            isCompressed && (ipv6Seperators + 1 < IPV6_MAX_SEPARATORS &&
+                                             (ip.charAt(0) != ':' || compressBegin <= 2)))) {
                 compressLength -= 2;
             } else {
                 return null;
@@ -802,14 +874,14 @@ public final class NetUtil {
         } else {
             tmp = ipLength - 1;
             if (begin > 0 && i - begin > IPV6_MAX_CHAR_BETWEEN_SEPARATOR ||
-                ipv6Seperators < IPV6_MIN_SEPARATORS ||
-                !isCompressed && (ipv6Seperators + 1 != IPV6_MAX_SEPARATORS  ||
-                                  ip.charAt(0) == ':' || ip.charAt(tmp) == ':') ||
-                isCompressed && (ipv6Seperators > IPV6_MAX_SEPARATORS ||
-                                 (ipv6Seperators == IPV6_MAX_SEPARATORS &&
-                                  (compressBegin <= 2 && ip.charAt(0) != ':' ||
-                                   compressBegin >= 14 && ip.charAt(tmp) != ':'))) ||
-                currentIndex + 1 >= bytes.length) {
+                    ipv6Seperators < IPV6_MIN_SEPARATORS ||
+                    !isCompressed && (ipv6Seperators + 1 != IPV6_MAX_SEPARATORS  ||
+                                      ip.charAt(0) == ':' || ip.charAt(tmp) == ':') ||
+                    isCompressed && (ipv6Seperators > IPV6_MAX_SEPARATORS ||
+                        (ipv6Seperators == IPV6_MAX_SEPARATORS &&
+                          (compressBegin <= 2 && ip.charAt(0) != ':' ||
+                           compressBegin >= 14 && ip.charAt(tmp) != ':'))) ||
+                    currentIndex + 1 >= bytes.length) {
                 return null;
             }
             if (begin >= 0 && i - begin <= IPV6_MAX_CHAR_BETWEEN_SEPARATOR) {
@@ -850,8 +922,8 @@ public final class NetUtil {
         }
 
         if (ipv4Mapped && ipv4Seperators > 0 &&
-            bytes[0] == 0 && bytes[1] == 0 && bytes[2] == 0 && bytes[3] == 0 && bytes[4] == 0 &&
-            bytes[5] == 0 && bytes[6] == 0 && bytes[7] == 0 && bytes[8] == 0 && bytes[9] == 0) {
+                bytes[0] == 0 && bytes[1] == 0 && bytes[2] == 0 && bytes[3] == 0 && bytes[4] == 0 &&
+                bytes[5] == 0 && bytes[6] == 0 && bytes[7] == 0 && bytes[8] == 0 && bytes[9] == 0) {
             bytes[10] = bytes[11] = (byte) 0xff;
         }
 
@@ -862,6 +934,38 @@ public final class NetUtil {
         }
     }
 
+    /**
+     * Returns the {@link String} representation of an {@link InetSocketAddress}.
+     * <p>
+     * The output does not include Scope ID.
+     * @param addr {@link InetSocketAddress} to be converted to an address string
+     * @return {@code String} containing the text-formatted IP address
+     */
+    public static String toSocketAddressString(InetSocketAddress addr) {
+        String port = String.valueOf(addr.getPort());
+        final StringBuilder sb;
+
+        if (addr.isUnresolved()) {
+            String hostString = PlatformDependent.javaVersion() >= 7 ? addr.getHostString() : addr.getHostName();
+            sb = newSocketAddressStringBuilder(hostString, port, !isValidIpV6Address(hostString));
+        } else {
+            InetAddress address = addr.getAddress();
+            String hostString = toAddressString(address);
+            sb = newSocketAddressStringBuilder(hostString, port, address instanceof Inet4Address);
+        }
+        return sb.append(':').append(port).toString();
+    }
+
+    private static StringBuilder newSocketAddressStringBuilder(String hostString, String port, boolean ipv4) {
+        if (ipv4) {
+            // Need to include enough space for hostString:port.
+            return new StringBuilder(hostString.length() + 1 + port.length()).append(hostString);
+        }
+        // Need to include enough space for [hostString]:port.
+        return new StringBuilder(
+                hostString.length() + 3 + port.length()).append('[').append(hostString).append(']');
+    }
+
     /**
      * Returns the {@link String} representation of an {@link InetAddress}.
      * <ul>
@@ -907,13 +1011,17 @@ public final class NetUtil {
             return ip.getHostAddress();
         }
         if (!(ip instanceof Inet6Address)) {
-            throw new IllegalArgumentException("Unhandled type: " + ip.getClass());
+            throw new IllegalArgumentException("Unhandled type: " + ip);
         }
 
-        final byte[] bytes = ip.getAddress();
+        return toAddressString(ip.getAddress(), 0, ipv4Mapped);
+    }
+
+    private static String toAddressString(byte[] bytes, int offset, boolean ipv4Mapped) {
         final int[] words = new int[IPV6_WORD_COUNT];
         int i;
-        for (i = 0; i < words.length; ++i) {
+        final int end = offset + words.length;
+        for (i = offset; i < end; ++i) {
             words[i] = ((bytes[i << 1] & 0xff) << 8) | (bytes[(i << 1) + 1] & 0xff);
         }
 
