@@ -57,6 +57,11 @@ import java.util.concurrent.atomic.AtomicReferenceFieldUpdater;
 import java.util.regex.Matcher;
 import java.util.regex.Pattern;
 
+import static io.netty.util.internal.PlatformDependent0.HASH_CODE_ASCII_SEED;
+import static io.netty.util.internal.PlatformDependent0.HASH_CODE_C1;
+import static io.netty.util.internal.PlatformDependent0.HASH_CODE_C2;
+import static io.netty.util.internal.PlatformDependent0.hashCodeAsciiSanitize;
+import static io.netty.util.internal.PlatformDependent0.unalignedAccess;
 
 /**
  * Utility that detects various properties specific to the current runtime
@@ -73,6 +78,9 @@ public final class PlatformDependent {
     private static final Pattern MAX_DIRECT_MEMORY_SIZE_ARG_PATTERN = Pattern.compile(
             "\\s*-XX:MaxDirectMemorySize\\s*=\\s*([0-9]+)\\s*([kKmMgG]?)\\s*$");
 
+    // this must be initialized before any code below triggers initialization of PlatformDependent0
+    private static final boolean IS_EXPLICIT_NO_UNSAFE = explicitNoUnsafe0();
+
     private static final boolean IS_ANDROID = isAndroid0();
     private static final boolean IS_WINDOWS = isWindows0();
     private static volatile Boolean IS_ROOT;
@@ -81,7 +89,6 @@ public final class PlatformDependent {
 
     private static final boolean CAN_ENABLE_TCP_NODELAY_BY_DEFAULT = !isAndroid();
 
-    private static final boolean IS_EXPLICIT_NO_UNSAFE = explicitNoUnsafe0();
     private static final boolean HAS_UNSAFE = hasUnsafe0();
     private static final boolean CAN_USE_CHM_V8 = HAS_UNSAFE && JAVA_VERSION < 8;
     private static final boolean DIRECT_BUFFER_PREFERRED =
@@ -93,7 +100,7 @@ public final class PlatformDependent {
     private static final int DEFAULT_MAX_MPSC_CAPACITY =  MPSC_CHUNK_SIZE * MPSC_CHUNK_SIZE;
     private static final int MAX_ALLOWED_MPSC_CAPACITY = Pow2.MAX_POW2;
 
-    private static final long ARRAY_BASE_OFFSET = arrayBaseOffset0();
+    private static final long BYTE_ARRAY_BASE_OFFSET = byteArrayBaseOffset0();
 
     private static final boolean HAS_JAVASSIST = hasJavassist0();
 
@@ -416,6 +423,102 @@ public final class PlatformDependent {
         return PlatformDependent0.getLong(data, index);
     }
 
+    private static long getLongSafe(byte[] bytes, int offset) {
+        if (BIG_ENDIAN_NATIVE_ORDER) {
+            return (long) bytes[offset] << 56 |
+                    ((long) bytes[offset + 1] & 0xff) << 48 |
+                    ((long) bytes[offset + 2] & 0xff) << 40 |
+                    ((long) bytes[offset + 3] & 0xff) << 32 |
+                    ((long) bytes[offset + 4] & 0xff) << 24 |
+                    ((long) bytes[offset + 5] & 0xff) << 16 |
+                    ((long) bytes[offset + 6] & 0xff) <<  8 |
+                    (long) bytes[offset + 7] & 0xff;
+        }
+        return (long) bytes[offset] & 0xff |
+                ((long) bytes[offset + 1] & 0xff) << 8 |
+                ((long) bytes[offset + 2] & 0xff) << 16 |
+                ((long) bytes[offset + 3] & 0xff) << 24 |
+                ((long) bytes[offset + 4] & 0xff) << 32 |
+                ((long) bytes[offset + 5] & 0xff) << 40 |
+                ((long) bytes[offset + 6] & 0xff) << 48 |
+                (long) bytes[offset + 7] << 56;
+    }
+
+    private static int getIntSafe(byte[] bytes, int offset) {
+        if (BIG_ENDIAN_NATIVE_ORDER) {
+            return bytes[offset] << 24 |
+                    (bytes[offset + 1] & 0xff) << 16 |
+                    (bytes[offset + 2] & 0xff) << 8 |
+                    bytes[offset + 3] & 0xff;
+        }
+        return bytes[offset] & 0xff |
+                (bytes[offset + 1] & 0xff) << 8 |
+                (bytes[offset + 2] & 0xff) << 16 |
+                bytes[offset + 3] << 24;
+    }
+
+    private static short getShortSafe(byte[] bytes, int offset) {
+        if (BIG_ENDIAN_NATIVE_ORDER) {
+            return (short) (bytes[offset] << 8 | (bytes[offset + 1] & 0xff));
+        }
+        return (short) (bytes[offset] & 0xff | (bytes[offset + 1] << 8));
+    }
+
+    /**
+     * Identical to {@link PlatformDependent0#hashCodeAsciiCompute(long, int)} but for {@link CharSequence}.
+     */
+    private static int hashCodeAsciiCompute(CharSequence value, int offset, int hash) {
+        if (BIG_ENDIAN_NATIVE_ORDER) {
+            return hash * HASH_CODE_C1 +
+                    // Low order int
+                    hashCodeAsciiSanitizeInt(value, offset + 4) * HASH_CODE_C2 +
+                    // High order int
+                    hashCodeAsciiSanitizeInt(value, offset);
+        }
+        return hash * HASH_CODE_C1 +
+                // Low order int
+                hashCodeAsciiSanitizeInt(value, offset) * HASH_CODE_C2 +
+                // High order int
+                hashCodeAsciiSanitizeInt(value, offset + 4);
+    }
+
+    /**
+     * Identical to {@link PlatformDependent0#hashCodeAsciiSanitize(int)} but for {@link CharSequence}.
+     */
+    private static int hashCodeAsciiSanitizeInt(CharSequence value, int offset) {
+        if (BIG_ENDIAN_NATIVE_ORDER) {
+            // mimic a unsafe.getInt call on a big endian machine
+            return (value.charAt(offset + 3) & 0x1f) |
+                   (value.charAt(offset + 2) & 0x1f) << 8 |
+                   (value.charAt(offset + 1) & 0x1f) << 16 |
+                   (value.charAt(offset) & 0x1f) << 24;
+        }
+        return (value.charAt(offset + 3) & 0x1f) << 24 |
+               (value.charAt(offset + 2) & 0x1f) << 16 |
+               (value.charAt(offset + 1) & 0x1f) << 8 |
+               (value.charAt(offset) & 0x1f);
+    }
+
+    /**
+     * Identical to {@link PlatformDependent0#hashCodeAsciiSanitize(short)} but for {@link CharSequence}.
+     */
+    private static int hashCodeAsciiSanitizeShort(CharSequence value, int offset) {
+        if (BIG_ENDIAN_NATIVE_ORDER) {
+            // mimic a unsafe.getShort call on a big endian machine
+            return (value.charAt(offset + 1) & 0x1f) |
+                    (value.charAt(offset) & 0x1f) << 8;
+        }
+        return (value.charAt(offset + 1) & 0x1f) << 8 |
+                (value.charAt(offset) & 0x1f);
+    }
+
+    /**
+     * Identical to {@link PlatformDependent0#hashCodeAsciiSanitize(byte)} but for {@link CharSequence}.
+     */
+    private static int hashCodeAsciiSanitizsByte(char value) {
+        return value & 0x1f;
+    }
+
     public static void putOrderedObject(Object object, long address, Object value) {
         PlatformDependent0.putOrderedObject(object, address, value);
     }
@@ -457,15 +560,15 @@ public final class PlatformDependent {
     }
 
     public static void copyMemory(byte[] src, int srcIndex, long dstAddr, long length) {
-        PlatformDependent0.copyMemory(src, ARRAY_BASE_OFFSET + srcIndex, null, dstAddr, length);
+        PlatformDependent0.copyMemory(src, BYTE_ARRAY_BASE_OFFSET + srcIndex, null, dstAddr, length);
     }
 
     public static void copyMemory(long srcAddr, byte[] dst, int dstIndex, long length) {
-        PlatformDependent0.copyMemory(null, srcAddr, dst, ARRAY_BASE_OFFSET + dstIndex, length);
+        PlatformDependent0.copyMemory(null, srcAddr, dst, BYTE_ARRAY_BASE_OFFSET + dstIndex, length);
     }
 
     public static void setMemory(byte[] dst, int dstIndex, long bytes, byte value) {
-        PlatformDependent0.setMemory(dst, ARRAY_BASE_OFFSET + dstIndex, bytes, value);
+        PlatformDependent0.setMemory(dst, BYTE_ARRAY_BASE_OFFSET + dstIndex, bytes, value);
     }
 
     public static void setMemory(long address, long bytes, byte value) {
@@ -547,54 +650,152 @@ public final class PlatformDependent {
     }
 
     /**
-     * Create a new optimized {@link AtomicReferenceFieldUpdater} or {@code null} if it
-     * could not be created. Because of this the caller need to check for {@code null} and if {@code null} is returned
-     * use {@link AtomicReferenceFieldUpdater#newUpdater(Class, Class, String)} as fallback.
+     * Compare two {@code byte} arrays for equality. For performance reasons no bounds checking on the
+     * parameters is performed.
+     *
+     * @param bytes1 the first byte array.
+     * @param startPos1 the position (inclusive) to start comparing in {@code bytes1}.
+     * @param bytes2 the second byte array.
+     * @param startPos2 the position (inclusive) to start comparing in {@code bytes2}.
+     * @param length the amount of bytes to compare. This is assumed to be validated as not going out of bounds
+     * by the caller.
      */
-    public static <U, W> AtomicReferenceFieldUpdater<U, W> newAtomicReferenceFieldUpdater(
-            Class<? super U> tclass, String fieldName) {
-        if (hasUnsafe()) {
-            try {
-                return PlatformDependent0.newAtomicReferenceFieldUpdater(tclass, fieldName);
-            } catch (Throwable ignore) {
-                // ignore
-            }
-        }
-        return null;
+    public static boolean equals(byte[] bytes1, int startPos1, byte[] bytes2, int startPos2, int length) {
+        return !hasUnsafe() || !unalignedAccess() ?
+                  equalsSafe(bytes1, startPos1, bytes2, startPos2, length) :
+                  PlatformDependent0.equals(bytes1, startPos1, bytes2, startPos2, length);
     }
 
     /**
-     * Create a new optimized {@link AtomicIntegerFieldUpdater} or {@code null} if it
-     * could not be created. Because of this the caller need to check for {@code null} and if {@code null} is returned
-     * use {@link AtomicIntegerFieldUpdater#newUpdater(Class, String)} as fallback.
+     * Compare two {@code byte} arrays for equality without leaking timing information.
+     * For performance reasons no bounds checking on the parameters is performed.
+     * <p>
+     * The {@code int} return type is intentional and is designed to allow cascading of constant time operations:
+     * <pre>
+     *     byte[] s1 = new {1, 2, 3};
+     *     byte[] s2 = new {1, 2, 3};
+     *     byte[] s3 = new {1, 2, 3};
+     *     byte[] s4 = new {4, 5, 6};
+     *     boolean equals = (equalsConstantTime(s1, 0, s2, 0, s1.length) &
+     *                       equalsConstantTime(s3, 0, s4, 0, s3.length)) != 0;
+     * </pre>
+     * @param bytes1 the first byte array.
+     * @param startPos1 the position (inclusive) to start comparing in {@code bytes1}.
+     * @param bytes2 the second byte array.
+     * @param startPos2 the position (inclusive) to start comparing in {@code bytes2}.
+     * @param length the amount of bytes to compare. This is assumed to be validated as not going out of bounds
+     * by the caller.
+     * @return {@code 0} if not equal. {@code 1} if equal.
      */
-    public static <T> AtomicIntegerFieldUpdater<T> newAtomicIntegerFieldUpdater(
-            Class<? super T> tclass, String fieldName) {
-        if (hasUnsafe()) {
-            try {
-                return PlatformDependent0.newAtomicIntegerFieldUpdater(tclass, fieldName);
-            } catch (Throwable ignore) {
-                // ignore
-            }
-        }
-        return null;
+    public static int equalsConstantTime(byte[] bytes1, int startPos1, byte[] bytes2, int startPos2, int length) {
+        return !hasUnsafe() || !unalignedAccess() ?
+                  ConstantTimeUtils.equalsConstantTime(bytes1, startPos1, bytes2, startPos2, length) :
+                  PlatformDependent0.equalsConstantTime(bytes1, startPos1, bytes2, startPos2, length);
     }
 
     /**
-     * Create a new optimized {@link AtomicLongFieldUpdater} or {@code null} if it
-     * could not be created. Because of this the caller need to check for {@code null} and if {@code null} is returned
-     * use {@link AtomicLongFieldUpdater#newUpdater(Class, String)} as fallback.
+     * Calculate a hash code of a byte array assuming ASCII character encoding.
+     * The resulting hash code will be case insensitive.
+     * @param bytes The array which contains the data to hash.
+     * @param startPos What index to start generating a hash code in {@code bytes}
+     * @param length The amount of bytes that should be accounted for in the computation.
+     * @return The hash code of {@code bytes} assuming ASCII character encoding.
+     * The resulting hash code will be case insensitive.
      */
-    public static <T> AtomicLongFieldUpdater<T> newAtomicLongFieldUpdater(
-            Class<? super T> tclass, String fieldName) {
-        if (hasUnsafe()) {
-            try {
-                return PlatformDependent0.newAtomicLongFieldUpdater(tclass, fieldName);
-            } catch (Throwable ignore) {
-                // ignore
-            }
+    public static int hashCodeAscii(byte[] bytes, int startPos, int length) {
+        return !hasUnsafe() || !unalignedAccess() ?
+                hashCodeAsciiSafe(bytes, startPos, length) :
+                PlatformDependent0.hashCodeAscii(bytes, startPos, length);
+    }
+
+    /**
+     * Calculate a hash code of a byte array assuming ASCII character encoding.
+     * The resulting hash code will be case insensitive.
+     * <p>
+     * This method assumes that {@code bytes} is equivalent to a {@code byte[]} but just using {@link CharSequence}
+     * for storage. The upper most byte of each {@code char} from {@code bytes} is ignored.
+     * @param bytes The array which contains the data to hash (assumed to be equivalent to a {@code byte[]}).
+     * @return The hash code of {@code bytes} assuming ASCII character encoding.
+     * The resulting hash code will be case insensitive.
+     */
+    public static int hashCodeAscii(CharSequence bytes) {
+        int hash = HASH_CODE_ASCII_SEED;
+        final int remainingBytes = bytes.length() & 7;
+        // Benchmarking shows that by just naively looping for inputs 8~31 bytes long we incur a relatively large
+        // performance penalty (only achieve about 60% performance of loop which iterates over each char). So because
+        // of this we take special provisions to unroll the looping for these conditions.
+        switch (bytes.length()) {
+            case 31:
+            case 30:
+            case 29:
+            case 28:
+            case 27:
+            case 26:
+            case 25:
+            case 24:
+                hash = hashCodeAsciiCompute(bytes, bytes.length() - 24,
+                        hashCodeAsciiCompute(bytes, bytes.length() - 16,
+                          hashCodeAsciiCompute(bytes, bytes.length() - 8, hash)));
+                break;
+            case 23:
+            case 22:
+            case 21:
+            case 20:
+            case 19:
+            case 18:
+            case 17:
+            case 16:
+                hash = hashCodeAsciiCompute(bytes, bytes.length() - 16,
+                         hashCodeAsciiCompute(bytes, bytes.length() - 8, hash));
+                break;
+            case 15:
+            case 14:
+            case 13:
+            case 12:
+            case 11:
+            case 10:
+            case 9:
+            case 8:
+                hash = hashCodeAsciiCompute(bytes, bytes.length() - 8, hash);
+                break;
+            case 7:
+            case 6:
+            case 5:
+            case 4:
+            case 3:
+            case 2:
+            case 1:
+            case 0:
+                break;
+            default:
+                for (int i = bytes.length() - 8; i >= remainingBytes; i -= 8) {
+                    hash = hashCodeAsciiCompute(bytes, i, hash);
+                }
+                break;
+        }
+        switch(remainingBytes) {
+            case 7:
+                return ((hash * HASH_CODE_C1 + hashCodeAsciiSanitizsByte(bytes.charAt(0)))
+                              * HASH_CODE_C2 + hashCodeAsciiSanitizeShort(bytes, 1))
+                              * HASH_CODE_C1 + hashCodeAsciiSanitizeInt(bytes, 3);
+            case 6:
+                return (hash * HASH_CODE_C1 + hashCodeAsciiSanitizeShort(bytes, 0))
+                             * HASH_CODE_C2 + hashCodeAsciiSanitizeInt(bytes, 2);
+            case 5:
+                return (hash * HASH_CODE_C1 + hashCodeAsciiSanitizsByte(bytes.charAt(0)))
+                             * HASH_CODE_C2 + hashCodeAsciiSanitizeInt(bytes, 1);
+            case 4:
+                return hash * HASH_CODE_C1 + hashCodeAsciiSanitizeInt(bytes, 0);
+            case 3:
+                return (hash * HASH_CODE_C1 + hashCodeAsciiSanitizsByte(bytes.charAt(0)))
+                             * HASH_CODE_C2 + hashCodeAsciiSanitizeShort(bytes, 1);
+            case 2:
+                return hash * HASH_CODE_C1 + hashCodeAsciiSanitizeShort(bytes, 0);
+            case 1:
+                return hash * HASH_CODE_C1 + hashCodeAsciiSanitizsByte(bytes.charAt(0));
+            default:
+                return hash;
         }
-        return null;
     }
 
     private static final class Mpsc {
@@ -866,6 +1067,10 @@ public final class PlatformDependent {
         }
     }
 
+    static boolean isExplicitNoUnsafe() {
+        return IS_EXPLICIT_NO_UNSAFE;
+    }
+
     private static boolean explicitNoUnsafe0() {
         final boolean noUnsafe = SystemPropertyUtil.getBoolean("io.netty.noUnsafe", false);
         logger.debug("-Dio.netty.noUnsafe: {}", noUnsafe);
@@ -911,19 +1116,13 @@ public final class PlatformDependent {
         }
     }
 
-    private static long arrayBaseOffset0() {
-        if (!hasUnsafe()) {
-            return -1;
-        }
-
-        return PlatformDependent0.arrayBaseOffset();
-    }
-
     private static long maxDirectMemory0() {
         long maxDirectMemory = 0;
+        ClassLoader systemClassLoader = null;
         try {
             // Try to get from sun.misc.VM.maxDirectMemory() which should be most accurate.
-            Class<?> vmClass = Class.forName("sun.misc.VM", true, getSystemClassLoader());
+            systemClassLoader = getSystemClassLoader();
+            Class<?> vmClass = Class.forName("sun.misc.VM", true, systemClassLoader);
             Method m = vmClass.getDeclaredMethod("maxDirectMemory");
             maxDirectMemory = ((Number) m.invoke(null)).longValue();
         } catch (Throwable ignored) {
@@ -938,9 +1137,9 @@ public final class PlatformDependent {
             // Now try to get the JVM option (-XX:MaxDirectMemorySize) and parse it.
             // Note that we are using reflection because Android doesn't have these classes.
             Class<?> mgmtFactoryClass = Class.forName(
-                    "java.lang.management.ManagementFactory", true, getSystemClassLoader());
+                    "java.lang.management.ManagementFactory", true, systemClassLoader);
             Class<?> runtimeClass = Class.forName(
-                    "java.lang.management.RuntimeMXBean", true, getSystemClassLoader());
+                    "java.lang.management.RuntimeMXBean", true, systemClassLoader);
 
             Object runtime = mgmtFactoryClass.getDeclaredMethod("getRuntimeMXBean").invoke(null);
 
@@ -1136,7 +1335,61 @@ public final class PlatformDependent {
         return PlatformDependent0.addressSize();
     }
 
+    private static long byteArrayBaseOffset0() {
+        if (!hasUnsafe()) {
+            return -1;
+        }
+        return PlatformDependent0.byteArrayBaseOffset();
+    }
+
+    private static boolean equalsSafe(byte[] bytes1, int startPos1, byte[] bytes2, int startPos2, int length) {
+        final int end = startPos1 + length;
+        for (int i = startPos1, j = startPos2; i < end; ++i, ++j) {
+            if (bytes1[i] != bytes2[j]) {
+                return false;
+            }
+        }
+        return true;
+    }
+
+    /**
+     * Package private for testing purposes only!
+     */
+    static int hashCodeAsciiSafe(byte[] bytes, int startPos, int length) {
+        int hash = HASH_CODE_ASCII_SEED;
+        final int remainingBytes = length & 7;
+        final int end = startPos + remainingBytes;
+        for (int i = startPos - 8 + length; i >= end; i -= 8) {
+            hash = PlatformDependent0.hashCodeAsciiCompute(getLongSafe(bytes, i), hash);
+        }
+        switch(remainingBytes) {
+        case 7:
+            return ((hash * HASH_CODE_C1 + hashCodeAsciiSanitize(bytes[startPos]))
+                          * HASH_CODE_C2 + hashCodeAsciiSanitize(getShortSafe(bytes, startPos + 1)))
+                          * HASH_CODE_C1 + hashCodeAsciiSanitize(getIntSafe(bytes, startPos + 3));
+        case 6:
+            return (hash * HASH_CODE_C1 + hashCodeAsciiSanitize(getShortSafe(bytes, startPos)))
+                         * HASH_CODE_C2 + hashCodeAsciiSanitize(getIntSafe(bytes, startPos + 2));
+        case 5:
+            return (hash * HASH_CODE_C1 + hashCodeAsciiSanitize(bytes[startPos]))
+                         * HASH_CODE_C2 + hashCodeAsciiSanitize(getIntSafe(bytes, startPos + 1));
+        case 4:
+            return hash * HASH_CODE_C1 + hashCodeAsciiSanitize(getIntSafe(bytes, startPos));
+        case 3:
+            return (hash * HASH_CODE_C1 + hashCodeAsciiSanitize(bytes[startPos]))
+                         * HASH_CODE_C2 + hashCodeAsciiSanitize(getShortSafe(bytes, startPos + 1));
+        case 2:
+            return hash * HASH_CODE_C1 + hashCodeAsciiSanitize(getShortSafe(bytes, startPos));
+        case 1:
+            return hash * HASH_CODE_C1 + hashCodeAsciiSanitize(bytes[startPos]);
+        default:
+            return hash;
+        }
+    }
+
     private static final class AtomicLongCounter extends AtomicLong implements LongCounter {
+        private static final long serialVersionUID = 4074772784610639305L;
+
         @Override
         public void add(long delta) {
             addAndGet(delta);
