@@ -16,6 +16,7 @@
 package io.netty.util.internal;
 
 import java.io.IOException;
+import java.util.ArrayList;
 import java.util.List;
 
 import static io.netty.util.internal.ObjectUtil.*;
@@ -25,12 +26,15 @@ import static io.netty.util.internal.ObjectUtil.*;
  */
 public final class StringUtil {
 
+    public static final String EMPTY_STRING = "";
     public static final String NEWLINE = System.getProperty("line.separator");
+
     public static final char DOUBLE_QUOTE = '\"';
     public static final char COMMA = ',';
     public static final char LINE_FEED = '\n';
     public static final char CARRIAGE_RETURN = '\r';
-    public static final String EMPTY_STRING = "";
+    public static final char TAB = '\t';
+
     private static final String[] BYTE2HEX_PAD = new String[256];
     private static final String[] BYTE2HEX_NOPAD = new String[256];
 
@@ -317,6 +321,76 @@ public final class StringUtil {
     }
 
     /**
+     * Unescapes the specified escaped CSV fields according to
+     * <a href="https://tools.ietf.org/html/rfc4180#section-2">RFC-4180</a>.
+     *
+     * @param value A string with multiple CSV escaped fields which will be unescaped according to
+     *              <a href="https://tools.ietf.org/html/rfc4180#section-2">RFC-4180</a>
+     * @return {@link List} the list of unescaped fields
+     */
+    public static List<CharSequence> unescapeCsvFields(CharSequence value) {
+        List<CharSequence> unescaped = new ArrayList<CharSequence>(2);
+        StringBuilder current = InternalThreadLocalMap.get().stringBuilder();
+        boolean quoted = false;
+        int last = value.length() - 1;
+        for (int i = 0; i <= last; i++) {
+            char c = value.charAt(i);
+            if (quoted) {
+                switch (c) {
+                    case DOUBLE_QUOTE:
+                        if (i == last) {
+                            // Add the last field and return
+                            unescaped.add(current.toString());
+                            return unescaped;
+                        }
+                        char next = value.charAt(++i);
+                        if (next == DOUBLE_QUOTE) {
+                            // 2 double-quotes should be unescaped to one
+                            current.append(DOUBLE_QUOTE);
+                            break;
+                        }
+                        if (next == COMMA) {
+                            // This is the end of a field. Let's start to parse the next field.
+                            quoted = false;
+                            unescaped.add(current.toString());
+                            current.setLength(0);
+                            break;
+                        }
+                        // double-quote followed by other character is invalid
+                        throw newInvalidEscapedCsvFieldException(value, i - 1);
+                    default:
+                        current.append(c);
+                }
+            } else {
+                switch (c) {
+                    case COMMA:
+                        // Start to parse the next field
+                        unescaped.add(current.toString());
+                        current.setLength(0);
+                        break;
+                    case DOUBLE_QUOTE:
+                        if (current.length() == 0) {
+                            quoted = true;
+                            break;
+                        }
+                        // double-quote appears without being enclosed with double-quotes
+                    case LINE_FEED:
+                    case CARRIAGE_RETURN:
+                        // special characters appears without being enclosed with double-quotes
+                        throw newInvalidEscapedCsvFieldException(value, i);
+                    default:
+                        current.append(c);
+                }
+            }
+        }
+        if (quoted) {
+            throw newInvalidEscapedCsvFieldException(value, last);
+        }
+        unescaped.add(current.toString());
+        return unescaped;
+    }
+
+    /**s
      * Validate if {@code value} is a valid csv field without double-quotes.
      *
      * @throws IllegalArgumentException if {@code value} needs to be encoded with double-quotes.
@@ -368,4 +442,16 @@ public final class StringUtil {
     private static boolean isDoubleQuote(char c) {
         return c == DOUBLE_QUOTE;
     }
+
+    /**
+     * Determine if the string {@code s} ends with the char {@code c}.
+     *
+     * @param s the string to test
+     * @param c the tested char
+     * @return true if {@code s} ends with the char {@code c}
+     */
+    public static boolean endsWith(CharSequence s, char c) {
+        int len = s.length();
+        return len > 0 && s.charAt(len - 1) == c;
+    }
 }
