@@ -28,7 +28,7 @@ public class RecyclerTest {
         return new Recycler<HandledObject>(max) {
             @Override
             protected HandledObject newObject(
-                    Recycler.Handle handle) {
+                    Recycler.Handle<HandledObject> handle) {
                 return new HandledObject(handle);
             }
         };
@@ -38,28 +38,28 @@ public class RecyclerTest {
     public void testMultipleRecycle() {
         Recycler<HandledObject> recycler = newRecycler(1024);
         HandledObject object = recycler.get();
-        recycler.recycle(object, object.handle);
-        recycler.recycle(object, object.handle);
+        object.recycle();
+        object.recycle();
     }
 
     @Test
     public void testRecycle() {
         Recycler<HandledObject> recycler = newRecycler(1024);
         HandledObject object = recycler.get();
-        recycler.recycle(object, object.handle);
+        object.recycle();
         HandledObject object2 = recycler.get();
         assertSame(object, object2);
-        recycler.recycle(object2, object2.handle);
+        object2.recycle();
     }
 
     @Test
     public void testRecycleDisable() {
         Recycler<HandledObject> recycler = newRecycler(-1);
         HandledObject object = recycler.get();
-        recycler.recycle(object, object.handle);
+        object.recycle();
         HandledObject object2 = recycler.get();
         assertNotSame(object, object2);
-        recycler.recycle(object2, object2.handle);
+        object2.recycle();
     }
 
     /**
@@ -83,7 +83,7 @@ public class RecyclerTest {
         }
 
         for (int i = 0; i < objects.length; i++) {
-            recycler.recycle(objects[i], objects[i].handle);
+            objects[i].recycle();
             objects[i] = null;
         }
 
@@ -96,7 +96,7 @@ public class RecyclerTest {
     public void testRecycleAtDifferentThread() throws Exception {
         final Recycler<HandledObject> recycler = new Recycler<HandledObject>(256, 10, 2, 10) {
             @Override
-            protected HandledObject newObject(Recycler.Handle handle) {
+            protected HandledObject newObject(Recycler.Handle<HandledObject> handle) {
                 return new HandledObject(handle);
             }
         };
@@ -106,8 +106,8 @@ public class RecyclerTest {
         final Thread thread = new Thread() {
             @Override
             public void run() {
-                recycler.recycle(o, o.handle);
-                recycler.recycle(o2, o2.handle);
+                o.recycle();
+                o2.recycle();
             }
         };
         thread.start();
@@ -132,14 +132,14 @@ public class RecyclerTest {
         }
 
         for (int i = 0; i < maxCapacity; i ++) {
-            recycler.recycle(array[i], array[i].handle);
+            array[i].recycle();
         }
 
         final Thread thread = new Thread() {
             @Override
             public void run() {
                 for (int i = maxCapacity; i < array.length; i ++) {
-                    recycler.recycle(array[i], array[i].handle);
+                    array[i].recycle();
                 }
             }
         };
@@ -164,7 +164,7 @@ public class RecyclerTest {
 
         final Recycler<HandledObject> recycler = new Recycler<HandledObject>(maxCapacity, 2) {
             @Override
-            protected HandledObject newObject(Recycler.Handle handle) {
+            protected HandledObject newObject(Recycler.Handle<HandledObject> handle) {
                 instancesCount.incrementAndGet();
                 return new HandledObject(handle);
             }
@@ -185,7 +185,7 @@ public class RecyclerTest {
             @Override
             public void run() {
                 for (HandledObject object: array) {
-                    recycler.recycle(object, object.handle);
+                    object.recycle();
                 }
             }
         };
@@ -207,10 +207,14 @@ public class RecyclerTest {
     }
 
     static final class HandledObject {
-        Recycler.Handle handle;
+        Recycler.Handle<HandledObject> handle;
 
-        HandledObject(Recycler.Handle handle) {
+        HandledObject(Recycler.Handle<HandledObject> handle) {
             this.handle = handle;
         }
+
+        void recycle() {
+            handle.recycle(this);
+        }
     }
 }
