@@ -273,7 +273,8 @@ public class TrafficShapingHandlerTest extends AbstractSocketTest {
      *            ensure correct testing)
      * @throws Throwable
      */
-    private static void testTrafficShapping0(ServerBootstrap sb, Bootstrap cb, final boolean additionalExecutor,
+    private static void testTrafficShapping0(
+            ServerBootstrap sb, Bootstrap cb, final boolean additionalExecutor,
             final boolean limitRead, final boolean limitWrite, final boolean globalLimit, int[] autoRead,
             long[] minimalWaitBetween, int[] multipleMessage) throws Throwable {
 
@@ -284,7 +285,7 @@ public class TrafficShapingHandlerTest extends AbstractSocketTest {
         final ServerHandler sh = new ServerHandler(autoRead, multipleMessage);
         Promise<Boolean> promise = group.next().newPromise();
         final ClientHandler ch = new ClientHandler(promise, minimalWaitBetween, multipleMessage,
-                autoRead);
+                                                   autoRead);
 
         final AbstractTrafficShapingHandler handler;
         if (limitRead) {
@@ -353,13 +354,13 @@ public class TrafficShapingHandlerTest extends AbstractSocketTest {
 
         if (autoRead == null && minimalWaitBetween != null) {
             assertTrue("Overall Traffic not ok since > " + maxfactor + ": " + average,
-                    average <= maxfactor);
+                       average <= maxfactor);
             if (additionalExecutor) {
                 // Oio is not as good when using additionalExecutor
                 assertTrue("Overall Traffic not ok since < 0.25: " + average, average >= 0.25);
             } else {
                 assertTrue("Overall Traffic not ok since < " + minfactor + ": " + average,
-                        average >= minfactor);
+                           average >= minfactor);
             }
         }
         if (handler != null && globalLimit) {
@@ -392,7 +393,7 @@ public class TrafficShapingHandlerTest extends AbstractSocketTest {
         final Promise<Boolean> promise;
 
         ClientHandler(Promise<Boolean> promise, long[] minimalWaitBetween, int[] multipleMessage,
-                int[] autoRead) {
+                      int[] autoRead) {
             this.minimalWaitBetween = minimalWaitBetween;
             this.multipleMessage = Arrays.copyOf(multipleMessage, multipleMessage.length);
             this.promise = promise;
@@ -426,7 +427,7 @@ public class TrafficShapingHandlerTest extends AbstractSocketTest {
             loggerClient.info("Step: " + step + " Interval: " + (lastTimestamp - currentLastTime) + " compareTo "
                               + minimalWait + " (" + ar + ')');
             assertTrue("The interval of time is incorrect:" + (lastTimestamp - currentLastTime) + " not> "
-                    + minimalWait, lastTimestamp - currentLastTime >= minimalWait);
+                       + minimalWait, lastTimestamp - currentLastTime >= minimalWait);
             currentLastTime = lastTimestamp;
             step++;
             if (multipleMessage.length > step) {
