@@ -16,6 +16,7 @@
 package io.netty.channel.epoll;
 
 import io.netty.buffer.ByteBuf;
+import io.netty.buffer.ByteBufAllocator;
 import io.netty.buffer.CompositeByteBuf;
 import io.netty.channel.AddressedEnvelope;
 import io.netty.channel.ChannelFuture;
@@ -25,7 +26,6 @@ import io.netty.channel.ChannelOutboundBuffer;
 import io.netty.channel.ChannelPipeline;
 import io.netty.channel.ChannelPromise;
 import io.netty.channel.DefaultAddressedEnvelope;
-import io.netty.channel.RecvByteBufAllocator;
 import io.netty.channel.socket.DatagramChannel;
 import io.netty.channel.socket.DatagramChannelConfig;
 import io.netty.channel.socket.DatagramPacket;
@@ -481,8 +481,6 @@ public final class EpollDatagramChannel extends AbstractEpollChannel implements
     }
 
     final class EpollDatagramChannelUnsafe extends AbstractEpollUnsafe {
-
-        private RecvByteBufAllocator.Handle allocHandle;
         private final List<Object> readBuf = new ArrayList<Object>();
 
         @Override
@@ -502,6 +500,9 @@ public final class EpollDatagramChannel extends AbstractEpollChannel implements
                     EpollDatagramChannel.this.local = fd().localAddress();
                     success = true;
 
+                    // First notify the promise before notifying the handler.
+                    channelPromise.trySuccess();
+
                     // Regardless if the connection attempt was cancelled, channelActive() event should be triggered,
                     // because what happened is what happened.
                     if (!wasActive && isActive()) {
@@ -511,12 +512,11 @@ public final class EpollDatagramChannel extends AbstractEpollChannel implements
                     if (!success) {
                         doClose();
                     } else {
-                        channelPromise.setSuccess();
                         connected = true;
                     }
                 }
             } catch (Throwable cause) {
-                channelPromise.setFailure(cause);
+                channelPromise.tryFailure(cause);
             }
         }
 
@@ -524,93 +524,70 @@ public final class EpollDatagramChannel extends AbstractEpollChannel implements
         void epollInReady() {
             assert eventLoop().inEventLoop();
             if (fd().isInputShutdown()) {
-                return;
-            }
-            DatagramChannelConfig config = config();
-            boolean edgeTriggered = isFlagSet(Native.EPOLLET);
-
-            if (!readPending && !edgeTriggered && !config.isAutoRead()) {
-                // ChannelConfig.setAutoRead(false) was called in the meantime
                 clearEpollIn0();
                 return;
             }
-
-            RecvByteBufAllocator.Handle allocHandle = this.allocHandle;
-            if (allocHandle == null) {
-                this.allocHandle = allocHandle = config.getRecvByteBufAllocator().newHandle();
-            }
+            DatagramChannelConfig config = config();
+            final EpollRecvByteAllocatorHandle allocHandle = recvBufAllocHandle();
+            allocHandle.edgeTriggered(isFlagSet(Native.EPOLLET));
 
             final ChannelPipeline pipeline = pipeline();
+            final ByteBufAllocator allocator = config.getAllocator();
+            allocHandle.reset(config);
+            epollInBefore();
+
             Throwable exception = null;
             try {
-                // if edgeTriggered is used we need to read all messages as we are not notified again otherwise.
-                final int maxMessagesPerRead = edgeTriggered
-                        ? Integer.MAX_VALUE : config.getMaxMessagesPerRead();
-                int messages = 0;
-                do {
-                    ByteBuf data = null;
-                    try {
-                        data = allocHandle.allocate(config.getAllocator());
-                        int writerIndex = data.writerIndex();
-                        DatagramSocketAddress remoteAddress;
+                ByteBuf data = null;
+                try {
+                    do {
+                        data = allocHandle.allocate(allocator);
+                        allocHandle.attemptedBytesRead(data.writableBytes());
+                        final DatagramSocketAddress remoteAddress;
                         if (data.hasMemoryAddress()) {
                             // has a memory address so use optimized call
                             remoteAddress = fd().recvFromAddress(data.memoryAddress(), data.writerIndex(),
-                                    data.capacity());
+                                                                 data.capacity());
                         } else {
                             ByteBuffer nioData = data.internalNioBuffer(data.writerIndex(), data.writableBytes());
                             remoteAddress = fd().recvFrom(nioData, nioData.position(), nioData.limit());
                         }
 
                         if (remoteAddress == null) {
+                            allocHandle.lastBytesRead(-1);
+                            data.release();
+                            data = null;
                             break;
                         }
 
-                        int readBytes = remoteAddress.receivedAmount();
-                        data.writerIndex(data.writerIndex() + readBytes);
-                        allocHandle.record(readBytes);
-                        readPending = false;
+                        allocHandle.incMessagesRead(1);
+                        allocHandle.lastBytesRead(remoteAddress.receivedAmount());
+                        data.writerIndex(data.writerIndex() + allocHandle.lastBytesRead());
 
                         readBuf.add(new DatagramPacket(data, (InetSocketAddress) localAddress(), remoteAddress));
                         data = null;
-                    } catch (Throwable t) {
-                        // We do not break from the loop here and remember the last exception,
-                        // because we need to consume everything from the socket used with epoll ET.
-                        exception = t;
-                    } finally {
-                        if (data != null) {
-                            data.release();
-                        }
-                        if (!edgeTriggered && !config.isAutoRead()) {
-                            // This is not using EPOLLET so we can stop reading
-                            // ASAP as we will get notified again later with
-                            // pending data
-                            break;
-                        }
+                    } while (allocHandle.continueReading());
+                } catch (Throwable t) {
+                    if (data != null) {
+                        data.release();
                     }
-                } while (++ messages < maxMessagesPerRead || isRdHup());
+                    exception = t;
+                }
 
                 int size = readBuf.size();
                 for (int i = 0; i < size; i ++) {
+                    readPending = false;
                     pipeline.fireChannelRead(readBuf.get(i));
                 }
-
                 readBuf.clear();
+                allocHandle.readComplete();
                 pipeline.fireChannelReadComplete();
 
                 if (exception != null) {
                     pipeline.fireExceptionCaught(exception);
                 }
             } finally {
-                // Check if there is a readPending which was not processed yet.
-                // This could be for two reasons:
-                // * The user called Channel.read() or ChannelHandlerContext.read() in channelRead(...) method
-                // * The user called Channel.read() or ChannelHandlerContext.read() in channelReadComplete(...) method
-                //
-                // See https://github.com/netty/netty/issues/2254
-                if (!readPending && !config.isAutoRead()) {
-                    clearEpollIn();
-                }
+                epollInFinally(config);
             }
         }
     }
