@@ -19,7 +19,7 @@ import io.netty.buffer.ByteBufAllocator;
 import io.netty.util.Attribute;
 import io.netty.util.AttributeKey;
 import io.netty.util.concurrent.EventExecutor;
-import io.netty.util.internal.OneTimeTask;
+import io.netty.util.internal.ThrowableUtil;
 import io.netty.util.internal.logging.InternalLogger;
 import io.netty.util.internal.logging.InternalLoggerFactory;
 
@@ -143,10 +143,17 @@ public class CombinedChannelDuplexHandler<I extends ChannelInboundHandler, O ext
                         // as well
                         outboundHandler.exceptionCaught(outboundCtx, cause);
                     } catch (Throwable error) {
-                        if (logger.isWarnEnabled()) {
+                        if (logger.isDebugEnabled()) {
+                            logger.debug(
+                                    "An exception {}" +
+                                    "was thrown by a user handler's exceptionCaught() " +
+                                    "method while handling the following exception:",
+                                    ThrowableUtil.stackTraceToString(error), cause);
+                        } else if (logger.isWarnEnabled()) {
                             logger.warn(
-                                    "An exception was thrown by a user handler's " +
-                                            "exceptionCaught() method while handling the following exception:", error);
+                                    "An exception '{}' [enable DEBUG level for full stacktrace] " +
+                                    "was thrown by a user handler's exceptionCaught() " +
+                                    "method while handling the following exception:", error, cause);
                         }
                     }
                 } else {
@@ -499,7 +506,7 @@ public class CombinedChannelDuplexHandler<I extends ChannelInboundHandler, O ext
 
         @Override
         public ChannelFuture deregister(ChannelPromise promise) {
-            return ctx.deregister();
+            return ctx.deregister(promise);
         }
 
         @Override
@@ -574,12 +581,17 @@ public class CombinedChannelDuplexHandler<I extends ChannelInboundHandler, O ext
             return ctx.attr(key);
         }
 
+        @Override
+        public <T> boolean hasAttr(AttributeKey<T> key) {
+            return ctx.hasAttr(key);
+        }
+
         final void remove() {
             EventExecutor executor = executor();
             if (executor.inEventLoop()) {
                 remove0();
             } else {
-                executor.execute(new OneTimeTask() {
+                executor.execute(new Runnable() {
                     @Override
                     public void run() {
                         remove0();
