@@ -18,7 +18,6 @@ package io.netty.channel.local;
 import io.netty.channel.AbstractChannel;
 import io.netty.channel.Channel;
 import io.netty.channel.ChannelConfig;
-import io.netty.channel.ChannelException;
 import io.netty.channel.ChannelMetadata;
 import io.netty.channel.ChannelOutboundBuffer;
 import io.netty.channel.ChannelPipeline;
@@ -29,11 +28,13 @@ import io.netty.channel.SingleThreadEventLoop;
 import io.netty.util.ReferenceCountUtil;
 import io.netty.util.concurrent.Future;
 import io.netty.util.concurrent.SingleThreadEventExecutor;
-import io.netty.util.internal.EmptyArrays;
 import io.netty.util.internal.InternalThreadLocalMap;
-import io.netty.util.internal.OneTimeTask;
 import io.netty.util.internal.PlatformDependent;
+import io.netty.util.internal.ThrowableUtil;
+import io.netty.util.internal.logging.InternalLogger;
+import io.netty.util.internal.logging.InternalLoggerFactory;
 
+import java.net.ConnectException;
 import java.net.SocketAddress;
 import java.nio.channels.AlreadyConnectedException;
 import java.nio.channels.ClosedChannelException;
@@ -46,16 +47,22 @@ import java.util.concurrent.atomic.AtomicReferenceFieldUpdater;
  * A {@link Channel} for the local transport.
  */
 public class LocalChannel extends AbstractChannel {
-
+    private static final InternalLogger logger = InternalLoggerFactory.getInstance(LocalChannel.class);
     @SuppressWarnings({ "rawtypes" })
-    private static final AtomicReferenceFieldUpdater<LocalChannel, Future> FINISH_READ_FUTURE_UPDATER;
+    private static final AtomicReferenceFieldUpdater<LocalChannel, Future> FINISH_READ_FUTURE_UPDATER =
+            AtomicReferenceFieldUpdater.newUpdater(LocalChannel.class, Future.class, "finishReadFuture");
     private static final ChannelMetadata METADATA = new ChannelMetadata(false);
     private static final int MAX_READER_STACK_DEPTH = 8;
-    private static final ClosedChannelException CLOSED_CHANNEL_EXCEPTION = new ClosedChannelException();
+    private static final ClosedChannelException DO_WRITE_CLOSED_CHANNEL_EXCEPTION = ThrowableUtil.unknownStackTrace(
+            new ClosedChannelException(), LocalChannel.class, "doWrite(...)");
+    private static final ClosedChannelException DO_CLOSE_CLOSED_CHANNEL_EXCEPTION = ThrowableUtil.unknownStackTrace(
+            new ClosedChannelException(), LocalChannel.class, "doClose()");
+
+    private enum State { OPEN, BOUND, CONNECTED, CLOSED }
 
     private final ChannelConfig config = new DefaultChannelConfig(this);
     // To further optimize this we could write our own SPSC queue.
-    private final Queue<Object> inboundBuffer = PlatformDependent.newMpscQueue();
+    private final Queue<Object> inboundBuffer = PlatformDependent.newSpscQueue();
     private final Runnable readTask = new Runnable() {
         @Override
         public void run() {
@@ -77,7 +84,7 @@ public class LocalChannel extends AbstractChannel {
         }
     };
 
-    private volatile int state; // 0 - open, 1 - bound, 2 - connected, 3 - closed
+    private volatile State state;
     private volatile LocalChannel peer;
     private volatile LocalAddress localAddress;
     private volatile LocalAddress remoteAddress;
@@ -87,24 +94,14 @@ public class LocalChannel extends AbstractChannel {
     private volatile boolean writeInProgress;
     private volatile Future<?> finishReadFuture;
 
-    static {
-        @SuppressWarnings({ "rawtypes" })
-        AtomicReferenceFieldUpdater<LocalChannel, Future> finishReadFutureUpdater =
-                PlatformDependent.newAtomicReferenceFieldUpdater(LocalChannel.class, "finishReadFuture");
-        if (finishReadFutureUpdater == null) {
-            finishReadFutureUpdater =
-                AtomicReferenceFieldUpdater.newUpdater(LocalChannel.class, Future.class, "finishReadFuture");
-        }
-        FINISH_READ_FUTURE_UPDATER = finishReadFutureUpdater;
-        CLOSED_CHANNEL_EXCEPTION.setStackTrace(EmptyArrays.EMPTY_STACK_TRACE);
-    }
-
     public LocalChannel() {
         super(null);
+        config().setAllocator(new PreferHeapByteBufAllocator(config.getAllocator()));
     }
 
-    LocalChannel(LocalServerChannel parent, LocalChannel peer) {
+    protected LocalChannel(LocalServerChannel parent, LocalChannel peer) {
         super(parent);
+        config().setAllocator(new PreferHeapByteBufAllocator(config.getAllocator()));
         this.peer = peer;
         localAddress = parent.localAddress();
         remoteAddress = peer.localAddress();
@@ -137,12 +134,12 @@ public class LocalChannel extends AbstractChannel {
 
     @Override
     public boolean isOpen() {
-        return state < 3;
+        return state != State.CLOSED;
     }
 
     @Override
     public boolean isActive() {
-        return state == 2;
+        return state == State.CONNECTED;
     }
 
     @Override
@@ -181,16 +178,16 @@ public class LocalChannel extends AbstractChannel {
             // See https://github.com/netty/netty/issues/2144
             final LocalChannel peer = this.peer;
             registerInProgress = true;
-            state = 2;
+            state = State.CONNECTED;
 
-            peer.remoteAddress = parent().localAddress();
-            peer.state = 2;
+            peer.remoteAddress = parent() == null ? null : parent().localAddress();
+            peer.state = State.CONNECTED;
 
             // Always call peer.eventLoop().execute() even if peer.eventLoop().inEventLoop() is true.
             // This ensures that if both channels are on the same event loop, the peer's channelActive
             // event is triggered *after* this channel's channelRegistered event, so that this channel's
             // pipeline is fully initialized by ChannelInitializer before any channelRead events.
-            peer.eventLoop().execute(new OneTimeTask() {
+            peer.eventLoop().execute(new Runnable() {
                 @Override
                 public void run() {
                     registerInProgress = false;
@@ -212,7 +209,7 @@ public class LocalChannel extends AbstractChannel {
         this.localAddress =
                 LocalChannelRegistry.register(this, this.localAddress,
                         localAddress);
-        state = 1;
+        state = State.BOUND;
     }
 
     @Override
@@ -223,7 +220,7 @@ public class LocalChannel extends AbstractChannel {
     @Override
     protected void doClose() throws Exception {
         final LocalChannel peer = this.peer;
-        if (state <= 2) {
+        if (state != State.CLOSED) {
             // Update all internal state before the closeFuture is notified.
             if (localAddress != null) {
                 if (parent() == null) {
@@ -234,52 +231,59 @@ public class LocalChannel extends AbstractChannel {
 
             // State change must happen before finishPeerRead to ensure writes are released either in doWrite or
             // channelRead.
-            state = 3;
+            state = State.CLOSED;
+
+            // Preserve order of event and force a read operation now before the close operation is processed.
+            finishPeerRead(this);
 
             ChannelPromise promise = connectPromise;
             if (promise != null) {
                 // Use tryFailure() instead of setFailure() to avoid the race against cancel().
-                promise.tryFailure(CLOSED_CHANNEL_EXCEPTION);
+                promise.tryFailure(DO_CLOSE_CLOSED_CHANNEL_EXCEPTION);
                 connectPromise = null;
             }
-
-            // To preserve ordering of events we must process any pending reads
-            if (writeInProgress && peer != null) {
-                finishPeerRead(peer);
-            }
         }
 
-        if (peer != null && peer.isActive()) {
+        if (peer != null) {
+            this.peer = null;
             // Need to execute the close in the correct EventLoop (see https://github.com/netty/netty/issues/1777).
             // Also check if the registration was not done yet. In this case we submit the close to the EventLoop
             // to make sure its run after the registration completes (see https://github.com/netty/netty/issues/2144).
-            if (peer.eventLoop().inEventLoop() && !registerInProgress) {
-                doPeerClose(peer, peer.writeInProgress);
+            EventLoop peerEventLoop = peer.eventLoop();
+            final boolean peerIsActive = peer.isActive();
+            if (peerEventLoop.inEventLoop() && !registerInProgress) {
+                peer.tryClose(peerIsActive);
             } else {
-                // This value may change, and so we should save it before executing the Runnable.
-                final boolean peerWriteInProgress = peer.writeInProgress;
                 try {
-                    peer.eventLoop().execute(new OneTimeTask() {
+                    peerEventLoop.execute(new Runnable() {
                         @Override
                         public void run() {
-                            doPeerClose(peer, peerWriteInProgress);
+                            peer.tryClose(peerIsActive);
                         }
                     });
-                } catch (RuntimeException e) {
-                    // The peer close may attempt to drain this.inboundBuffers. If that fails make sure it is drained.
+                } catch (Throwable cause) {
+                    logger.warn("Releasing Inbound Queues for channels {}-{} because exception occurred!",
+                            this, peer, cause);
                     releaseInboundBuffers();
-                    throw e;
+                    if (peerEventLoop.inEventLoop()) {
+                        peer.releaseInboundBuffers();
+                    } else {
+                        // inboundBuffers is a SPSC so we may leak if the event loop is shutdown prematurely or rejects
+                        // the close Runnable but give a best effort.
+                        peer.close();
+                    }
+                    PlatformDependent.throwException(cause);
                 }
             }
-            this.peer = null;
         }
     }
 
-    private void doPeerClose(LocalChannel peer, boolean peerWriteInProgress) {
-        if (peerWriteInProgress) {
-            finishPeerRead0(this);
+    private void tryClose(boolean isActive) {
+        if (isActive) {
+            unsafe().close(unsafe().voidPromise());
+        } else {
+            releaseInboundBuffers();
         }
-        peer.unsafe().close(peer.unsafe().voidPromise());
     }
 
     @Override
@@ -320,20 +324,25 @@ public class LocalChannel extends AbstractChannel {
         } else {
             try {
                 eventLoop().execute(readTask);
-            } catch (RuntimeException e) {
-                releaseInboundBuffers();
-                throw e;
+            } catch (Throwable cause) {
+                logger.warn("Closing Local channels {}-{} because exception occurred!", this, peer, cause);
+                close();
+                peer.close();
+                PlatformDependent.throwException(cause);
             }
         }
     }
 
     @Override
     protected void doWrite(ChannelOutboundBuffer in) throws Exception {
-        if (state < 2) {
+        switch (state) {
+        case OPEN:
+        case BOUND:
             throw new NotYetConnectedException();
-        }
-        if (state > 2) {
-            throw CLOSED_CHANNEL_EXCEPTION;
+        case CLOSED:
+            throw DO_WRITE_CLOSED_CHANNEL_EXCEPTION;
+        case CONNECTED:
+            break;
         }
 
         final LocalChannel peer = this.peer;
@@ -348,11 +357,11 @@ public class LocalChannel extends AbstractChannel {
                 try {
                     // It is possible the peer could have closed while we are writing, and in this case we should
                     // simulate real socket behavior and ensure the write operation is failed.
-                    if (peer.state == 2) {
+                    if (peer.state == State.CONNECTED) {
                         peer.inboundBuffer.add(ReferenceCountUtil.retain(msg));
                         in.remove();
                     } else {
-                        in.remove(CLOSED_CHANNEL_EXCEPTION);
+                        in.remove(DO_WRITE_CLOSED_CHANNEL_EXCEPTION);
                     }
                 } catch (Throwable cause) {
                     in.remove(cause);
@@ -382,7 +391,7 @@ public class LocalChannel extends AbstractChannel {
     private void runFinishPeerReadTask(final LocalChannel peer) {
         // If the peer is writing, we must wait until after reads are completed for that peer before we can read. So
         // we keep track of the task, and coordinate later that our read can't happen until the peer is done.
-        final Runnable finishPeerReadTask = new OneTimeTask() {
+        final Runnable finishPeerReadTask = new Runnable() {
             @Override
             public void run() {
                 finishPeerRead0(peer);
@@ -394,19 +403,21 @@ public class LocalChannel extends AbstractChannel {
             } else {
                 peer.eventLoop().execute(finishPeerReadTask);
             }
-        } catch (RuntimeException e) {
-            peer.releaseInboundBuffers();
-            throw e;
+        } catch (Throwable cause) {
+            logger.warn("Closing Local channels {}-{} because exception occurred!", this, peer, cause);
+            close();
+            peer.close();
+            PlatformDependent.throwException(cause);
         }
     }
 
     private void releaseInboundBuffers() {
-        for (;;) {
-            Object o = inboundBuffer.poll();
-            if (o == null) {
-                break;
-            }
-            ReferenceCountUtil.release(o);
+        assert eventLoop() == null || eventLoop().inEventLoop();
+        readInProgress = false;
+        Queue<Object> inboundBuffer = this.inboundBuffer;
+        Object msg;
+        while ((msg = inboundBuffer.poll()) != null) {
+            ReferenceCountUtil.release(msg);
         }
     }
 
@@ -444,7 +455,7 @@ public class LocalChannel extends AbstractChannel {
                 return;
             }
 
-            if (state == 2) {
+            if (state == State.CONNECTED) {
                 Exception cause = new AlreadyConnectedException();
                 safeSetFailure(promise, cause);
                 pipeline().fireExceptionCaught(cause);
@@ -457,7 +468,7 @@ public class LocalChannel extends AbstractChannel {
 
             connectPromise = promise;
 
-            if (state != 1) {
+            if (state != State.BOUND) {
                 // Not bound yet and no localAddress specified - get one.
                 if (localAddress == null) {
                     localAddress = new LocalAddress(LocalChannel.this);
@@ -476,7 +487,7 @@ public class LocalChannel extends AbstractChannel {
 
             Channel boundChannel = LocalChannelRegistry.get(remoteAddress);
             if (!(boundChannel instanceof LocalServerChannel)) {
-                Exception cause = new ChannelException("connection refused");
+                Exception cause = new ConnectException("connection refused: " + remoteAddress);
                 safeSetFailure(promise, cause);
                 close(voidPromise());
                 return;
